; Module name: ALFAGRAF.ASM                   Library:   GRAPHSOL.LIB
; Created: March, 1992
;********************************************************************
;********************************************************************
;                          ALFAGRAF.ASM
;********************************************************************
;********************************************************************
; Description:
; Library module for the following functions:
;     1. Video services using direct buffer access
;     2. Aids for alphanumeric graphics programming
;
; Note: The library routines that are contained in this module can
;       be used from the high level languages C, Pascal, and BASIC
;       as well as in programs coded directly in Assembly Language.
;       Appendix B explains interfacing the library routines with
;       these languages.
;*******************************************************************
;                          ROUTINES
;*******************************************************************
; Public procedures in this module:
; NAME             DESCRIPTION
; SET_CURSOR       Set system cursor to screen position
; GET_CURSOR       Read current cursor position
; CURSOR_OFF       Turn-off system cursor
; CLEAR_SCREEN     Clear video display and set display attribute
; CLEAR_AREA       Clear rectangular area and set attribute
; SHOW_BLOCK       Display a formate message block
; REPEAT_HOR       Repeat a character horizontally
; REPEAT_VER       Repeat a character vertically
; DRAW_BOX         Display a screen box using alpha characters
; SHADE_BOX        Display a rectangular, shaded area
; BOX_LINE_HOR     Display a horizontal line inside a screen box
; BOX_LINE_VER     Display a vertical line inside a screen box
;
; Local procedures
; ES_TO_VIDEO      Set ES register to video buffer base
; GET_OFFSET       Set DI to a row/column offset
;
;*******************************************************************
;                         W A R N I N G S !
;*******************************************************************
; 1. Some routines in this module use the ES register as a pointer
;    to the video buffer base. If the user's software requires ES
;    it must preserve it before calling any routine in this library
;*******************************************************************
;*******************|
;      PUBLICS      |
;*******************|
PUBLIC  SET_CURSOR
PUBLIC  GET_CURSOR
PUBLIC  CURSOR_OFF
PUBLIC  CLEAR_SCREEN
PUBLIC  CLEAR_AREA
PUBLIC  SHOW_BLOCK
PUBLIC  REPEAT_HOR
PUBLIC  REPEAT_VER
PUBLIC  DRAW_BOX
PUBLIC  SHADE_BOX
PUBLIC  BOX_LINE_HOR
PUBLIC  BOX_LINE_VER
;
;******************************************************************
;
P_CODE  SEGMENT PUBLIC
        ASSUME  CS:P_CODE
;
;*******************************************************************
;                        local procedures
;*******************************************************************
GET_OFFSET      PROC    FAR
; Set DI to the offset address in the video buffer that corresponds
; with a certain screen row and column position
; On entry:
;         DH = screen row (0 to 24)
;         DL = screen column (0 to 79)
; On exit:
;         DI --> buffer offset (row * 160) + (column * 2)
;         carry clear
; If input out of range carry flag set and AX destroyed.
;
        PUSH    AX              ; Save accumulator
        PUSH    DX              ; and DX
        MOV     AL,160          ; Bytes per row
        MUL     DH              ; AX = 160 * DH
        MOV     DI,AX           ; Save partial result in DI
        MOV     AL,2            ; Bytes per column
        MUL     DL              ; AX = DL * 2
        ADD     DI,AX           ; Add in previous result
        POP     DX              ; Restore DX
        POP     AX              ; And accumulator
        CLC                     ; No error detection
        RET
GET_OFFSET      ENDP
;*******************************************************************
;
ES_TO_VIDEO     PROC    FAR
; Set ES register to the base address of the video buffer according
; to the video mode stored in BIOS address 0040:0049
; On entry:
;         nothing
; On exit:
;         ES --> segment address of video buffer for alphanumeric
;                display
;         carry clear
        PUSH    AX              ; Save register
        MOV     AX,0040H        ; Segment base of data area
        MOV     ES,AX           ; To extra segment
        MOV     AL,ES:[0049H]   ; Current mode
        CMP     AL,7            ; Code for monochrome card
        JE      SET_MONO_BASE   ; Set base for mono card
;********************|
;     color card     |
;********************|
        MOV     AX,0B800H       ; Color system base
        MOV     ES,AX           ; to ES
        POP     AX              ; Restore entry register
        CLC                     ; No error detection
        RET
;*********************|
;  monochrome card    |
;*********************|
SET_MONO_BASE:
        MOV     AX,0B000H       ; Video base in monochrome systems
        MOV     ES,AX           ; To ES
        POP     AX              ; Restore entry register
        CLC                     ; No error detection
        RET
ES_TO_VIDEO     ENDP
;******************************************************************
;                         public procedures
;******************************************************************
SET_CURSOR      PROC    FAR
; Set system cursor to a row and column coordinate
; On entry:
;         DH = screen row (range 0 to 24)
;         DL = screen column (range 0 to 79)
; Note: routine does not test for valid parameters
; On exit:
;         cursor set
;
        PUSH    AX              ; Save entry registers
        PUSH    BX
        PUSH    DX
        MOV     AH,2            ; Service request number
        MOV     BH,0            ; Assume display page No. 0
        INT     10H
        POP     DX              ; Restore registers
        POP     BX
        POP     AX
        CLC                     ; No error
        RET
SET_CURSOR      ENDP
;******************************************************************
;
GET_CURSOR      PROC    FAR
; Read coordinates of current cursor position and start and
; end lines of cursor size
; On entry:
;         BH = active page number (normally 0)
; On exit:
;         CH = starting line of cursor size
;         CL = ending line of cursor size
;         DH = row number of current cursor position
;         DL = column number of current cursor position
;         carry clear
;
        PUSH    AX              ; Save entry registers
        MOV     AH,3            ; Service request number
        INT     10H
        POP     AX              ; Restore AX
        CLC                     ; No error
        RET
GET_CURSOR      ENDP
;*******************************************************************
;
CURSOR_OFF      PROC    FAR
; Turn off cursor by setting the cursor address outside of the
; current display page
        PUSH    AX              ; Save entry registers
        PUSH    BX
        PUSH    DX
        MOV     DH,25           ; y coordinate out of page
        MOV     DL,0
        MOV     AH,2            ; Service request number
        MOV     BH,0            ; Assume display page No. 0
        INT     10H
        POP     DX              ; Restore registers
        POP     BX
        POP     AX
        CLC                     ; No error
CURSOR_OFF      ENDP
;*******************************************************************
;
CLEAR_SCREEN    PROC    FAR
; Clear the video display with a block move
; On entry:
;         AL = attribute to use
;              Use AL = 7 for normal attribute
; On exit:
;         Screen cleared, attribute set
        CALL    ES_TO_VIDEO     ; Local procedure
        PUSH    AX              ; Save entry registers
        PUSH    CX
        PUSH    DI
        XCHG    AL,AH           ; Move attribute to AH
        MOV     DI,0            ; Start at offset 0 in buffer
        MOV     AL,20H          ; Blank byte
        MOV     CX,2000         ; Repeat 2000 times
        CLD                     ; Forward direction in memory
        REP     STOSW           ; Store 2000 words in buffer
        POP     DI
        POP     CX
        POP     AX              ; Restore entry registers
        CLC                     ; No error detection
        RET
CLEAR_SCREEN    ENDP
;*******************************************************************
;
CLEAR_AREA      PROC    FAR
; Clear a rectangular screen area and set the display attribute
; On entry:
;         AL = character to use in cleared area
;              Use AL = 20H for blanking function
;                  AL = 0 to preserve displayed characters
;         AH = attribute to use
;              Use AH = 7 for normal attribute
;         CH = top left corner, screen row (0 to 24)
;         CL = top left corner, screen column (0 to 79)
;         DH = bottom right corner, screen row (0 to 24)
;         DL = bottom right corner, screen column (0 to 79)
; Legal parameters:
;         The values assigned to the bottom-right corner of the
;         area must be at least one (row or column) greater than
;         those for the top-left corner. If the parameters are
;         not legal, the procedure will set the carry flag and exit
; On exit:
;         Carry clear (no error), shaded box displayed
;         Carry set (error) invalid or illegal input
; On exit:
;         Screen area cleared, attribute set
;**********************|
;    test for valid    |
;     parameters       |
;**********************|
; See valid data values in routine header
        CMP     CH,24           ; Test screen row range
        JBE     OK_ROW_A        ; Go if in range
;**********************|
;  invalid or illegal  |
;      input           |
;**********************|
BAD_AREA:
        STC                     ; Error flag set
        RET
OK_ROW_A:
        CMP     DH,24           ; Row range for end point
        JA      BAD_AREA        ; Go if in range
        CMP     CL,79           ; Test first column range
        JA      BAD_AREA
        CMP     DL,79           ; Test column range for end point
        JA      BAD_AREA
; At this point row and column ranges are valid. Now test for
; legal parameters
        CMP     CH,DH           ; Start row and end row
        JBE     OK_ROW_ATTR     ; Go if legal row parameters
        JMP     BAD_AREA        ; Exit if not
OK_ROW_ATTR:
        CMP     CL,DL           ; Start column and end column
        JBE     OK_COL_ATTR     ; Go if legal column parameters
        JMP     BAD_AREA        ; Error exit
;**********************|
;  store coordinates   |
;**********************|
; Area coordinates are stored in the code segment variables
; defined for the DRAW_BOX procedure
OK_COL_ATTR:
        MOV     CS:LEFT_CORNER,CX       ; Save coordinates
        CALL    ES_TO_VIDEO             ; Local procedure
; Calculate and store horizontal dimension of shaded box
        SUB     DL,CL           ; End column minus start column
        MOV     CS:HORZ_LEN,DL  ; Store horizontal length
; Calculate and store vertical dimensions of box
        SUB     DH,CH           ; End row minus start row
        MOV     CS:VERT_LEN,DH  ; Store vertical dimension
; Attribute is in AH, blanking character in AL
; If AL = 0 then screen characters are not changed
;**********************|
;    set attribute     |
;    and character     |
;**********************|
; Get top left parameters of area
        MOV     DX,CS:LEFT_CORNER       ; Restore left coordinates
; Display horizontal lines using local procedure
ATTR_ROW:
        CALL    GET_OFFSET      ; DI --> video buffer offset
        MOV     CL,CS:HORZ_LEN  ; Get horizontal byte count
        MOV     CH,0            ; Clear high byte of counter
; Horizontal line of shade characters displayed
; Test for preserve text mode (AL = 0)
        CMP     AL,0            ; Entry code passed by caller
        JE      NO_CHAR_DISPLAY ; Go if special entry code
;*********************|
;   attribute in AH   |
;*********************|
CHAR_CHANGE:
        CLD                     ; Forward direction
        REP     STOSW           ; Store AX by [DI], bump twice
                                ; and repeat
        JMP     NEXT_ATTR_ROW
;********************|
; change attribute   |
;       only         |
;********************|
NO_CHAR_DISPLAY:
        INC     DI              ; Skip character byte
        MOV     ES:[DI],AH      ; Store attribute byte
        INC     DI              ; Bump pointer
        LOOP    NO_CHAR_DISPLAY
NEXT_ATTR_ROW:
        INC     DH              ; Bump to next row
        DEC     CS:VERT_LEN     ; Decrement row counter
        JNZ     ATTR_ROW        ; Continue until last row
        RET
CLEAR_AREA      ENDP
;*****************************************************************
;
SHOW_BLOCK      PROC    FAR
; Display a pre-formatted block message
; On entry:
;          DS:SI --> start of block message
; Message format:
;       Offset:         Function:
;              0 ------ Display row for start of message
;              1 ------ Display column for start of message
;              2 ------ Attribute for block
; Embedded codes:
;              00H = End of message
;              FFH = End of row
; On exit:
;        carry clear
;
        CALL    ES_TO_VIDEO
        PUSH    AX              ; Save entry registers
        PUSH    CX
        PUSH    DX
        MOV     DH,[SI]         ; Get row
        INC     SI              ; Point to start column byte
        MOV     DL,[SI]         ; Column
        MOV     CL,DL           ; Save start column in CL
        CALL    GET_OFFSET      ; Set DI to offset in buffer
        INC     SI              ; Point to attribute byte
        MOV     AH,[SI]         ; To AH
        INC     SI
GET_AND_SHOW:
        MOV     AL,[SI]         ; Character
;*********************|
;   test for control  |
;      characters     |
;*********************|
        CMP     AL,00H          ; Terminator 
        JE      SHOW_END
        CMP     AL,0FFH         ; End of line
        JE      LINE_END
; Display it by storing character in video buffer
        CLD                     ; Forward direction
        STOSW                   ; Store word
        INC     DL              ; Next column
;*********************|
;   bump pointer      |
;*********************|
BUMP_PTRS:
        CALL    GET_OFFSET
        INC     SI              ; Bump message pointer
        JMP     GET_AND_SHOW
;********************|
;    end of line     |
;********************|
LINE_END:
        INC     DH              ; Bump row pointer
        MOV     DL,CL           ; Reset column to start column
        JMP     BUMP_PTRS
;********************|
;   end of routine   |
;********************|
SHOW_END:
        POP     DX              ; Restore registers
        POP     CX
        POP     AX
        CLC                     ; No error detection
        RET
SHOW_BLOCK      ENDP
;*******************************************************************
;
REPEAT_HOR      PROC    FAR
; Display CX number of characters
; On entry:
;         AL = Character to display
;         AH = Attribute (if AH = 0 then attribute is not changed)
;         DH = screen row
;         DL = screen column
;         CX = number of characters to repeat
; On exit:
;        carry clear if no error
;        carry set if error. AX destroyed
;
        CALL    ES_TO_VIDEO             ; Local procedure
        CALL    GET_OFFSET
        JNC     OK_SHOW_CX      ; No error reported
        RET                     ; Error return
OK_SHOW_CX:
        CMP     AH,0            ; Test for attribute code
        JE      USE_BYTE_MOVE   ; Use installed attributes
;*********************|
;   attribute in AH   |
;*********************|
        CLD                     ; Forward direction
        REP     STOSW           ; Store AX by [DI], bump twice
                                ; and repeat
        CLC                     ; No error detection
        RET
;********************|
;  current attribute |
;********************|
USE_BYTE_MOVE:
        CLD                     ; Forward direction
        STOSB                   ; Store AL by [DI] and bump once
        INC     DI              ; Skip attribute byte
        LOOP    USE_BYTE_MOVE
        CLC                     ; No error detection
        RET
REPEAT_HOR      ENDP
;*******************************************************************
;
REPEAT_VER      PROC    FAR
; Display CX number of characters vertically using current attribute
; On entry:
;         AL = Character to display
;         DH = screen row
;         DL = screen column
;         CX = number of characters to repeat
; On exit:
;        carry clear
        CALL    ES_TO_VIDEO             ; Local procedure
        CALL    GET_OFFSET
        JNC     SET_VERTS       ; No error reported
        RET                     ; Error return
;*********************|
;   attribute in AH   |
;*********************|
SET_VERTS:
        CLD                     ; Forward
        STOSB                   ; Store AL by [DI] and bump once
        INC     DI              ; Skip attribute
        ADD     DI,158          ; Bump to next row
        LOOP    SET_VERTS
        CLC                     ; No error detection
        RET
REPEAT_VER      ENDP
;******************************************************************
;
; CS DATA STORAGE FOR DRAW_BOX PROCEDURE
LEFT_CORNER     DW      0       ; Coordinates of left box corner
                                ; (Also used by BOX_LINE procedure)
HORZ_LEN        DB      0       ; Number of horizontal bytes
VERT_LEN        DB      0       ; Number of vertical bytes
; Box drawing symbols for single and double line boxes
; Default values are for double box
ULC_SYM         DB      0C9H    ; Top-left corner
URC_SYM         DB      0BBH    ; Top-right corner
LLC_SYM         DB      0C8H    ; Bottom-left corner
LRC_SYM         DB      0BCH    ; Bottom-right corner
V_SYM           DB      0BAH    ; Vertical bar
H_SYM           DB      0CDH    ; Horizontal bar
; Initialization strings for single and double-line boxes
SINGLE_LINE     DB      0DAH,0BFH,0C0H,0D9H,0B3H,0C4H
DOUBLE_LINE     DB      0C9H,0BBH,0C8H,0BCH,0BAH,0CDH
;
DRAW_BOX        PROC    FAR
; Draw a screen box using alphanumeric characters
; On entry:
;         AL = 1 to draw a single-line box (default)
;         AL = 2 to draw a double-line box
;         CH = box top left corner, screen row (0 to 24)
;         CL = box top left corner, screen column (0 to 79)
;         DH = box bottom right corner, screen row (0 to 24)
;         DL = box bottom right corner, screen column (0 to 79)
; Legal parameters:
;         The values assigned to the bottom-right corner of the
;         box must be at least one (row or column) greater than
;         those for the top-left corner. If the box parameters are
;         not legal the procedure will set the carry flag and exit
; On exit:
;         Carry clear (no error), screen box displayed
;         Carry set (error) invalid or illegal input
; Note:
;         The procedure uses data variables in CS
;
;**********************|
;    test for valid    |
;     parameters       |
;**********************|
; See valid data values in routine header
        CMP     CH,24           ; Test screen row range
        JBE     OK_ROW1         ; Go if in range
;**********************|
;  invalid or illegal  |
;      input           |
;**********************|
BAD_VALUES:
        STC                     ; Error flag set
        RET
OK_ROW1:
        CMP     DH,24           ; Row range for end point
        JA      BAD_VALUES      ; Go if in range
        CMP     CL,79           ; Test first column range
        JA      BAD_VALUES
        CMP     DL,79           ; Test column range for end point
        JA      BAD_VALUES
; At this point row and column ranges are valid. Now test for
; legal parameters
        CMP     CH,DH           ; Start row and end row
        JB      OK_ROW_PARAM    ; Go if legal row parameters
        JMP     BAD_VALUES      ; Exit if not
OK_ROW_PARAM:
        CMP     CL,DL           ; Start column and end column
        JB      OK_COL_PARAM    ; Go if legal column parameters
        JMP     BAD_VALUES      ; Error exit
;**********************|
;  select box type     |
;**********************|
; The entry value in AL determines of the box is drawn using
; single or double line. AL = 2 for double lines
OK_COL_PARAM:
        PUSH    CX              ; Save left corner coordinates
        LEA     SI,CS:SINGLE_LINE       ; Default to single-line
        CMP     AL,2            ; Test for double line box
        JNE     INIT_SYM_TABLE  ; Go if not 2
        LEA     SI,CS:DOUBLE_LINE       ; Switch pointer
INIT_SYM_TABLE:
        MOV     CX,6            ; Symbols to move
        LEA     DI,CS:ULC_SYM   ; Start of symbols area
MOVE_SIX:
        MOV     AL,CS:[SI]      ; Get symbol from table
        MOV     CS:[DI],AL      ; Move into symbols area
        INC     SI              ; Bump pointers
        INC     DI
        LOOP    MOVE_SIX
        POP     CX              ; Restore left corner
;**********************|
;      store box       |
;     coordinates      |
;**********************|
        MOV     CS:LEFT_CORNER,CX       ; Save coordinates
        CALL    ES_TO_VIDEO             ; Local procedure
; Calculate and store horizontal dimension of box
        SUB     DL,CL           ; End column minus start column
        MOV     CS:HORZ_LEN,DL  ; Store horizontal length
; Calculate and store vertical dimensions of box
        SUB     DH,CH           ; End row minus start row
        MOV     CS:VERT_LEN,DH  ; Store vertical dimension
;**********************|
; draw horizontal line |
;      at box top      |
;**********************|
; Draw top-left corner
        MOV     DX,CS:LEFT_CORNER       ; Restore left coordinates
        CALL    GET_OFFSET      ; DI --> video buffer offset
; Display top-left corner symbol
        MOV     AL,CS:ULC_SYM           ; Symbol to AL
        MOV     BYTE PTR ES:[DI],AL
        INC     DL              ; Bump column counter
; Display horizontal line using local procedure
        MOV     AH,0            ; Do not change attribute
        MOV     CL,CS:HORZ_LEN  ; Get horizontal byte count
        MOV     CH,0            ; Clear high byte of counter
        MOV     AL,CS:H_SYM     ; Double line character
        CALL    REPEAT_HOR      ; Local display procedure
; Display top-right corner symbol
        MOV     AL,CS:URC_SYM           ; Symbol to AL
        MOV     BYTE PTR ES:[DI],AL
;**********************|
;  draw vertical line  |
;     at box left      |
;**********************|
        MOV     DX,CS:LEFT_CORNER       ; Left corner coordinates
        INC     DH              ; Skip top row
        CALL    GET_OFFSET      ; DI --> video buffer offset
        MOV     CL,CS:VERT_LEN  ; Get vertical byte count
        MOV     CH,0            ; Clear high byte of counter
        MOV     AL,CS:V_SYM     ; Vertical line symbol
        CALL    REPEAT_VER      ; Local display procedure
;**********************|
;  draw vertical line  |
;     at box right     |
;**********************|
        MOV     DX,CS:LEFT_CORNER       ; Left corner coordinates
        INC     DH              ; Skip top row
        MOV     AL,CS:HORZ_LEN  ; Get box horizontal dimension
        ADD     DL,AL           ; Add to left corner coordinate
        INC     DL              ; Bump once
        CALL    GET_OFFSET      ; DI --> video buffer offset
        MOV     CL,CS:VERT_LEN  ; Get vertical byte count
        MOV     CH,0            ; Clear high byte of counter
        MOV     AL,CS:V_SYM     ; Vertical line symbol
        CALL    REPEAT_VER      ; Local display procedure
;**********************|
; draw horizontal line |
;    at box bottom     |
;**********************|
; Draw bottom-left corner
        MOV     DX,CS:LEFT_CORNER       ; Restore left coordinates
        ADD     DH,CS:VERT_LEN  ; Add vertical dimension
        INC     DH              ; And one more row
        CALL    GET_OFFSET      ; DI --> video buffer offset
; Symbol for lower left corner
        MOV     AL,CS:LLC_SYM           ; Symbol to AL
        MOV     BYTE PTR ES:[DI],AL
        INC     DL              ; Bump column counter
; Display horizontal line using local procedure
        MOV     AH,0            ; Do not change attribute
        MOV     CL,CS:HORZ_LEN  ; Get horizontal byte count
        MOV     CH,0            ; Clear high byte of counter
        MOV     AL,CS:H_SYM     ; Double line character
        CALL    REPEAT_HOR      ; Local display procedure
; Display symbol for lower-right corner
        MOV     AL,CS:LRC_SYM           ; Symbol to AL
        MOV     BYTE PTR ES:[DI],AL
;
        RET
DRAW_BOX        ENDP
;******************************************************************
;
; CS DATA STORAGE FOR SHADE_BOX PROCEDURE
SHADE_CHAR      DB      0B0H    ; Shading codes as follows:
                                ; B0H = lighter shade (default)
                                ; B1H = medium shade
                                ; B2H = darker shade
;
SHADE_BOX       PROC    FAR
; Draw a rectangular area using the shading characters B0H, B1H,
; and B2H
; On entry:
;         AL = 1 for lighter shade box
;         AL = 2 for medium shade box
;         AL = 3 for darker shade box
;         CH = box top left corner, screen row (0 to 24)
;         CL = box top left corner, screen column (0 to 79)
;         DH = box bottom right corner, screen row (0 to 24)
;         DL = box bottom right corner, screen column (0 to 79)
; Legal parameters:
;         The values assigned to the bottom-right corner of the
;         box must be at least one (row or column) greater than
;         those for the top-left corner. If the box parameters are
;         not legal the procedure will set the carry flag and exit
; On exit:
;         Carry clear (no error), shaded box displayed
;         Carry set (error) invalid or illegal input
; Note:
;         The procedure uses data variables in CS
;
;**********************|
;    test for valid    |
;     parameters       |
;**********************|
; See valid data values in routine header
        CMP     CH,24           ; Test screen row range
        JBE     OK_ROWS         ; Go if in range
;**********************|
;  invalid or illegal  |
;      input           |
;**********************|
BAD_SHADES:
        STC                     ; Error flag set
        RET
OK_ROWS:
        CMP     DH,24           ; Row range for end point
        JA      BAD_SHADES      ; Go if in range
        CMP     CL,79           ; Test first column range
        JA      BAD_SHADES
        CMP     DL,79           ; Test column range for end point
        JA      BAD_SHADES
; At this point row and column ranges are valid. Now test for
; legal parameters
        CMP     CH,DH           ; Start row and end row
        JB      OK_ROW_SHADES   ; Go if legal row parameters
        JMP     BAD_SHADES      ; Exit if not
OK_ROW_SHADES:
        CMP     CL,DL           ; Start column and end column
        JB      OK_COL_SHADES   ; Go if legal column parameters
        JMP     BAD_SHADES      ; Error exit
;**********************|
;  select shade char   |
;**********************|
; The entry value in AL determines the shading character (see
; note in routine header comments)
OK_COL_SHADES:
        MOV     AH,0B0H         ; Assume lighter shade
        CMP     AL,1            ; Is it this shade?
        JE      SET_SHADE       ; Go if lighter shade requested
        MOV     AH,0B1H         ; Switch to medium shade
        CMP     AL,2            ; Is it this shade?
        JE      SET_SHADE       ; Go if medium requested
        MOV     AH,0B2H         ; Set darker shade
SET_SHADE:
        MOV     CS:SHADE_CHAR,AH        ; Store shade in variable
;**********************|
;  store coordinates   |
;**********************|
; Shaded box coordinates are stored in the code segment variables
; defined for the DRAW_BOX procedure
        MOV     CS:LEFT_CORNER,CX       ; Save coordinates
        CALL    ES_TO_VIDEO             ; Local procedure
; Calculate and store horizontal dimension of shaded box
        SUB     DL,CL           ; End column minus start column
        MOV     CS:HORZ_LEN,DL  ; Store horizontal length
; Calculate and store vertical dimensions of box
        SUB     DH,CH           ; End row minus start row
        MOV     CS:VERT_LEN,DH  ; Store vertical dimension
;**********************|
;    draw shaded box   |
;**********************|
; Draw top-left corner
        MOV     DX,CS:LEFT_CORNER       ; Restore left coordinates
; Display horizontal lines using local procedure
SHADE_ROW:
        CALL    GET_OFFSET      ; DI --> video buffer offset
        MOV     AH,0            ; Do not change attribute
        MOV     CL,CS:HORZ_LEN  ; Get horizontal byte count
        MOV     CH,0            ; Clear high byte of counter
        MOV     AL,CS:SHADE_CHAR        ; Stored shade
        CALL    REPEAT_HOR      ; Local display procedure
; Horizontal line of shade characters displayed
        INC     DH              ; Bump to next row
        DEC     CS:VERT_LEN     ; Decrement row counter
        JNZ     SHADE_ROW       ; Continue until last row
        RET
SHADE_BOX       ENDP
;*****************************************************************
;
; CS DATA FOR THE BOX_LINE_HOR PROCEDURE
; Left-replacement table. Default values for double line case. See
; replacement tables in the routine header comment
FOUND_LEFT      DB      0B3H,0BAH,0B4H,0B5H,0B6H,0B9H
REPLACE_LEFT    DB      0C6H,0CCH,0C5H,0D8H,0D7H,0CEH
LEFT1_CODES     DW      0C7C3H
LEFT2_CODES     DW      0CCC6H
;
FOUND_RIGHT     DB      0B3H,0BAH,0C3H,0C6H,0C7H,0CCH
REPLACE_RIGHT   DB      0B5H,0B9H,0C5H,0D8H,0D7H,0CEH
RIGHT1_CODES    DW      0B6B4H
RIGHT2_CODES    DW      0B9B5H
;
LINE_CODE       DB      0CDH    ; Default is double line
                                ; Single line code is C4H
;
BOX_LINE_HOR    PROC    FAR
; Draw a horizontal line inside an existing screen box
; On entry:
;         AL = 1 to draw a single-line (default)
;         AL = 2 to draw a double-line
;         CH = screen row for line start point (0 to 24)
;         CL = screen column for line top start point (0 to 79)
; Note: the routine assumes that there is either a single or double
;       valid box symbol at the line's start coordinate. Also that
;       there is at least one more such symbol before the end
;       of the screen line
;
; Symbol substitution at left of box
;        F O U N D                  R E P L A C E D   B Y
;       code        graphics        code         graphics  if AL =
;       B3H          |              C3H            |-         1
;                                   C6H            |=         2
;       BAH          ||             C7H            ||-        1
;                                   CCH            ||=        2
;       B4H          -|             C5H           -|-       1 or 2
;       B5H          =|             D8H           =|=       1 or 2
;       B6H          -||            D7H          -||-       1 or 2
;       B9H          =||            CEH          =||=       1 or 2
;
; Symbol substitution at right of box
;        F O U N D                  R E P L A C E D   B Y
;       code        graphics        code         graphics  if AL =
;       B3H          |              B4H            -|         1
;                                   B5H            =|         2
;       BAH          ||             B6H            -||        1
;                                   B9H            =||        2
;       C3H          |-             C5H           -|-       1 or 2
;       C6H          |=             D8H           =|=       1 or 2
;       C7H          ||-            D7H          -||-       1 or 2
;       CCH          ||=            CEH          =||=       1 or 2
; On exit:
;         Carry clear (no error), line drawn
;         Carry set (error) invalid or illegal input
; Note:
;         The procedure uses data variables in CS
;
;**********************|
;   test for single or |
;     double line      |
;**********************|
        MOV     AH,0C4H         ; Assume single line
        MOV     BX,CS:LEFT1_CODES       ; Replacement codes
        MOV     DX,CS:RIGHT1_CODES
        CMP     AL,2            ; Test for double-line entry code
        JNE     SINGLE_OK       ; Go if not double line
        MOV     AH,0CDH         ; Replace with double-line code
        MOV     BX,CS:LEFT2_CODES       ; Replacement codes
        MOV     DX,CS:RIGHT2_CODES
SINGLE_OK:
        MOV     CS:LINE_CODE,AH ; Store in variable
; Install replacement codes in tables
        MOV     WORD PTR CS:REPLACE_LEFT,BX
        MOV     WORD PTR CS:REPLACE_RIGHT,DX
;**********************|
;   save parameters    |
;**********************|
        MOV     CS:LEFT_CORNER,CX       ; Store entry parameters
        CALL    ES_TO_VIDEO     ; Local procedure
;**********************|
; search for left-hand |
;      box codes       |
;**********************|
; The procedure searches for a code in the table named FOUND_LEFT
; The code is replace with the corresponding one in the table
; named REPLACE_LEFT. If no matching code is found, the carry flag
; is set and no line is drawn
        MOV     DX,CS:LEFT_CORNER       ; Restore left coordinates
        CALL    GET_OFFSET      ; DI --> video buffer offset
        LEA     SI,CS:FOUND_LEFT
        MOV     CX,6            ; Counter for 6 valid codes
        MOV     AL,ES:[DI]      ; Box character on screen
        MOV     DX,0            ; DL is offset counter
READ_SCREEN_6:
        MOV     AH,CS:[SI]      ; Table code
        CMP     AL,AH           ; Compare screen code with table
        JE      CODE_MATCHED    ; Go if code found
        INC     SI              ; Bump code table pointer
        INC     DX              ; Bump offset counter
        LOOP    READ_SCREEN_6   ; Continue testing
;**********************|
;   ERROR - no match   |
;**********************|
NO_MATCH:
        STC                     ; Carry is error flag
        RET
;**********************|
;    match found       |
;**********************|
; At this point a screen character matches the table of valid
; left-hand box characters. CL holds the character offset from
; the start of the table
CODE_MATCHED:
        INT     03H
        LEA     SI,CS:REPLACE_LEFT      ; Table of replacements
        ADD     SI,DX           ; Add offset of found character
        MOV     DX,CS:LEFT_CORNER       ; Restore left coordinates
        CALL    GET_OFFSET      ; DI --> video buffer offset
        MOV     AL,CS:[SI]      ; Get replacement code from table
        MOV     ES:[DI],AL      ; Display replacement
;**********************|
;   draw box line      |
;**********************|
        ADD     DI,2            ; Bump screen position
; Count characters to end of line. 78 is maximum search area
        MOV     CX,78           ; Test 78 screen positions
        MOV     AH,CS:LINE_CODE ; Graphics line character
TEST_78:
        MOV     AL,ES:[DI]      ; Screen character to AL
        CMP     AL,0B3H         ; First box character
        JAE     END_OF_BOX      ; Go if a box graphics character
        MOV     ES:[DI],AH      ; Display one line character
        ADD     DI,2            ; Bump screen pointer
        LOOP    TEST_78
        JMP     NO_MATCH        ; ERROR, no box
; AL holds graphics character that marks the end of the box
; ES:DI => screen box character to be replaced
END_OF_BOX:
        LEA     SI,CS:FOUND_RIGHT       ; Table for right-hand codes
        MOV     CX,6            ; Counter for 6 valid codes
        MOV     DX,0            ; DL is offset counter
TEST_RIGHT_6:
        MOV     AH,CS:[SI]      ; Table code
        CMP     AL,AH           ; Compare screen code with table
        JE      RIGHT_MATCH     ; Go if code found
        INC     SI              ; Bump code table pointer
        INC     DX              ; Bump offset counter
        LOOP    TEST_RIGHT_6    ; Continue testing
        JMP     NO_MATCH        ; ERROR, invalid end character
RIGHT_MATCH:
        LEA     SI,CS:REPLACE_RIGHT     ; Table of replacements
        ADD     SI,DX           ; Add offset of found character
        MOV     AL,CS:[SI]      ; Get replacement code from table
        MOV     ES:[DI],AL      ; Display replacement
        RET
;
BOX_LINE_HOR    ENDP
;******************************************************************
;
; CS DATA FOR THE BOX_LINE_VER PROCEDURE
; Left-replacement table. Default values for double line case. See
; replacement tables in the routine header comment
FOUND_TOP       DB      0C4H,0CDH,0C1H,0C5H,0CFH,0D0H
REPLACE_TOP     DB      0D2H,0CBH,0C5H,0CEH,0D8H,0D7H
TOP1_CODES      DW      0D1C2H
TOP2_CODES      DW      0CBD2H
;
FOUND_BOT       DB      0C4H,0CDH,0C2H,0CAH,0CFH,0D2H
REPLACE_BOT     DB      0D0H,0CAH,0C5H,0CEH,0D8H,0D7H
BOT1_CODES      DW      0CFC1H
BOT2_CODES      DW      0CAD0H
;
LINE_CODE_VER   DB      0BAH    ; Default is double line
                                ; Single line code is B3H
;
BOX_LINE_VER    PROC    FAR
; Draw a vertical line inside an existing screen box
; On entry:
;         AL = 1 to draw a single-line (default)
;         AL = 2 to draw a double-line
;         CH = screen row for line start point (0 to 24)
;         CL = screen column for line top start point (0 to 79)
; Note: the routine assumes that there is either a single or double
;       valid box symbol at the line's start coordinate. Also that
;       there is at least one more such symbol before the end
;       of the screen line
;
; Symbol substitution at top of box
;        F O U N D                  R E P L A C E D   B Y
;       code                        code                   if AL =
;       C4H                         C2H                       1
;                                   D2H                       2
;       CDH                         D1H                       1
;                                   CBH                       2
;       C1H                         C5H                     1 or 2
;       B5H                         CEH                     1 or 2
;       CFH                         D8H                     1 or 2
;       D0H                         D7H                     1 or 2
;
; Symbol substitution at bottom of box
;        F O U N D                  R E P L A C E D   B Y
;       code                        code                   if AL =
;       C4H                         C1H                       1
;                                   D0H                       2
;       CDH                         CFH                       1
;                                   CAH                       2
;       C2H                         C5H                     1 or 2
;       CAH                         CEH                     1 or 2
;       CFH                         D8H                     1 or 2
;       D2H                         D7H                     1 or 2
;
; On exit:
;         Carry clear (no error), line drawn
;         Carry set (error) invalid or illegal input
; Note:
;         The procedure uses data variables in CS
;
;**********************|
;   test for single or |
;     double line      |
;**********************|
        MOV     AH,0B3H         ; Assume single line
        MOV     BX,CS:TOP1_CODES        ; Replacement codes
        MOV     DX,CS:BOT1_CODES
        CMP     AL,2            ; Test for double-line entry code
        JNE     SINGLE_OKV      ; Go if not double line
        MOV     AH,0BAH         ; Replace with double-line code
        MOV     BX,CS:TOP2_CODES        ; Replacement codes
        MOV     DX,CS:BOT2_CODES
SINGLE_OKV:
        MOV     CS:LINE_CODE_VER,AH ; Store in variable
; Install replacement codes in tables
        MOV     WORD PTR CS:REPLACE_TOP,BX
        MOV     WORD PTR CS:REPLACE_BOT,DX
;**********************|
;   save parameters    |
;**********************|
        MOV     CS:LEFT_CORNER,CX       ; Store entry parameters
        CALL    ES_TO_VIDEO     ; Local procedure
;**********************|
;    search for top    |
;      box codes       |
;**********************|
; The procedure searches for a code in the table named FOUND_TOP
; The code is replace with the corresponding one in the table
; named REPLACE_TOP. If no matching code is found, the carry flag
; is set and no line is drawn
        MOV     DX,CS:LEFT_CORNER       ; Restore left coordinates
        CALL    GET_OFFSET      ; DI --> video buffer offset
        LEA     SI,CS:FOUND_TOP
        MOV     CX,6            ; Counter for 6 valid codes
        MOV     AL,ES:[DI]      ; Box character on screen
        MOV     DX,0            ; DL is offset counter
READ_SCREEN_6V:
        MOV     AH,CS:[SI]      ; Table code
        CMP     AL,AH           ; Compare screen code with table
        JE      CODE_MATCHED_V  ; Go if code found
        INC     SI              ; Bump code table pointer
        INC     DX              ; Bump offset counter
        LOOP    READ_SCREEN_6V  ; Continue testing
;**********************|
;   ERROR - no match   |
;**********************|
NO_MATCH_V:
        STC                     ; Carry is error flag
        RET
;**********************|
;    match found       |
;**********************|
; At this point a screen character matches the table of valid
; left-hand box characters. CL holds the character offset from
; the start of the table
CODE_MATCHED_V:
        LEA     SI,CS:REPLACE_TOP       ; Table of replacements
        ADD     SI,DX           ; Add offset of found character
        MOV     DX,CS:LEFT_CORNER       ; Restore left coordinates
        CALL    GET_OFFSET      ; DI --> video buffer offset
        MOV     AL,CS:[SI]      ; Get replacement code from table
        MOV     ES:[DI],AL      ; Display replacement
;**********************|
;   draw box line      |
;**********************|
        ADD     DI,160          ; Bump screen position
; Count characters to end of line. 23 is maximum search area
        MOV     CX,23           ; Test 23 screen positions
        MOV     AH,CS:LINE_CODE_VER ; Graphics line character
TEST_23:
        MOV     AL,ES:[DI]      ; Screen character to AL
        CMP     AL,0B3H         ; First box character
        JAE     END_OF_BOX_V    ; Go if a box graphics character
        MOV     ES:[DI],AH      ; Display one line character
        ADD     DI,160          ; Bump screen pointer
        LOOP    TEST_23
        JMP     NO_MATCH_V      ; ERROR, no box
; AL holds graphics character that marks the bottom of the box
; ES:DI => screen box character to be replaced
END_OF_BOX_V:
        LEA     SI,CS:FOUND_BOT         ; Table for bottom codes
        MOV     CX,6            ; Counter for 6 valid codes
        MOV     DX,0            ; DX is offset counter
TEST_BOT_6:
        MOV     AH,CS:[SI]      ; Table code
        CMP     AL,AH           ; Compare screen code with table
        JE      BOT_MATCH       ; Go if code found
        INC     SI              ; Bump code table pointer
        INC     DX              ; Bump offset counter
        LOOP    TEST_BOT_6      ; Continue testing
        JMP     NO_MATCH_V      ; ERROR, invalid end character
BOT_MATCH:
        LEA     SI,CS:REPLACE_BOT       ; Table of replacements
        ADD     SI,DX           ; Add offset of found character
        MOV     AL,CS:[SI]      ; Get replacement code from table
        MOV     ES:[DI],AL      ; Display replacement
        RET
;
BOX_LINE_VER    ENDP
;
P_CODE  ENDS
        END
