; Module name: OUTPUT.ASM                     Library:   SOLUTION.LIB
; Created:
; Update:
;
;********************************************************************
;********************************************************************
;                             OUTPUT.ASM
;********************************************************************
;********************************************************************
;
; Library module for printer and serial line operations
;
;*******************|
;      PUBLICS      |
;*******************|
PUBLIC	SEND_PRN
PUBLIC	STATUS_PRN
PUBLIC	SET_PROTOCOL
PUBLIC	STATUS_RS232
PUBLIC	SEND_RS232
PUBLIC	GET_RS232
;
;*******************************************************************
;                              code
;*******************************************************************
;
P_CODE	SEGMENT PUBLIC
	ASSUME	CS:P_CODE
;
;********************************************************************
;                           printer routines 
;********************************************************************
SEND_PRN	PROC	FAR
; Test printer status and send character when printer ready
; On entry:
;         AL = character to send
;         DX = printer port (valid range is 0, 1, and 2)
; On exit:
;         carry clear if no error
;         carry set if print operation failed
;                               
;******************|
; initialize port  |
;******************|
INIT_PORT:
	MOV	AH,1		; Service request number
	INT	17H		; Transfer to BIOS
	TEST	AH,80H		; Test printer not busy bit
	JNZ	NOT_BUSY
	TEST	AH,1		; Time out bit
	JNZ	STATUS_ERROR
	JMP	INIT_PORT	; Repeat initialization and status
				; check
;*******************|
;    error exit     |
;*******************|
STATUS_ERROR:
	STC			; Carry set to signal error
	RET
;******************|
;  send character  |
;******************|
NOT_BUSY:
	MOV	AH,0		; Function request code
	INT	17H		; Transfer to BIOS
	CLC			; No error in send operation
	RET
SEND_PRN	ENDP
;*******************************************************************
;
STATUS_PRN	PROC	FAR
; Obtain printer status using BIOS service number of the specified
; port
; On entry:
;         DX = printer number (valid range is 0, 1, and 2)
; On exit:
;           AH = 7  6  5  4  3  2  1  0  printer status bit map
;                |  |  |  |  |  |  |  |
;                |  |  |  |  |  |  |  |__ time out
;                |  |  |  |  |  |__|_____ UNUSED
;                |  |  |  |  |___________ input/output error
;                |  |  |  |______________ printer selected
;                |  |  |_________________ out of paper
;                |  |____________________ acknowledge
;                |_______________________ printer not busy
;           carry clear if valid status
;           carry set if invalid port request
; Test for invalid value in DX
	CMP	DX,2		; 2 is maximum legal input
	JLE	OK_PORT_VAL	; Input is valid
	STC			; Set carry to signal error  
	RET
OK_PORT_VAL:
	MOV	DX,0		; Port 0
	MOV	AH,1		; Service request number
	INT	17H		; Transfer to BIOS
	CLC			; Carry clear if no error
	RET
STATUS_PRN	ENDP
;
;********************************************************************
;                       serial port routines
;********************************************************************
;
SET_PROTOCOL	PROC	FAR
; Set communications protocol for the serial line
; On entry:
; BAUD RATE:
;          AL = 1 for 110 baud          AL = 5 for 1200 baud
;          AL = 2 for 150 baud          AL = 6 for 2400 baud 
;          AL = 3 for 300 baud          AL = 7 for 4800 baud
;          AL = 4 for 600 baud          AL = 8 for 9600 baud
;---------------------------------------------------------------------
; PARITY:
;          BL = 1 for parity ODD        BL = 2 for no parity
;          BL = 3 for parity EVEN
;---------------------------------------------------------------------
; STOP BITS:
;          BH = 1 for 1 stop bit        BH = 2 for 2 stop bits
;---------------------------------------------------------------------
; WORD LENGTH:
;          DL = 1 for 7 bit words       DL = 2 for 8 bit words
;---------------------------------------------------------------------
; On exit:
;         carry clear if no error
;         carry set if invalid input values
;        
	PUSH	AX		; Save caller's registers
	PUSH	BX
	PUSH	CX
	PUSH	DX
;******************|
;    baud rate     |
;******************|
; AL holds baud rate code. Valid range is 1 to 8
	CMP	AL,1		; Carry set if less than 1
	JL	BAD_VALUES	; Error exit if less than 1
	CMP	AL,9		; or greater than 8
	JLE	OK_BAUD	
;******************|
;    error exit    |
;******************|
BAD_VALUES:
	POP	DX		; Restore caller's register
	POP	CX
	POP	BX
	POP	AX
	STC			; Illegal baud rate code
	RET
; Input is in valid range.
OK_BAUD:
	SUB	AL,1		; To range 0 to 7
	MOV	CL,5		; Shift bit counter
	SHL	AL,CL		; Baud rate bits left 5 bits
;*******************|
;      parity       |
;*******************|
; BL holds parity code. Valid range is 1, 2 or 3
	CMP	BL,1		; Carry set if less than 1
	JL	BAD_PARITY	; Error exit if less than 1
	CMP	BL,3		; or greater than 8
	JLE	OK_PARITY
BAD_PARITY:
	JMP	BAD_VALUES
OK_PARITY:
	MOV	CL,3		; Shift bit counter
	SHL	BL,CL		; Shift left 3 bits
	OR	AL,BL		; OR with baud rate
;*******************|
;    stop bits      |
;*******************|
; BH holds stop bits code. Valid values are 1 and 2
	CMP	BH,1		; Carry set if less than 1
	JL	BAD_STOP	; Error exit if less than 1
	CMP	BH,2		; or greater than 2
	JLE	OK_STOP
BAD_STOP:
	JMP	BAD_VALUES
OK_STOP:
	MOV	CL,2
	SHL	BH,CL		; Shift digits
	OR	AL,BH		; OR with stored parameters
;*******************|
;    word length    |
;*******************|
; DL holds word length code. Valid values are 1 and 2
	CMP	DL,1		; Carry set if less than 1
	JL	BAD_WORD	; Error exit if less than 1
	CMP	BH,2		; or greater than 2
	JLE	OK_WORD
BAD_WORD:
	JMP	BAD_VALUES
OK_WORD:
	INC	DL		; Bring to range 2 or 3
	OR	AL,DL		; OR with stored parameters
;*******************|
;    install new    |
;     protocol      |
;*******************|
; AL holds bit code for BIOS service number 0, INT 14H
	MOV	DX,0		; Always COMM1 card
	MOV	AH,0		; Request code
	INT	14H		; BIOS service
	CLC			; No error
	RET
SET_PROTOCOL	ENDP
;********************************************************************

STATUS_RS232	PROC	FAR
; Obtain RS-232C port status using BIOS service number of the specified
; port
; On entry:
;         DX = RS-232C port number (valid range is 0 to 3)
;              0 = COMM1, 1 = COMM2, etc.
; On exit:
;           AH = 7  6  5  4  3  2  1  0  RS-232C line status bit map
;                |  |  |  |  |  |  |  |
;                |  |  |  |  |  |  |  |__ data ready
;                |  |  |  |  |  |  |_____ overrun error
;                |  |  |  |  |  |________ parity error
;                |  |  |  |  |___________ framing error
;                |  |  |  |______________ break detect
;                |  |  |_________________ transmission hold register
;                |  |                     empty
;                |  |____________________ transmission shift register
;                |                        empty
;                |_______________________ time-out
;
;           AL = 7  6  5  4  3  2  1  0  modem status bit map
;                |  |  |  |  |  |  |  |
;                |  |  |  |  |  |  |  |__ delta clear to send
;                |  |  |  |  |  |  |_____ delta data set ready
;                |  |  |  |  |  |________ trailing edge indicator
;                |  |  |  |  |___________ receive line signal detect
;                |  |  |  |______________ clear to send
;                |  |  |_________________ data set ready
;                |  |____________________ ring indicator
;                |_______________________ received line signal detect
;
;           carry clear if valid status
;           carry set if invalid port request
;
; Test for invalid value in DX
	CMP	DX,3		; 3 is maximum legal input
	JLE	OK_RSPORT	; Input is valid
	STC			; Set carry to signal error  
	RET
OK_RSPORT:
	MOV	AH,3		; Service request number
	INT	14H		; Transfer to BIOS
	CLC			; Carry clear if no error
	RET
STATUS_RS232	ENDP
;********************************************************************
;
SEND_RS232	PROC	FAR
; Send character through RS-232C port
; On entry:
;         AL = character to send
;         DX = RS-232C port number (valid range is 0 to 3)
;              0 = COMM1, 1 = COMM2, etc.
; On exit:
;         AH = line status bit map (as in STATUS_RS232)
;              time-out bit set if character not sent
;         carry clear
;
	MOV	AH,1		; Service request number
	INT	14H		; Transfer to BIOS
	CLC			; Carry clear if no error
	RET
SEND_RS232	ENDP
;********************************************************************
;
GET_RS232	PROC	FAR
; Wait for a character and receive it through the RS-232C port
; On entry:
;         DX = RS-232C port number (valid range is 0 to 3)
;              0 = COMM1, 1 = COMM2, etc.
; On exit:
;         AL = character received
;         AH = line status bit map (as in STATUS_RS232)
;              time-out bit set if character not sent
;         carry clear
;
; Note: The routine waits until character is received
	MOV	AH,2		; Service request number
	INT	14H		; Transfer to BIOS
	CLC			; Carry clear if no error
	RET
GET_RS232	ENDP
;
P_CODE	ENDS
	END
ce request number
	INT	14H		; Transfer to BIOS
	CLC			; Carry clear if no error
	RET
GET_RS232	E