; Module name: SVGA.ASM                       Library:   GRAPHSOL.LIB
; Created: July 15, 1992
;********************************************************************
;********************************************************************
;                              SVGA.ASM
;********************************************************************
;********************************************************************
; Description:
; Library module for SVGA video operations using VESA BIOS services
; and direct access to video memory
;    1. Initialization and setup
;       a. Initialize SVGA and report system configuration
;       b. Set SVGA mode
;    2. Display operations while in 1024 by 768 resolution,
;       256 colors mode (VESA mode number 105H)
;       a. Display pixel 
;       b. Clear screen
;       c. read pixel
;
; Note: The library routines that are contained in this module can
;       be used from the high level languages C, Pascal, and BASIC
;       as well as in programs coded directly in Assembly Language.
;       Appendix B explains interfacing the library routines with
;       these languages.
;*******************************************************************
;                          ROUTINES
;*******************************************************************
; Public procedures in this module:
; NAME             DESCRIPTION
; VESA_MODE        Obtain VESA BIOS and mode information
; VESA_105         Set VESA SuperVGA mode 105H
; SVGA_PIX_105     Set pixel in VESA mode 105H
; SVGA_CLS_105     Clear screen in VESA mode 105H
; SVGA_READ_105    Read screen pixel in VESA mode 105H
;
;*******************|
;      PUBLICS      |
;*******************|
PUBLIC  VESA_MODE
PUBLIC  VESA_105
PUBLIC  SVGA_PIX_105
PUBLIC  SVGA_CLS_105
PUBLIC  SVGA_READ_105
;
;*****************************************************************
;*****************************************************************
;                      segment for SVGA data
;*****************************************************************
;*****************************************************************
SVGA_DATA        SEGMENT
;
VESA_BUFFER     DB      '    '          ; VESA signature
VESA_VERSION    DW      ?       ; Version number
OEM_PTR_OFF     DW      ?       ; OEM string offset pointer
OEM_PTR_SEG     DW      ?       ; OEM string segment pointer
CAPABILITES     DB      4 DUP (00H)  ; Reserved field
MODES_PTR_OFF   DW      ?       ; Pointer to modes list, offset
MODES_PTR_SEG   DW      ?       ; Segment for idem
MEM_BLOCKS      DW      ?       ; Count of 64K memory blocks
                                ; (Only in June 2, 1990 revision)
                DB      242 DUP (0H)
;
VESA_INFO       DW      ?       ; Mode attributes, mapped at follows:
                        ; ..4 3 2 1 0 <= bits
                        ;   | | | | |__ 0 = mode not supported
                        ;   | | | |     1 = mode supported
                        ;   | | | |____ 0 = no extended mode info
                        ;   | | |       1 = extended mode info
                        ;   | | |_______0 = no output functions
                        ;   | |         1 = output functions
                        ;   | |________ 0 = monochrome mode
                        ;   |           1 = color mode
                        ;   |__________ 0 = text mode
                        ;               1 = graphics mode
                        ; 15..5 = RESERVED
WIN_A_ATTS      DB      ?       ; Window A attributes
WIN_B_ATTS      DB      ?       ; Window B attributes
WIN_GRAIN       DW      ?       ; Window granularity
WIN_SIZE        DW      ?       ; Window size
WIN_A_SEG       DW      ?       ; Segment address for window A
WIN_B_SEG       DW      ?       ; Segment address for window B
SWITCH_BANK     DD      ?       ; Far pointer to window function
BYTES_PER_ROW   DW      ?       ; Bytes per screen row
; Extended mode optional data
X_RES           DW      ?       ; Horizontal resolution
Y_RES           DW      ?       ; Vertical resolution
X_CHAR_SIZE     DB      ?       ; Pixel width of character cell
Y_CHAR_SIZE     DB      ?       ; Pixel height of character cell
BIT_PLANES      DB      ?       ; Number of bit planes
BITS_PER_PIX    DB      ?       ; Bits per pixel in this mode
NUM_OF_BANKS    DB      ?       ; Number of video memory banks
MEM_MODEL       DB      ?       ; Memory model
BANK_SIZE       DB      ?       ; Kilobytes per bank
                DW      0       ; Padding
;
THIS_BANK       DB      0       ; Current video bank number
;
;

SVGA_DATA        ENDS
;
;*****************************************************************
;*****************************************************************
;                          procedures
;*****************************************************************
;*****************************************************************
P_CODE  SEGMENT PUBLIC
        ASSUME  CS:P_CODE
;
;
VESA_MODE       PROC    FAR
; Call VESA BIOS sub-service number 0 to obtain SuperVGA
; information and sub-service number 1 to obtain mode information
; On entry:
;          CX = mode number, as follows:
;               number       resolution      colors
;                100H        640 by 400       256
;                101H        640 by 480       256
;                102H        800 by 600        16
;                103H        800 by 600       256
;                104H       1024 by 768        16
;                105H       1024 by 768       256
;                106H       1280 by 1224       16
;                107H       1280 by 1224      256
; On exit:
;         Carry clear if no error
;         Data stored in the buffers VESA_BUFFER and VESA_INFO
;         ES:SI -> VESA_BUFFER
;         ES:DI -> VESA_INFO
;
;**********************|
;   save caller's DS   |
; set DS to SVGA_DATA  |
;**********************|
        PUSH    DS
        MOV     AX,SVGA_DATA            ; Local data segment
        MOV     DS,AX                   ; to DS
        ASSUME  DS:SVGA_DATA
;**********************|
;   setup registers    |
;**********************|
; Set pointers to data storage buffers
        LEA     DI,VESA_BUFFER  ; Pointer to data buffer
        LEA     SI,VESA_INFO
; VESA BIOS sub-service number 0 uses ES as a segment base
        PUSH    DS              ; Local data segment
        POP     ES              ; To ES
;**********************|
; get VESA information |
;**********************|
        MOV     AH,79           ; VESA BIOS service number
        MOV     AL,0            ; This sub-service
        INT     10H             ; BIOS video service
; At this point AX must hold 004FH if the call executed
        CMP     AX,004FH        ; Returned code
        JNE     BAD_VESA        ; Go if invalid value
; Test buffer for a valid 'VESA' signature
        CMP     WORD PTR [DI],'EV'      ; First two letters
        JE      OK_VE           ; Go if matched
        JMP     BAD_VESA        ; Exit if not matched
OK_VE:
        CMP     WORD PTR [DI+2],'AS'    ; Last two letters
        JE      OK_VESA         ; Go if signature matched
        JMP     BAD_VESA        ; Go if not matched
;**********************|
;  get VESA mode info  |
;**********************|
OK_VESA:
; At this point there is a valid call and signature
; CX holds requested mode number
; VESA BIOS sub-service number 1 is used to obtain mode information
        XCHG    SI,DI           ; Set DI as pointer
        MOV     AH,79           ; VESA BIOS service number
        MOV     AL,1            ; This sub-service
        INT     10H             ; BIOS video service
;
        POP     DS              ; Restore caller's DS
        XCHG    SI,DI           ; Pointers to original registers
        CLC
        RET
BAD_VESA:
; Error exit
        POP     DS              ; Restore caller's DS
        STC                     ; Error flag
        RET
VESA_MODE       ENDP
;****************************************************************
;
VESA_105        PROC    FAR
; Procedure to set SuperVGA mode number 105H with a resolution
; of 1024 by 768 pixels in 256 colors
; This procedure assumes that the data variables in the buffers
; VESA_BUFFER and VESA_INFO have been filled by a previous call
; to the VESA_MODE procedure with CX = 105H
;
;**********************|
;   save caller's DS   |
; set DS to SVGA_DATA  |
;**********************|
        PUSH    DS
        MOV     AX,SVGA_DATA            ; Local data segment
        MOV     DS,AX                   ; to DS
        ASSUME  DS:SVGA_DATA
;**********************|
; test standard values |
;**********************|
        MOV     AX,WIN_SIZE     ; Standard size is 64K (40H)
        CMP     AX,040H         ; Is it 64K
        JE      OK_WIN_SIZE     ; Go if standard size
; Error exit. Non-standard parameters
;**********************|
;     ERROR exit       |
;**********************|
NON_STANDARD:
        POP     DS              ; Restore caller's DS
        STC                     ; Carry is error flag
        RET
; Now test for a start address of video buffer at A000H
OK_WIN_SIZE:
        MOV     AX,WIN_A_SEG    ; Segment for window A
        CMP     AX,0A000H       ; Test for standard address
        JE      SET_MODE_105    ; Go if standard
        JMP     NON_STANDARD
;**********************|
;  select video mode   |
;**********************|
; Mode variables are standard. Select mode 105H using VESA BIOS
; sub-service number 2
SET_MODE_105:
        MOV     BX,0105H        ; Mode number and high bit = 0
                                ; to request clear video
        MOV     AH,79           ; VESA BIOS service number
        MOV     AL,2            ; This sub-service
        INT     10H             ; BIOS video service
; Test for valid returned value
        CMP     AX,004FH        ; Status for no error
        JE      OK_MODE_105     ; No error during mode set
        STC                     ; Error flag. Mode not set
        POP     DS              ; Restore caller's DS
        RET
OK_MODE_105:
        POP     DS              ; Restore caller's DS
        CLC                     ; No error flag
        RET
VESA_105        ENDP
;******************************************************************
;
SVGA_PIX_105   PROC  FAR
; Write a screen pixel accessing SVGA memory directly and using
; VESA BIOS service to select bank
; On entry:
;       CX = x coordinate of pixel
;       DX = y coordinate of pixel
;       BL = pixel color in 8-bit format
; Code assumes:
;       1. SVGA is in a 1024 by 768 pixel mode in 256 colors
;           (mode number 105H)
;       2. video bank size is 64K
;       3. ES holds base address of video buffer (A000H)
;
;**********************|
;   save caller's DS   |
; set DS to SVGA_DATA  |
;**********************|
; This manipulation is required in order access the address of
; the bank switching routine
        PUSH    DS
        MOV     AX,SVGA_DATA            ; Local data segment
        MOV     DS,AX                   ; to DS
        ASSUME  DS:SVGA_DATA
;
        PUSH    BX              ; Save entry registers
        PUSH    CX
        PUSH    DX
        MOV     AL,BL           ; Color to AL
; Get address in SVGA memory space
        CLC                     ; Clear carry flag
        PUSH    AX              ; Save color value
        MOV     AX,1024         ; Pixels per scan line
        MUL     DX              ; DX holds line count of address
        ADD     AX,CX           ; Add pixels in current line
        ADC     DX,0            ; Answer in DX:AX
                                ; DL = bank, AX = offset
        MOV     BX,AX           ; Offset to BX
;**********************|
;    change banks      |
;**********************|
; Bank switching is performed by means of a far call to the
; service routine located at the label SWITCH_BANK
        PUSH    BX              ; Save entry registers
        PUSH    DX
; VESA BIOS sub-service number 5 register setup:
; BH = 0 to select bank
; BL = 0 to select window A
; DX = bank number
        MOV     BX,0            ; Select bank in window A
        CALL    DWORD PTR SWITCH_BANK
        POP     DX              ; Restore entry registers
        POP     BX
        POP     AX              ; Restore color
;**********************|
;   set the pixel      |
;**********************|
        MOV     ES:[BX],AL      ; Write the dot
        POP     DX              ; Restore entry values
        POP     CX
        POP     BX
        POP     DS              ; Restore caller's DS
        RET
SVGA_PIX_105    ENDP
;
;******************************************************************
SVGA_CLS_105    PROC    FAR
; Clear video memory while in SVGA mode number 105H
; On entry:
;         AL = 8-bit color code to use for initialization
; Code assumes:
;       1. SVGA is in a 1024 by 768 pixel mode in 256 colors
;           (mode number 105H)
;       2. Video bank size is 64K
;       3. ES holds base address of video buffer (A000H)
;
;**********************|
;   save caller's DS   |
; set DS to SVGA_DATA  |
;**********************|
; This manipulation is required in order access the address of
; the bank switching routine
        PUSH    DS
        MOV     DX,SVGA_DATA            ; Local data segment
        MOV     DS,DX                   ; to DS
        ASSUME  DS:SVGA_DATA
; BL holds bank number
        MOV     DX,0            ; Initialize to first bank
;**********************|
;    change banks      |
;**********************|
; Bank switching is performed by means of a far call to the
; service routine located at the label SWITCH_BANK
NEXT_BANK:
        PUSH    AX              ; Save color code
        PUSH    BX              ; Save entry registers
        PUSH    DX
; VESA BIOS sub-service number 5 register setup:
; BH = 0 to select bank
; BL = 0 to select window A
; DX = bank number
        MOV     BX,0            ; Select bank in window A
        CALL    DWORD PTR SWITCH_BANK
        POP     DX              ; Restore entry registers
        POP     BX
        POP     AX              ; Restore color code
; Write 65536 bytes of 00H in current bank
        MOV     CX,0FFFFH       ; CX is byte counter
        CLD                     ; Forward direction
        MOV     DI,0            ; Start of block
        REP     STOSB           ; Store 65536 bytes
; Bump bank
        INC     DX              ; Next bank
        CMP     DX,12           ; 12 is last bank
        JNE     NEXT_BANK
        POP     DS              ; Restore caller's DS
        RET
SVGA_CLS_105    ENDP
;*****************************************************************
;
SVGA_READ_105   PROC    FAR
; Read a screen pixel accessing SVGA memory directly and using
; VESA BIOS service to select bank
; On entry:
;       CX = x coordinate of pixel
;       DX = y coordinate of pixel
; Code assumes:
;       1. SVGA is in a 1024 by 768 pixel mode in 256 colors
;           (mode number 105H)
;       2. video bank size is 64K
;       3. ES holds base address of video buffer (A000H)
; On exit:
;       AL = 8-bit pixel color code
;
;**********************|
;   save caller's DS   |
; set DS to SVGA_DATA  |
;**********************|
; This manipulation is required in order access the address of
; the bank switching routine
        PUSH    DS
        MOV     AX,SVGA_DATA            ; Local data segment
        MOV     DS,AX                   ; to DS
        ASSUME  DS:SVGA_DATA
;
        PUSH    BX              ; Save entry registers
        PUSH    CX
        PUSH    DX
; Get address in SVGA memory space
        CLC                     ; Clear carry flag
        MOV     AX,1024         ; Pixels per scan line
        MUL     DX              ; DX holds line count of address
        ADD     AX,CX           ; Add pixels in current line
        ADC     DX,0            ; Answer in DX:AX
                                ; DL = bank, AX = offset
        MOV     BX,AX           ; Offset to BX
;**********************|
;    change banks      |
;**********************|
; Bank switching is performed by means of a far call to the
; service routine located at the label SWITCH_BANK
        PUSH    BX              ; Save entry registers
        PUSH    DX
; VESA BIOS sub-service number 5 register setup:
; BH = 0 to select bank
; BL = 0 to select window A
; DX = bank number
        MOV     BX,0            ; Select bank in window A
        CALL    DWORD PTR SWITCH_BANK
        POP     DX              ; Restore entry registers
        POP     BX
;**********************|
;   read the pixel     |
;**********************|
        MOV     AL,BYTE PTR ES:[BX]    ; Read memory
        POP     DX              ; Restore entry values
        POP     CX
        POP     BX
        POP     DS              ; Restore caller's DS
        RET
SVGA_READ_105   ENDP
;
P_CODE  ENDS
        END
