; Module name: VGA1.ASM                       Library:   GRAPHSOL.LIB
; Created: March, 1992
;
;********************************************************************
;********************************************************************
;                           VGA1.ASM
;********************************************************************
;********************************************************************
; Description:
; Library module for the following functions:
;     1. VGA mode number 18 device drivers
;        a. calculate pixel address
;        b. calculate tile (8-by-8 pixels) address
;        c. write pixel
;        d. write tile
;        e. read pixel
;     2. VGA mode number 19 device drivers
;        a. write pixel
;        b. read pixel
;        c. calculate address (at pixel level)
;        d. Initialize DAC for IIRRGGBB color code
;        e. Initialize DAC for 64-step gray scale
;        f. Perform sum-to-gray function
;        g. Save installed DAC colors
;        h. Restore previous DAC colors
;        i. Disable DAC changes during mode set
;        j, Enable DAC changes during mode set
;     3. VGA setup, inquiry, and control primitives
;        a. obtain BIOS graphics mode
;        b. set BIOS graphics mode
;        c. select read mode
;        d. select write mode (16- and 256-color modes)
;        e. select logical operation mode (NORMAL/OR/AND/XOR)
;        f. VGA timing (wait for start of vertical retrace cycle)
;
; Note: The library routines that are contained in this module can
;       be used from the high level languages C, Pascal, and BASIC
;       as well as in programs coded directly in Assembly Language.
;       Appendix B explains interfacing the library routines with
;       these languages.
;*******************************************************************
;                          ROUTINES
;*******************************************************************
; Public procedures in this module:
; VGA device drivers
; NAME             DESCRIPTION
; Mode number 18 device drivers
; PIXEL_ADD_18     Calculate pixel-level buffer offset
; WRITE_PIX_18     Set one pixel in VGA mode 18
; TILE_ADD_18      Calculate coarse grain offset (8-by-8 pixels)
; WRITE_TILE_18    Set one tile (8-by-8 pixels) in mode 18
; READ_PIX_18      Read screen pixel in VGA mode 18
; READ_MAPS_18     Read IRGB maps in VGA mode number 18
;
; Mode number 19 device drivers
; TWO_BIT_IRGB     Init DAC registers to IIRRGGBB format
; GRAY_256         Init DAC registers to 64 gray shades
; SUM_TO_GRAY      Perform BIOS sum-to-gray-shades function
; SAVE_DAC         Save contents of DAC registers into buffer
; RESTORE_DAC      Restore DAC registers from buffer
; PIXEL_ADD_19     Mode number 19 address computation
; TILE_ADD_19      Mode number 19 coarse grain address
; FREEZE_DAC       Disable Palette and DAC changes by BIOS
; THAW_DAC         Enable Palette and DAC changes by BIOS
;
; VGA setup, inquiry, and control primitives
; NAME             DESCRIPTION
; SET_MODE         Set video mode
; GET_MODE         Read current video mode
; SET_WRITE_MODE   Set Graphics Controller write mode
; SET_WRITE_256    Set Graphics Controller write mode in 256 colors
; SET_READ_MODE    Set Graphics Controller read mode
; LOGICAL_MODE     Set Graphics Controller Data Rotate Register to
;                  XOR, OR, AND or NORMAL mode
; TIME_VRC         Wait for vertical retrace cycle of CRT controller
;
; Support procedures
; ES_TO_VIDEO      Set ES register to video buffer base
; ES_TO_APA        Set ES to graphics mode segment base (A000H)
;*******************************************************************
;                         W A R N I N G S !
;*******************************************************************
; 1. Some routines in this module use the ES register as a pointer
;    to the video buffer base. If the user's software requires ES
;    it must preserve it before calling the routine in this library
;*******************************************************************
;*******************|
;      PUBLICS      |
;*******************|
PUBLIC  ES_TO_VIDEO
PUBLIC  ES_TO_APA
PUBLIC  SET_MODE
PUBLIC  GET_MODE
PUBLIC  TIME_VRC
PUBLIC  SET_WRITE_MODE
PUBLIC  SET_WRITE_256
PUBLIC  SET_READ_MODE
PUBLIC  LOGICAL_MODE
PUBLIC  PIXEL_ADD_18
PUBLIC  TILE_ADD_18
PUBLIC  TILE_ADD_19
PUBLIC  WRITE_PIX_18
PUBLIC  WRITE_TILE_18
PUBLIC  READ_PIX_18
PUBLIC  TWO_BIT_IRGB
PUBLIC  GRAY_256
PUBLIC  SUM_TO_GRAY
PUBLIC  SAVE_DAC
PUBLIC  RESTORE_DAC
PUBLIC  PIXEL_ADD_19
PUBLIC  FREEZE_DAC
PUBLIC  THAW_DAC
;*****************************************************************
;                           externals
;*****************************************************************
EXTRN   SET_CURSOR:     FAR     ; In the ALFAGRAF module
;
;******************************************************************
;                       graphic data segment
;******************************************************************
P_DATA  SEGMENT
;                      | R   G   B | R   G   B | R   G   B |
INTENSITY_0     DB      009,009,009,009,009,018,009,009,027 ; 2
                DB      009,009,036,009,018,009,009,018,018 ; 5
                DB      009,018,027,009,018,036,009,027,009 ; 8
                DB      009,027,018,009,027,027,009,027,036 ; 11
                DB      009,036,009,009,036,018,009,036,027 ; 14
                DB      009,036,036,018,009,009,018,009,018 ; 17
                DB      018,009,027,018,009,036,018,018,009 ; 20
                DB      018,018,018,018,018,027,018,018,036 ; 23
                DB      018,027,009,018,027,018,018,027,027 ; 26
                DB      018,027,036,018,036,009,018,036,018 ; 29
                DB      018,036,027,018,036,036,027,009,009 ; 32
                DB      027,009,018,027,009,027,027,009,036 ; 35
                DB      027,018,009,027,018,018,027,018,027 ; 38
                DB      027,018,036,027,027,009,027,027,018 ; 41
                DB      027,027,027,027,027,036,009,036,009 ; 44
                DB      027,036,018,027,036,027,027,036,036 ; 47
                DB      036,009,009,036,009,018,036,009,027 ; 50
                DB      036,009,036,036,018,009,036,018,018 ; 53
                DB      036,018,027,036,018,036,036,027,009 ; 56
                DB      036,027,018,036,027,027,036,027,036 ; 59
                DB      036,036,009,036,036,018,036,036,027 ; 62
                DB      036,036,036;                        ; 63
;
INTENSITY_1     DB      192 DUP (00)
INTENSITY_2     DB      192 DUP (00)
INTENSITY_3     DB      192 DUP (00)
;
;                      | R   G   B | R   G   B | R   G   B |
GRAY_MAP        DB      000,000,000,001,001,001,002,002,002 ; 2
                DB      003,003,003,004,004,004,005,005,005 ; 5
                DB      006,006,006,007,007,007,008,008,008 ; 8 
                DB      009,009,009,010,010,010,011,011,011 ; 11
                DB      012,012,012,013,013,013,014,014,014 ; 14
                DB      015,015,015,016,016,016,017,017,017 ; 17
                DB      018,018,018,019,019,019,020,020,020 ; 20
                DB      021,021,021,022,022,022,023,023,023 ; 23
                DB      024,024,024,025,025,025,026,026,026 ; 26
                DB      027,027,027,028,028,028,029,029,029 ; 29
                DB      030,030,030,031,031,031,032,032,032 ; 32
                DB      033,033,033,034,034,034,035,035,035 ; 35
                DB      036,036,036,037,037,037,038,038,038 ; 38
                DB      039,039,039,040,040,040,041,041,041 ; 41
                DB      042,042,042,043,043,043,044,044,044 ; 44
                DB      045,045,045,046,046,046,047,047,047 ; 47
                DB      048,048,048,049,049,049,050,050,050 ; 50
                DB      051,051,051,052,052,052,053,053,053 ; 53
                DB      054,054,054,055,055,055,056,056,056 ; 56
                DB      057,057,057,058,058,058,059,059,059 ; 59
                DB      060,060,060,061,061,061,062,062,062 ; 62
                DB      063,063,063                         ; 63
                DB      000,000,000,001,001,001,002,002,002 ; 2
                DB      003,003,003,004,004,004,005,005,005 ; 5
                DB      006,006,006,007,007,007,008,008,008 ; 8 
                DB      009,009,009,010,010,010,011,011,011 ; 11
                DB      012,012,012,013,013,013,014,014,014 ; 14
                DB      015,015,015,016,016,016,017,017,017 ; 17
                DB      018,018,018,019,019,019,020,020,020 ; 20
                DB      021,021,021,022,022,022,023,023,023 ; 23
                DB      024,024,024,025,025,025,026,026,026 ; 26
                DB      027,027,027,028,028,028,029,029,029 ; 29
                DB      030,030,030,031,031,031,032,032,032 ; 32
                DB      033,033,033,034,034,034,035,035,035 ; 35
                DB      036,036,036,037,037,037,038,038,038 ; 38
                DB      039,039,039,040,040,040,041,041,041 ; 41
                DB      042,042,042,043,043,043,044,044,044 ; 44
                DB      045,045,045,046,046,046,047,047,047 ; 47
                DB      048,048,048,049,049,049,050,050,050 ; 50
                DB      051,051,051,052,052,052,053,053,053 ; 53
                DB      054,054,054,055,055,055,056,056,056 ; 56
                DB      057,057,057,058,058,058,059,059,059 ; 59
                DB      060,060,060,061,061,061,062,062,062 ; 62
                DB      063,063,063                         ; 63
                DB      000,000,000,001,001,001,002,002,002 ; 2
                DB      003,003,003,004,004,004,005,005,005 ; 5
                DB      006,006,006,007,007,007,008,008,008 ; 8 
                DB      009,009,009,010,010,010,011,011,011 ; 11
                DB      012,012,012,013,013,013,014,014,014 ; 14
                DB      015,015,015,016,016,016,017,017,017 ; 17
                DB      018,018,018,019,019,019,020,020,020 ; 20
                DB      021,021,021,022,022,022,023,023,023 ; 23
                DB      024,024,024,025,025,025,026,026,026 ; 26
                DB      027,027,027,028,028,028,029,029,029 ; 29
                DB      030,030,030,031,031,031,032,032,032 ; 32
                DB      033,033,033,034,034,034,035,035,035 ; 35
                DB      036,036,036,037,037,037,038,038,038 ; 38
                DB      039,039,039,040,040,040,041,041,041 ; 41
                DB      042,042,042,043,043,043,044,044,044 ; 44
                DB      045,045,045,046,046,046,047,047,047 ; 47
                DB      048,048,048,049,049,049,050,050,050 ; 50
                DB      051,051,051,052,052,052,053,053,053 ; 53
                DB      054,054,054,055,055,055,056,056,056 ; 56
                DB      057,057,057,058,058,058,059,059,059 ; 59
                DB      060,060,060,061,061,061,062,062,062 ; 62
                DB      063,063,063                         ; 63
                DB      000,000,000,001,001,001,002,002,002 ; 2
                DB      003,003,003,004,004,004,005,005,005 ; 5
                DB      006,006,006,007,007,007,008,008,008 ; 8 
                DB      009,009,009,010,010,010,011,011,011 ; 11
                DB      012,012,012,013,013,013,014,014,014 ; 14
                DB      015,015,015,016,016,016,017,017,017 ; 17
                DB      018,018,018,019,019,019,020,020,020 ; 20
                DB      021,021,021,022,022,022,023,023,023 ; 23
                DB      024,024,024,025,025,025,026,026,026 ; 26
                DB      027,027,027,028,028,028,029,029,029 ; 29
                DB      030,030,030,031,031,031,032,032,032 ; 32
                DB      033,033,033,034,034,034,035,035,035 ; 35
                DB      036,036,036,037,037,037,038,038,038 ; 38
                DB      039,039,039,040,040,040,041,041,041 ; 41
                DB      042,042,042,043,043,043,044,044,044 ; 44
                DB      045,045,045,046,046,046,047,047,047 ; 47
                DB      048,048,048,049,049,049,050,050,050 ; 50
                DB      051,051,051,052,052,052,053,053,053 ; 53
                DB      054,054,054,055,055,055,056,056,056 ; 56
                DB      057,057,057,058,058,058,059,059,059 ; 59
                DB      060,060,060,061,061,061,062,062,062 ; 62
                DB      063,063,063                         ; 63
;**********************|
;  storage for DAC     |
;      registers       |
;**********************|
DAC_REGS        DB      768 DUP (00H)   ; Storage for 256
                                ; registers in 3 color bytes per
                                ; register
                DW      0       ; Padding
;
P_DATA  ENDS
;*****************************************************************
;                          code segment
;*****************************************************************
P_CODE  SEGMENT PUBLIC
        ASSUME  CS:P_CODE
;
;*******************************************************************
;                        local procedures
;*******************************************************************
ES_TO_VIDEO     PROC    FAR
; Set ES register to the base address of the video buffer in an
; alphanumeric mode, according to the mode stored in BIOS address
; 0040:0049
; On entry:
;         nothing
; On exit:
;         ES --> segment address of video buffer for alphanumeric
;                display
;         carry clear
        PUSH    AX              ; Save register
        MOV     AX,0040H        ; Segment base of data area
        MOV     ES,AX           ; To extra segment
        MOV     AL,ES:[0049H]   ; Current mode
        CMP     AL,7            ; Code for monochrome card
        JE      SET_MONO_BASE   ; Set base for mono card
;********************|
;     color card     |
;********************|
        MOV     AX,0B800H       ; Color system base
        MOV     ES,AX           ; to ES
        POP     AX              ; Restore entry register
        CLC                     ; No error detection
        RET
;*********************|
;  monochrome card    |
;*********************|
SET_MONO_BASE:
        MOV     AX,0B000H       ; Video base in monochrome systems
        MOV     ES,AX           ; To ES
        POP     AX              ; Restore entry register
        CLC                     ; No error detection
        RET
ES_TO_VIDEO     ENDP
;*****************************************************************
;
ES_TO_APA       PROC    FAR
; Set the ES register to the base address of the video buffer
; in VGA graphics mode
        PUSH    AX              ; Save accumulator
        MOV     AX,0A000H       ; Video segment base for
                                ; graphics modes
        MOV     ES,AX           ; To ES
        POP     AX              ; Restore accumulator
        RET
ES_TO_APA       ENDP
;
;*****************************************************************
;*****************************************************************
;           VGA setup, inquiry, and control primitives
;*****************************************************************
;*****************************************************************
SET_MODE        PROC     FAR
; Set video display mode using BIOS service number 0 of INT 10H
; On entry:
;         AL = number of vide mode to set
; On exit:
;         carry clear
;
        PUSH     AX             ; Save AX
        MOV      AH,0           ; BIOS service request number
        INT      10H
        POP      AX             ; Restore caller's register
        RET
SET_MODE        ENDP
;*******************************************************************
;
GET_MODE        PROC      FAR
; Obtain current video mode using BIOS service number 15 of INT 10H
; On entry:
;         nothing
; On exit:
;         AH = character columns on screen (40 or 80)
;         AL = active video mode
;         BH = active video page
;         carry clear
;
        MOV     AH,15           ; BIOS service request number
        INT     10H
        RET
GET_MODE        ENDP
;******************************************************************
TIME_VRC        PROC     FAR
; Test for start of the vertical retrace cycle of the CRT
; controller
; Bit 7 of the Input Status Register 0 is set if a vertical cycle
; is in progress
        MOV     DX,3C2H         ; Input status register 0
                                ; In VGA color modes
VRC_CLEAR:
        IN      AL,DX           ; Read byte at port
        JMP     SHORT $+2       ; I/O delay
        TEST    AL,10000000B    ; Is bit 7 set?
        JNZ     VRC_CLEAR       ; Wait until bit clear
; At this point the vertical retrace ended. Wait for it to restart
VRC_START:
        IN      AL,DX           ; Read byte at port
        JMP     SHORT $+2       ; I/O delay
        TEST    AL,10000000B    ; Is bit 7 set?
        JZ      VRC_START       ; Wait until bit set
        RET
TIME_VRC        ENDP
;****************************************************************
;
SET_WRITE_MODE  PROC     FAR
; Set the Graphics Controller's Graphic Mode Register to the
; desired write mode
; On entry:
;         AL = write mode requested
; Note: Write mode bits are 0 and 1, all others are preserved
;
; Also set default bit mask
;
        MOV     AH,AL           ; Mode to AH
        MOV     DX,3CEH     ; Graphic Controller Address
                            ; Register
        MOV     AL,5        ; Offset of the Mode Register
        OUT     DX,AL       ; Select this register
        INC     DX          ; Point to data register
; Read contents of mode register
; Write modes bits are 000000xxB
        IN      AL,DX           ; AL has current bit map
        AND     AL,11111100B    ; Clear write mode bits
        OR      AL,AH           ; OR with write mode bits
                                ; passed by caller
; AL now holds bit pattern for mode register
        OUT     DX,AL           ; Set register
; Set Bit Mask Register to default setting
        MOV     DX,3CEH     ; Graphic controller latch
        MOV     AL,8
        OUT     DX,AL       ; Select data register 8
        INC     DX          ; To 3CFH
        MOV     AL,0FFH     ; Default mask
        OUT     DX,AL       ; Load bit mask
        RET
SET_WRITE_MODE  ENDP
;*****************************************************************
;
SET_WRITE_256   PROC     FAR
; Set the Graphics Controller's Graphic Mode Register to the
; desired write mode in 256 colors
; On entry:
;         AL = write mode requested
; Also set default bit mask
;
        PUSH    AX          ; Save mode
        MOV     DX,3CEH     ; Graphic Controller Address
                            ; Register
        MOV     AL,5        ; Offset of the Mode Register
        OUT     DX,AL       ; Select this register
        INC     DX          ; Point to data register
        POP     AX          ; Recover mode in AL
; Set bit 6 to enable 256 colors
        OR      AL,01000000B    ; Mask for bit 6
        OUT     DX,AL       ; Selected
; Set Bit Mask Register to default setting
        MOV     DX,3CEH     ; Graphic controller latch
        MOV     AL,8
        OUT     DX,AL       ; Select data register 8
        INC     DX          ; To 3CFH
        MOV     AL,0FFH     ; Default mask
        OUT     DX,AL       ; Load bit mask
        RET
SET_WRITE_256   ENDP
;*****************************************************************
;
SET_READ_MODE   PROC    FAR
; Set the Graphics Controller Graphic Mode Select register to read
; mode 0 or 1. All other bits are preserved
; On entry:
;         AL = read mode requested
;
; Set bit 3 according to mode
;         Read mode 0 - bit 3 = 0 (AL = 00H)
;         Read mode 1 - bit 3 = 1 (AL = 08H)
;**********************|
;   test entry code    |
;**********************|
        CMP     AL,1            ; If entry value is not 1
        JNE     OK_BIT3         ; read mode 0 is forced
        MOV     AL,08H          ; 00001000B to set bit 3
;**********************|
;    set GC register   |
;**********************|
OK_BIT3:
        MOV     AH,AL           ; Bit mask to AH
        MOV     DX,3CEH         ; Graphic controller address
                                ; register
        MOV     AL,5            ; Offset of the mode register
        OUT     DX,AL           ; Select this register
        INC     DX              ; Point to data register
; Read contents of mode register
        IN      AL,DX           ; AL has current bit map
        AND     AL,11110111B    ; Clear bit 3
        OR      AL,AH           ; OR setting with mask
; AL now holds bit pattern for mode register
        OUT     DX,AL           ; Set register
        RET
SET_READ_MODE       ENDP
;*****************************************************************
;
LOGICAL_MODE    PROC     FAR
; Set the Graphics Controller Data Rotate register to the XOR,
; OR, AND, or NORMAL mode
; On entry:
;         AL = 0 to set data unmodified (NORMAL mode)
;         AL = 1 to AND with latched data
;         AL = 2 to OR with latched data
;         AL = 3 to XOR with latched data
;
        PUSH    CX              ; Save caller's context
        MOV     CL,3            ; Number of bits to shift left
        SHL     AL,CL           ; Shift input 3 bit positions
                                ; to reach bits 3 and 4
        PUSH    AX              ; Save AL in stack
        MOV     DX,03CEH        ; Graphic controller port address
        MOV     AL,3            ; Select data rotate register
        OUT     DX,AL
        INC     DX              ; 03CFH register
        POP     AX              ; Restore AL
        OUT     DX,AL
        POP     CX              ; Restore caller's context
        RET
LOGICAL_MODE    ENDP
;****************************************************************
;
READ_MAPS_18    PROC    FAR
; Procedure to read the contents of the 4 color maps in VGA mode
; number 18
;
; On entry:
;               ES = A000H
;               BX = byte offset into video map
;
; On exit:
;               CL = byte stored in intensity map
;               CH = byte stored in red map
;               DL = byte stored in green map
;               DH = byte stored in blue map
;               AX is destroyed
;
; The code assumes that read mode 0 is set
;
; Preserve entry DI
        PUSH    DI
; Set counter and map selector
        MOV     CX,4        ; Counter for 4 maps to read
        MOV     DI,0        ; Map selector code
;**********************|
;   read four maps     |
;**********************|
READ_IRGB:
; Select map from which to read
        MOV     DX,3CEH     ; Graphic controller address
                            ; register
        MOV     AL,4        ; Read map select register
        OUT     DX,AL       ; Activate
;
        INC     DX          ; Graphic controller = 3CFH
        MOV     AX,DI       ; AL = map selector code DI
        OUT     DX,AL       ; IRGB color map selected
; Read 8 bits from selected map
        MOV     AL,ES:[BX]  ; Get byte from bit plane
        PUSH    AX          ; Store it in the stack
        INC     DI          ; Bump selector to next map
        LOOP    READ_IRGB   ; Execute loop 4 times
; 4 maps are stored in stack
;**********************|
;  move stack data to  |
;    exit registers    |
;**********************|
        POP     AX          ; B map byte in AL
        MOV     DH,AL       ; Move B map byte to DH
        POP     AX          ; G map byte in AL
        MOV     DL,AL       ; Move G map byte to DL
        POP     AX          ; R map byte in AL
        MOV     CH,AL       ; Move R map byte to CH
        POP     AX          ; I map byte in AL
        MOV     CL,AL       ; Move I map byte to CL
; Restore entry DI and exit
        POP     DI
        RET
READ_MAPS_18    ENDP
;****************************************************************
;****************************************************************
;              VGA device drivers for mode number 18
;****************************************************************
;****************************************************************
PIXEL_ADD_18    PROC     FAR
; Address computation from x and y pixel coordinates 
; On entry: 
;               CX = x coordinate of pixel (range 0 to 639) 
;               DX = y coordinate of pixel (range 0 to 479) 
; On exit:
;               BX = byte offset into video buffer 
;               AH = bit mask for the write operation using
;                    VGA write modes 0 or 2
;               AL is preserved
; Save all entry registers 
        PUSH    CX 
        PUSH    DX 
; Compute address
        PUSH    AX          ; Save accumulator 
        PUSH    CX          ; Save x coordinate 
        MOV     AX,DX       ; y coordinate to AX 
        MOV     CX,80       ; Multiplier (80 bytes per row)
        MUL     CX          ; AX = y times 80 
        MOV     BX,AX       ; Free AX and hold in BX 
        POP     AX          ; x coordinate from stack 
; Prepare for division 
        MOV     CL,8        ; Divisor
        DIV     CL          ; AX / CL = quotient in AL and 
                            ; remainder in AH 
; Add in quotient 
        MOV     CL,AH       ; Save remainder in CL 
        MOV     AH,0        ; Clear high byte 
        ADD     BX,AX       ; Offset into buffer to BX 
        POP     AX          ; Restore AX
; Compute bit mask from remainder 
        MOV     AH,10000000B    ; Unit mask for 0 remainder 
        SHR     AH,CL       ; Shift right CL times 
; Restore all entry registers 
        POP     DX 
        POP     CX 
        RET
PIXEL_ADD_18    ENDP
;****************************************************************
;
WRITE_PIX_18    PROC     FAR
; VGA mode number 18 device driver for writing an individual
; pixel or a pixel pattern to the graphics screen
;
; On entry:
;            ES:BX = byte offset into the video buffer
;               AL = pixel color in IRGB format
;               AH = bit pattern to set
;
; This routine assumes that write mode 2 has been set
;
; Note: programs using this procedure will usually precede its
;       call by one to ES_TO_APA (to set the video segment base)
;       and another one to PIXEL_ADD_18 (to obtain the byte offset
;       and pixel mask).
;       This procedure does not reset the default write mode nor
;       the contents of the Bit Mask register
;
        PUSH    DX          ; Save outer loop counter
        PUSH    AX          ; Color byte
        PUSH    AX          ; Twice
; Set Bit Mask Register according to mask in AH
        MOV     DX,3CEH     ; Graphic controller latch
        MOV     AL,8
        OUT     DX,AL       ; Select data register 8
        JMP     SHORT $+2
        INC     DX          ; To 3CFH
        POP     AX          ; AX once from stack
        MOV     AL,AH       ; Bit pattern
        OUT     DX,AL       ; Load bit mask
        JMP     SHORT $+2
; Write color code
        MOV     AL,ES:[BX]  ; Dummy read to load latch
                            ; registers
        POP     AX          ; Restore color code
        MOV     ES:[BX],AL  ; Write the pixel with the
                            ; color code in AL
        POP     DX          ; Restore outer loop counter
        RET
WRITE_PIX_18    ENDP
;
;****************************************************************
;
TILE_ADD_18     PROC    FAR
; Procedure to calculate the coarse-grain address at an 8-by-8
; pixel level in VGA mode number 18
;
; On entry:
;         CH = horizontal tile number (range 0 to 79) = x coordinate
;         CL = vertical tile number (range 0 to 59) = y coordinate
;
; Compute coarse-grain address (in BX) as follows:
;     BX = (CL * 640) + CH
;
; On exit:
;          BX = tile offset into video buffer
;          CX is destroyed
;
        PUSH    AX          ; Save accumulator
        PUSH    DX          ; For word multiply
        PUSH    CX          ; To save CH for addition
        MOV     AX,CX       ; Copy CX in AX
; AL = CL
        MOV     AH,0        ; Clear high byte
        MOV     CX,640            ; Multiplier
        MUL     CX          ; AX * CX results in AX
; The multiplier (640) is the product of 80 tiles columns
; times 8 vertical pixels in each tile row
        POP     CX          ; Restore CH
        POP     DX          ; and DX
        MOV     CL,CH       ; Prepare to add in CH
        MOV     CH,0 
        ADD     AX,CX       ; Add
        MOV     BX,AX       ; Move sum to BX
        POP     AX          ; Restore accumulator
        RET
TILE_ADD_18     ENDP
;
;*****************************************************************
;
WRITE_TILE_18   PROC    FAR
; Write an 8-by-8 pixel block addressed at a coarse-grain level
; On entry:
;            ES:BX = byte offset into the video buffer
;                    (see TILE_ADD_18 procedure)
;               AL = pixel color in IRGB format
;
; This routine assumes that write mode 2 has been set
;
; Note: programs using this procedure will usually precede its
;       call by one to ES_TO_APA (to set the video segment base)
;       and another one to TILE_ADD_18 (to obtain the coarse-grain
;       offset into the video buffer)
;       This procedure does not reset the default write mode nor
;       the contents of the Bit Mask register
;
        PUSH    DX          ; Save caller's context
        PUSH    CX
        PUSH    BX
        PUSH    AX          ; Color code byte
; Set Bit Mask Register to all one bits
        MOV     DX,3CEH     ; Graphic controller latch
        MOV     AL,8
        OUT     DX,AL       ; Select data register 8
        INC     DX          ; To 3CFH
        MOV     AL,0FFH     ; Bit pattern of all ones
        OUT     DX,AL       ; Load bit mask
; Set counter for 8 pixel rows
        MOV     CX,8        ; Counter initialized
        POP     AX          ; Restore color code
;**********************|
;     set 8 pixels     |
;**********************|
SET_EIGHT:
        MOV     AH,ES:[BX]  ; Dummy read to load latch
                            ; registers
        MOV     ES:[BX],AL  ; Write the pixel with the
                            ; color code in AL
        ADD     BX,80       ; Index to next row
        LOOP    SET_EIGHT
; Tile is set
        POP     BX          ; Restore caller's context
        POP     CX
        POP     DX
        RET
WRITE_TILE_18   ENDP
;*****************************************************************
;
READ_PIX_18     PROC    FAR
; Procedure to read a pixel color value in VGA mode 18
; On entry:
;            ES:BX = byte offset into the video buffer
;               AH = bit pattern for mask
; On exit:
;               CL = 4 low bits hold pixel color in IRGB format
;               CH = 0
; Code assumes that read mode 0 has been set
;
; Move bit mask to CH
        MOV     CH,AH       ; CH = bit mask for pixel
;**********************|
;   set-up sequencer   |
;**********************|
; Bit 3 of the sequencer memory mode register must be clear for
; read mode 0 operations
        MOV     DX,3C4H     ; Sequencer address register
        MOV     AL,4        ; Select memory mode register
        OUT     DX,AL       ; Activate
        INC     DX          ; To Sequencer data register base
        IN      AL,DX       ; Read present value
        AND     AL,11110111B    ; Clear bit 3, preserve others
        OUT     DX,AL       ; Write to memory mode register
;***********************|
;   set-up read loop    |
;***********************|
        MOV     AH,4        ; Reset map counter register
        MOV     CL,0        ; Clear pixel color return register
;***********************|
; execute 4 read cycles |
;***********************|
; AH has number for current IRGB map (range 0 to 3)
READ_MAPS:
; Select map from which to read
        MOV     DX,3CEH     ; Graphic controller address
                            ; register
        MOV     AL,4        ; Read map select register
        OUT     DX,AL       ; Activate
        INC     DX          ; Graphic controller = 3CFH
        MOV     AL,AH       ; AH = counter for 4 maps
        DEC     AL          ; Adjust to range 0 to 3
        OUT     DX,AL       ; IRGB color map selected
;***********************|
;      read one byte    |
;***********************|
; Read 8 bits from selected map
        MOV     AL,ES:[BX]  ; Get byte from bit plane
;***********************|
; shift return register |
;***********************|
; Previous color code is in bit 0. The shift operation will free
; the low order bit and mode previous bit codes to higher positions
        SHL     CL,1
;**********************|
;    mask out pixels   |
;**********************|
        AND     AL,CH       ; Pixel mask in CH
        JZ      NO_PIX_SET  ; Jump if no pixel in map
; Pixel was set in bit map
        OR      CL,00000001B    ; Set bit 0 in pixel color
                                ; return register
NO_PIX_SET:
        DEC     AH          ; Bump counter to next map
        JNZ     READ_MAPS   ; Continue if not last map
; 4 low bits in CL hold pixel color in IRGB format
        MOV     CH,0        ; Clear CH
        RET
READ_PIX_18     ENDP
;*****************************************************************
;*****************************************************************
;              VGA device drivers for mode number 19
;*****************************************************************
;*****************************************************************
TWO_BIT_IRGB    PROC    FAR
; Initialize DAC registers for 256-color mode in the following
; format:
;               7 6 5 4 3 2 1 0  <= bits
;               |_| |_| |_| |_|
;                I   R   G   B
;
;**********************|
;     save caller's    |
;       context        |
;**********************|
        PUSH    ES              ; Registers used by routine
        PUSH    AX
        PUSH    BX
        PUSH    CX
        PUSH    DX
        PUSH    SI
        PUSH    DI
        PUSH    BP
;**********************|
;    set ES to CS      |
;**********************|
        MOV     AX,P_DATA       ; Local data segment
        MOV     ES,AX           ; To DS
        ASSUME  ES:P_DATA       ; Assume this DS
;**********************|
;  expand color table  |
;**********************|
; The code segment table named INTENSITY_0 contains the register
; values for 00 intensity bits. The values for 01, 10, and 11
; intensity bits are calculated by adding 9, 18, and 27 to the
; values in the INTENSITY_0 table
        LEA     SI,ES:INTENSITY_0  ; Set pointer to source table
        LEA     DI,ES:INTENSITY_1  ; Pointer to intensity 01 table
        LEA     BX,ES:INTENSITY_2  ; Pointer to intensity 10 table
        LEA     BP,ES:INTENSITY_3  ; Pointer to intensity 11 table
        MOV     CX,192          ; Number of values to move
        MOV     AH,9            ; Constant value to add
DO_TABLES:
        MOV     AL,ES:[SI]      ; Value from intensity 00 table
        ADD     AL,AH           ; Add 9, once
        MOV     ES:[DI],AL      ; Place in intensity 01 table
        ADD     AL,AH           ; Add 18
        MOV     ES:[BX],AL      ; Place in intensity 10 table
        ADD     AL,AH           ; Add 27
        MOV     ES:[BP],AL      ; Place in intensity 11 table
        INC     SI              ; Bump all pointers
        INC     DI
        INC     BX
        INC     BP
        LOOP    DO_TABLES
;**********************|
;  set DAC registers   |
;**********************|
; Change first 64 DAC registers using BIOS service number 16
; of interrupt 10H
        MOV     AH,16           ; Service number
        MOV     AL,18           ; Sub-service to set group of
                                ; DAC color registers
        LEA     DX,ES:INTENSITY_0  ; Pointer to table of 256 DAC
                                ; colors
        MOV     BX,0            ; Start with register
        MOV     CX,256          ; Set all 256 DAC registers
        INT     10H
;**********************|
;   restore caller's   |
;       context        |
;**********************|
        POP     BP              ; Registers in stack
        POP     DI
        POP     SI
        POP     DX
        POP     CX
        POP     BX
        POP     AX
        POP     ES
        RET
TWO_BIT_IRGB    ENDP
;*****************************************************************
;
GRAY_256        PROC    FAR
; Initialize DAC registers for 64 shades of gray, as follows:
;               7 6 5 4 3 2 1 0  <= bits
;               | | |_|_|_|_|_|___ Gray scale in the range 0-63
;               | |
;               |_|_______________ Not significant
;
; Note: This procedure sets four blocks of 64 DAC registers to
;       the same gray scale so that bits 6 and 7 of video color
;       data are ignored
;
;**********************|
;     save caller's    |
;       context        |
;**********************|
        PUSH    ES              ; Registers used by routine
        PUSH    AX
        PUSH    BX
        PUSH    CX
        PUSH    DX
;**********************|
;    set ES to CS      |
;**********************|
        MOV     AX,P_DATA       ; Local data segment
        MOV     ES,AX           ; To DS
        ASSUME  ES:P_DATA       ; Assume this eS
;**********************|
;  set DAC registers   |
;**********************|
; Change first 256 DAC registers using BIOS service number 16
; of interrupt 10H. DAC registers are set to a 64 step gray
; scale, repeated four times
        MOV     AH,16           ; Service number
        MOV     AL,18           ; Sub-service to set group of
                                ; DAC color registers
        LEA     DX,ES:GRAY_MAP  ; Pointer to table of 256 DAC
                                ; colors, in groups of 64 shades
                                ; of gray
        MOV     BX,0            ; Start with register
        MOV     CX,256          ; Set 256 DAC registers
        INT     10H
;**********************|
;   restore caller's   |
;       context        |
;**********************|
        POP     DX              ; Registers in stack
        POP     CX
        POP     BX
        POP     AX
        POP     ES
        RET
GRAY_256        ENDP
;*****************************************************************
;
SUM_TO_GRAY     PROC    FAR
; Uses BIOS service number 16, sub-service number 27, to perform
; a sum-to-gray-shades operation on all 256 DAC color registers
; The entry values in the DAC registers is not preserved
;**********************|
;     save caller's    |
;       context        |
;**********************|
        PUSH    AX      ; Save registers used by routine
        PUSH    BX
        PUSH    CX
; Initialize registers for BIOS service
        MOV     AH,16           ; Service request number
        MOV     AL,27           ; Sub-service is sum-to-gray
        MOV     BX,0            ; Start with first register
        MOV     CX,256          ; All 256 DAC color registers
        INT     10H
;**********************|
;   restore caller's   |
;       context        |
;**********************|
        POP     CX      ; Caller's registers in stack
        POP     BX
        POP     AX
        RET
SUM_TO_GRAY     ENDP
;*****************************************************************
;
SAVE_DAC                PROC    FAR
; Save current values in DAC color registers
; Data is saved in the local segment named P_DATA, under the
; variable name DAC_REGS
;**********************|
;     save caller's    |
;       context        |
;**********************|
        PUSH    ES              ; Registers used by routine
        PUSH    AX
        PUSH    BX
        PUSH    CX
        PUSH    DX
;**********************|
;    set ES to CS      |
;**********************|
        MOV     AX,P_DATA       ; Local data segment
        MOV     ES,AX           ; To DS
        ASSUME  ES:P_DATA       ; Assume this DS
; Initialize registers for BIOS service
        MOV     AH,16           ; Service request number
        MOV     AL,23           ; Sub-service is read DAC group
        MOV     BX,0            ; Start with first register
        MOV     CX,256          ; All 256 DAC color registers
        LEA     DX,ES:DAC_REGS  ; Pointer to storage area
        INT     10H
;**********************|
;   restore caller's   |
;       context        |
;**********************|
        POP     DX              ; Registers in stack
        POP     CX
        POP     BX
        POP     AX
        POP     ES
        RET
SAVE_DAC        ENDP
;*****************************************************************
;
RESTORE_DAC     PROC    FAR
; Restore DAC color registers to values stored by the SAVE_DAC
; procedure
;**********************|
;     save caller's    |
;       context        |
;**********************|
        PUSH    ES              ; Registers used by routine
        PUSH    AX
        PUSH    BX
        PUSH    CX
        PUSH    DX
;**********************|
;    set ES to CS      |
;**********************|
        MOV     AX,P_DATA       ; Local data segment
        MOV     ES,AX           ; To DS
        ASSUME  ES:P_DATA       ; Assume this DS
; Initialize registers for BIOS service
        MOV     AH,16           ; Service request number
        MOV     AL,18           ; Sub-service is write DAC group
        MOV     BX,0            ; Start with first register
        MOV     CX,256          ; All 256 DAC color registers
        LEA     DX,ES:DAC_REGS  ; Pointer to storage area
        INT     10H
;**********************|
;   restore caller's   |
;       context        |
;**********************|
        POP     DX              ; Registers in stack
        POP     CX
        POP     BX
        POP     AX
        POP     ES
        RET
RESTORE_DAC     ENDP
;*****************************************************************
;
PIXEL_ADD_19    PROC FAR
; Address computation for VGA mode number 19
; On entry: 
;               CX = x coordinate of pixel (range 0 to 319) 
;               DX = y coordinate of pixel (range 0 to 199) 
; On exit:
;               BX = offset into video buffer 
;
        PUSH    AX      ; Save caller's context
        PUSH    CX
        PUSH    DX
        PUSH    CX          ; Save x coordinate 
        MOV     AX,DX       ; y coordinate to AX 
        MOV     CX,320      ; Multiplier is 320 bytes per row)
        MUL     CX          ; AX = y times 320
        MOV     BX,AX       ; Free AX and hold in BX 
        POP     AX          ; x coordinate from stack 
        ADD     BX,AX       ; Add in column value
        POP     DX          ; Restore caller's context
        POP     CX
        POP     AX
        RET
PIXEL_ADD_19    ENDP
;*****************************************************************
;
TILE_ADD_19     PROC    FAR
; Procedure to calculate the coarse-grain address at an 8-by-8
; pixel level in VGA mode number 19
;
; On entry:
;         CH = horizontal tile number (range 0 to 39) = x coordinate
;         CL = vertical tile number (range 0 to 24) = y coordinate
;
; Compute coarse-grain address (in BX) as follows:
;     BX = (CL * 2560) + (CH * 8)
;
; On exit:
;          BX = tile offset into video buffer
;          CX is destroyed
;
        PUSH    AX          ; Save accumulator
        PUSH    DX          ; Save caller's DX
        PUSH    CX          ; and entry values
        MOV     AX,CX       ; Copy CX in AX
; AL = CL
        MOV     AH,0        ; Clear high byte
        MOV     CX,2560           ; Multiplier
        MUL     CX          ; AX * CX results in AX
; The multiplier (2560) is the product of 40 tiles columns
; times 8 vertical in each tile, times 8 rows of tiles
        MOV     BX,AX      ; Store partial product in BX
;**********************|
;    add tile number   |
;       times 8        |
;**********************|
        POP     AX          ; Restore entry tile number (AL)
        MOV     AL,AH       ; Transfer column to AL
        MOV     AH,0        ; Clear high byte
        MOV     CL,8        ; Multiplier in formula
        MUL     CL          ; AX = AX * 8
        ADD     BX,AX       ; Add in tile offset
        POP     DX          ; Restore caller's DX
        POP     AX          ; and AX
        RET
TILE_ADD_19     ENDP
;************************************************************
;
FREEZE_DAC      PROC    FAR
; Disable changes to Palette and DAC registers during BIOS mode
; set operations
        MOV     AH,18           ; Alternate select BIOS service
        MOV     BL,49           ; Sub-service to enable or disable
        MOV     AL,1            ; Code to disable changes
        INT     10H
        RET
FREEZE_DAC      ENDP
;
;*****************************************************************
;
THAW_DAC        PROC    FAR
; Enable changes to Palette and DAC registers during BIOS mode
; set operations
        MOV     AH,18           ; Alternate select BIOS service
        MOV     BL,49           ; Sub-service to enable or disable
        MOV     AL,0            ; Code to enable
        INT     10H
        RET
THAW_DAC        ENDP
;
;
P_CODE  ENDS
        END
