; Module name: VGA2.ASM                       Library:   GRAPHSOL.LIB
; Created: April 17, 1992
;********************************************************************
;********************************************************************
;                              VGA2.ASM
;********************************************************************
;********************************************************************
; Description:
; Library module for the following functions:
;     1. VGA mode number 18 and 19 text display primitives
;        a. display text using BIOS service number 9, interrupt 10H
;        b. display text at a screen tile boundary (VGA font)
;        c. display text at a pixel boundary (VGA and PCL fonts)
;        d. load BIOS font to RAM
;        e. load PCL disk font to RAM
;     2. VGA mode number 18 and 19 bit-map and fill primitives
;        a. display a monochrome bit map in mode 18
;        b. display a color bit-map in modes 18 and 19
;        c. initialize full screen (clear)
;        b. initialize a screen rectangle, at tile level
;
; Note: The library routines that are contained in this module can
;       be used from the high level languages C, Pascal, and BASIC
;       as well as in programs coded directly in Assembly Language.
;       Appendix B explains interfacing the library routines with
;       these languages.
;*******************************************************************
;                          ROUTINES
;*******************************************************************
; Public procedures in this module:
; NAME             DESCRIPTION
; Mode number 18 and 19 text display primitives
; GRAPHIC_TEXT     Display text while in APA mode using BIOS service
; Mode number 18 text display primitives
; FINE_TEXT        Display single text line at a pixel boundary
;                  using a RAM resident symmetrical font. Optional
;                  skewing of individual characters. Can be used
;                  by XGA and SuperVGA systems
; COARSE_TEXT      Display multi-line text at a tile boundary using
;                  a RAM-resident symmetrical font
; MULTI_TEXT       Display a multi-line text message at a pixel
;                  boundary using a RAM-resident symmetrical font.
;                  Can be used by XGA and SuperVGA systems
; FINE_TEXTHP      Display single text line at a pixel boundary
;                  using a RAM resident PCL font. Optional skewing
;                  of individual characters. Can be used by XGA and
;                  SuperVGA systems
; Font manipulation primitives
; FONT_TO_RAM      Load BIOS font to RAM
; READ_HPFONT      Load disk-based PCL font to RAM
; Bit-map display primitives
; MONO_MAP_18      Display a monochrome bit-map while in mode 18
; COLOR_MAP_18     Display a color bit-map while in mode 18
; COLOR_MAP_19     Display a color bit-map while in mode 19
; Clear screen primitives
; CLS_18           Clear or initialize screen in mode 18
; CLS_19           Clear or initialize screen in mode 19
; Rectangular fill primitives
; TILE_FILL_18     Fill a screen rectangle (tile level) in mode 18
; TILE_FILL_19     Fill a screen rectangle (tile level) in mode 19
; 
; Local support procedures
; FINE_PATTERN     Display a bit map at a pixel boundary (mode 18)
;*******************************************************************
;                         W A R N I N G S !
;*******************************************************************
; 1. Some routines in this module use the ES register as a pointer
;    to the video buffer base. If the user's software requires ES
;    it must preserve it before calling the routine in this library
;*******************************************************************
;*******************|
;      PUBLICS      |
;*******************|
PUBLIC  GRAPHIC_TEXT
PUBLIC  FINE_TEXT
PUBLIC  COARSE_TEXT
PUBLIC  MULTI_TEXT
PUBLIC  FINE_TEXTHP
PUBLIC  READ_HPFONT
PUBLIC  FONT_TO_RAM
;
PUBLIC  MONO_MAP_18
PUBLIC  COLOR_MAP_18
PUBLIC  COLOR_MAP_19
PUBLIC  CLS_18
PUBLIC  CLS_19
PUBLIC  TILE_FILL_18
PUBLIC  TILE_FILL_19
;*****************************************************************
;            externals in other libraries or modules
;*****************************************************************
; In SOLUTION.LIB
EXTRN   OPEN_FILE:      FAR
EXTRN   READ_128:       FAR
EXTRN   CLOSE_FILE:     FAR
;
; In other modules of GRAPHSOL.LIB
EXTRN   SET_CURSOR:     FAR
EXTRN   PIXEL_ADD_18:   FAR
EXTRN   PIXEL_ADD_19:   FAR
EXTRN   WRITE_PIX_18:   FAR
EXTRN   TILE_ADD_18:    FAR
EXTRN   TILE_ADD_19:    FAR
EXTRN   ES_TO_APA:      FAR
EXTRN   PIXEL_WRITE:    FAR
;
INCLUDELIB SOLUTION
;*****************************************************************
;
P_CODE  SEGMENT PUBLIC
        ASSUME  CS:P_CODE
;
;*****************************************************************
;*****************************************************************
;            support procedures for text primitives
;*****************************************************************
;*****************************************************************
;
; FINE_PATTERN PROCEDURE
; Scratchpad variables for FINE_PATTERN
COUNT_8         DB      0       ; 8 bit counter
BYTES           DB      0       ; Block byte-width storage
X_COORD         DW      0
;
FINE_PATTERN    PROC    FAR
; Display a bit map at a pixel boundary
;
; On entry:
;       CX = x coordinate to start display
;       DX = y coordinate to start display
;       BH = number of bytes per bit map row (map's x dimension)
;       BL = number of rows in bit map (map's y dimension)
;       DI --> start of bit map
;       AL = I R G B color code
;
; Note: this procedure uses the device independent procedure named
;       PIXEL_WRITE which is compatible with VGA, XGA, and
;       SuperVGA systems
;
; Initialize registers
        MOV     CS:COUNT_8,8    ; Prime bit counter
        MOV     CS:BYTES,BH     ; Store byte count
        MOV     CS:X_COORD,CX   ; and x coordinate of block
DISPLAY_BYTE:
        MOV     AH,[DI]         ; Byte to be displayed
TEST_BIT:
        TEST    AH,10000000B    ; Is high bit set?
        JZ      NEXT_BIT        ; Bit not set
; Set the pixel
        PUSH    AX              ; Save entry registers
        PUSH    BX
; The PIXEL_WRITE routine in the VGA3.ASM module allows display
; in XGA and SuperVGA systems
        CALL    PIXEL_WRITE
        POP     BX              ; Restore registers
        POP     AX
NEXT_BIT:
        SAL     AH,1            ; Shift AH to test next bit
        INC     CX              ; Bump x coordinate counter
        DEC     CS:COUNT_8      ; Bit counter
        JZ      NEXT_BYTE1      ; Exit if counter rewound
        JMP     TEST_BIT        ; Continue
; Index to next byte in row, if not at end of row
NEXT_BYTE1:
        DEC     BH              ; Bytes per row counter
        JZ      NEXT_ROW1       ; End of graphic row
BYTE_ENTRY:
        INC     DI              ; Bump graphic code pointer
        MOV     CS:COUNT_8,8    ; Reset bits counter
        JMP     DISPLAY_BYTE
; Index to next row
NEXT_ROW1:
; Test for last graphic row
        DEC     BL              ; Row counter
        JZ      GRAPH_END       ; Done, exit
        MOV     BH,CS:BYTES     ; Reset bytes counter
        INC     DX              ; Bump y coordinate control
        MOV     CX,CS:X_COORD   ; Reset x coordinate control
        JMP     BYTE_ENTRY
GRAPH_END:
        RET
FINE_PATTERN    ENDP
;*****************************************************************
;*****************************************************************
;                   VGA text display primitives
;*****************************************************************
;*****************************************************************
;  CS data for GRAPHIC_TEXT procedure
START_COL       DB      0
;
GRAPHIC_TEXT    PROC    FAR
; Display a message pointed at by DS:DI on the graphic screen
; Message format as follows:
; OFFSET:       CONTENTS:
;       0 ----> Screen row
;       1 ----> Screen column
;       2 ----> Color code
; DISPLAY CODES:
;       00H --> End of message
;       FFH --> End of screen line
; On entry:
;          DI --> Graphic display block
;
; Note: code assumes that write mode 0 is active
;
        MOV     DH,[DI]         ; Get row into DH
        INC     DI              ; Bump pointer
        MOV     DL,[DI]         ; And column into DL
        MOV     CS:START_COL,DL ; Store start column
; Set system cursor using BIOS service number 2 of INT 10H
        CALL    SET_CURSOR      ; Procedure in the ALFAGRAF module
        INC     DI              ; Bump pointer to attribute
        MOV     BL,[DI]         ; Get color code into BL
CHAR_WRITE:
        INC     DI              ; Bump to message start
        MOV     AL,[DI]         ; Get character
        CMP     AL,0FFH         ; End of line ?
        JE      BUMP_ROW        ; Next row
        CMP     AL,0            ; Test for terminator
        JZ      END_TEXT        ; Exit routine
;***********************|
;    display character  |
;***********************|
; Display character in AL and using the color code in BL
        MOV     AH,9            ; BIOS service request number
        MOV     BH,0            ; Page
        MOV     CX,1            ; No repeat
        INT     10H
;**********************|
;     bump cursor      |
;**********************|
        INC     DL
        CALL    SET_CURSOR      ; Procedure in ALFAGRAF module
        JMP     CHAR_WRITE
END_TEXT:
        RET
BUMP_ROW:
        INC     DH              ; Row control register
        MOV     DL,CS:START_COL ; Column control to start column
        CALL    SET_CURSOR      ; Procedure in the ALFAGRAF module
        JMP     CHAR_WRITE
GRAPHIC_TEXT    ENDP
;*******************************************************************
;
; FINE_TEXT PROCEDURE
; Character generator for displaying text messages using a font
; table located in RAM. All bit maps in the font must have the
; same dimensions. Display position is defined at a pixel boundary
; Coding assumes an initialized VGA graphics system
; Notes: 
;       1. This routine is for displaying a single text line. To
;       display a multi-line message use the MULTI_TEXT procedure
;       2. This routine cannot be used to display a Hewlett Packard
;       printer font. Use FINE_TEXTHP instead
;
; Message format:
;  OFFSET  STORAGE   CONTENTS
;    0      word     x coordinate start address (pixel number)
;    2      word     y coordinate start address (pixel number)
;    4      word     x axis pixel spacing between characters
;    6      word     y axis pixel spacing between characters
;    8      byte     I R G B color code
;    9      ----     first character in text message
;
; Font table format:
; Each font table is preceded by two bytes that determine its
; dimensions, as follows:
;   Byte at font table - 1 = number of pixel rows in each bit map
;   Byte at font table - 2 = number of horizontal bytes in each map
;
; Embedded codes:
;                    00H = end of message
;                    Note: no end of line code is implemented
; On entry:
;     DS:SI ==> message
;     DS:DI ==> RAM font table
;
; Note: this procedure uses the device independent procedure named
;       PIXEL_WRITE which is compatible with VGA, XGA, and
;       SuperVGA systems
;
; Code segment data for scratchpad variables
FINE_ADD        DW      0       ; Offset of font table in RAM
MAP_X           DB      0       ; Horizontal bytes in bit map
MAP_Y           DB      0       ; Number of pixel rows in map
X_DISPL         DW      0       ; x character displacement
Y_DISPL         DW      0       ; y character displacement
IRGB_CODE       DB      0
;
FINE_TEXT       PROC    FAR
; Save start address of font table
        MOV     CS:FINE_ADD,DI     ; In memory variable
; Load font dimension into registers and store in variables
        MOV     BH,[DI-2]       ; x dimension into BH (bytes)
        MOV     BL,[DI-1]       ; y dimension into BL (bits)
        MOV     CS:MAP_X,BH        ; Save map dimensions
        MOV     CS:MAP_Y,BL
; Set up pointers
        MOV     CX,WORD PTR [SI]    ; x axis pointer
        ADD     SI,2
        MOV     DX,WORD PTR [SI]    ; y axis pointer
        ADD     SI,2
        MOV     AX,WORD PTR [SI]    ; x axis character spacing
        MOV     CS:X_DISPL,AX       ; Store in variable
        ADD     SI,2
        MOV     AX,WORD PTR [SI]    ; y axis character spacing
        MOV     CS:Y_DISPL,AX
        ADD     SI,2
        MOV     AL,BYTE PTR [SI]    ; IRGB color code
        MOV     CS:IRGB_CODE,AL     ; Store color code
        INC     SI                  ; SI ==> to message text
; At this point
;       CX = x coordinate pixel address to start display
;       DX = y coordinate pixel
;       SI ==> first character in text message
NEXT_FINE:
        MOV     DI,CS:FINE_ADD  ; Offset of RAM font table
        SUB     AX,AX       ; Clear high byte
        MOV     AL,BYTE PTR [SI]     ; ASCII symbol
; Test for embedded control code
        CMP     AL,0        ; End of message
        JNE     NOT_FINE_END
        RET
NOT_FINE_END:
; At this point AL holds ASCII character code, which is the
; character's offset in the font
; Calculate character bit map offset in font
        MUL     CS:MAP_X    ; Offset times number of columns
        MUL     CS:MAP_Y    ; and times number of rows
                            ; in bit map
        ADD     DI,AX       ; DI ==> font to be displayed
; DI ==> bit map to be displayed
        MOV     BL,CS:MAP_Y     ; Map's y dimension
        MOV     BH,CS:MAP_X     ; Map's x dimension
        MOV     AL,CS:IRGB_CODE ; Color code for display
        PUSH    CX              ; Save x and y coordinates
        PUSH    DX
        CALL    FAR PTR FINE_PATTERN
        POP     DX              ; Restore coordinates
        POP     CX
; Index to next character
        ADD     CX,CS:X_DISPL   ; Add x displacement
        ADD     DX,CS:Y_DISPL   ; and y displacement
        INC     SI          ; Bump message pointer
        JMP     NEXT_FINE
;
FINE_TEXT       ENDP
;*****************************************************************
;
; MULTI_TEXT PROCEDURE
; Character generator for displaying text messages using a font
; table located in RAM. All bit maps in the font must have the
; same dimensions. Display position is defined at a pixel boundary
; Coding assumes an initialized VGA graphics system
; Notes: 
;       1. This routine is for displaying a several text lines
;
; Message format:
;  OFFSET  STORAGE   CONTENTS
;    0      word     x coordinate start address (pixel number)
;    2      word     y coordinate start address (pixel number)
;    4      word     x axis pixel spacing between characters
;    6      word     y axis pixel spacing between lines
;    8      byte     I R G B color code
;    9      ----     first character in text message
;
; Font table format:
; Each font table is preceded by two bytes that determine its
; dimensions, as follows:
;   Byte at font table - 1 = number of pixel rows in each bit map
;   Byte at font table - 2 = number of horizontal bytes in each map
;
; Embedded codes:
;                    00H = end of message
;                    FFH = end of text line
; On entry:
;     DS:SI ==> message
;     DS:DI ==> RAM font table
;
; Note: this procedure uses the device independent procedure named
;       PIXEL_WRITE which is compatible with VGA, XGA, and
;       SuperVGA systems
;
; Code segment data for scratchpad variables
; These variables are declared in the FINE_TEXT procedure
; FINE_ADD        DW      0       ; Offset of font table in RAM
; MAP_X           DB      0       ; Horizontal bytes in bit map
; MAP_Y           DB      0       ; Number of pixel rows in map
; X_DISPL         DW      0       ; x character displacement
; IRGB_CODE       DB      0
;
; Variables for this procedure
LINE_SPACE      DW      0       ; Pixel spacing between text
                                ; lines
LINE_X          DW      0       ; x coordinate of line start
;
MULTI_TEXT      PROC    FAR
; Save start address of font table
        MOV     CS:FINE_ADD,DI     ; In memory variable
; Load font dimension into registers and store in variables
        MOV     BH,[DI-2]       ; x dimension into BH (bytes)
        MOV     BL,[DI-1]       ; y dimension into BL (bits)
        MOV     CS:MAP_X,BH        ; Save map dimensions
        MOV     CS:MAP_Y,BL
; Set up pointers
        MOV     CX,WORD PTR [SI]    ; x axis pointer
        MOV     CS:LINE_X,CX        ; Store line start
        ADD     SI,2
        MOV     DX,WORD PTR [SI]    ; y axis pointer
        ADD     SI,2
        MOV     AX,WORD PTR [SI]    ; x axis character spacing
        MOV     CS:X_DISPL,AX       ; Store in variable
        ADD     SI,2
        MOV     AX,WORD PTR [SI]    ; y axis character spacing
        MOV     CS:LINE_SPACE,AX
        ADD     SI,2
        MOV     AL,BYTE PTR [SI]    ; IRGB color code
        MOV     CS:IRGB_CODE,AL     ; Store color code
        INC     SI                  ; SI ==> to message text
; At this point
;       CX = x coordinate pixel address to start display
;       DX = y coordinate pixel
;       SI ==> first character in text message
NEXT_MULT:
        MOV     DI,CS:FINE_ADD  ; Offset of RAM font table
        SUB     AX,AX       ; Clear high byte
        MOV     AL,BYTE PTR [SI]     ; ASCII symbol
; Test for embedded control code
        CMP     AL,0        ; End of message
        JNE     NOT_MULT_END
        RET
NOT_MULT_END:
; Test for end of line code
        CMP     AL,0FFH         ; Code for end of text line
        JNE     NOT_EOL         ; Continue if not end of line
;**********************|
;      end-of-line     |
;**********************|
        ADD     DX,CS:LINE_SPACE        ; Add line spacing factor
        MOV     CX,CS:LINE_X    ; x coordinate to block start
        INC     SI              ; Bump pointer
        JMP     NEXT_MULT       ; Continue with next character
NOT_EOL:
; At this point AL holds ASCII character code, which is the
; character's offset in the font
; Calculate character bit map offset in font
        MUL     CS:MAP_X    ; Offset times number of columns
        MUL     CS:MAP_Y    ; and times number of rows
                            ; in bit map
        ADD     DI,AX       ; DI ==> font to be displayed
; DI ==> bit map to be displayed
        MOV     BL,CS:MAP_Y     ; Map's y dimension
        MOV     BH,CS:MAP_X     ; Map's x dimension
        MOV     AL,CS:IRGB_CODE ; Color code for display
        PUSH    CX              ; Save x and y coordinates
        PUSH    DX
        CALL    FAR PTR FINE_PATTERN
        POP     DX              ; Restore coordinates
        POP     CX
; Index to next character
        ADD     CX,CS:X_DISPL   ; Add x displacement
        INC     SI          ; Bump message pointer
        JMP     NEXT_MULT
MULTI_TEXT      ENDP
;*****************************************************************
;
; FINE_TEXTHP PROCEDURE
; Character generator for displaying text messages using a RAM
; resident printer font in Hewlett Packard PCL format
; Coding assumes an initialized VGA graphic system
;
; Message format:
;  OFFSET  STORAGE   CONTENTS
;    0      word     x coordinate start address (pixel number)
;    2      word     y coordinate start address (pixel number)
;    4      word     x axis pixel spacing between characters
;    6      word     y axis pixel spacing between characters
;    8      byte     I R G B color code
;    9      ----     first character in text message
;
; Embedded codes:
;                    00H = end of message
;                    Note: no end of line code is implemented
; On entry:
;     DS:SI ==> message
;     DS:DI ==> RAM font table
;
; Code segment data for scratchpad variables
; These variables are declared in the FINE_TEXT procedure
; FINE_ADD        DW      0       ; Offset of font table in RAM
; MAP_X           DB      0       ; Horizontal bytes in bit map
; MAP_Y           DB      0       ; Number of pixel rows in map
; X_DISPL         DW      0       ; x character displacement
; Y_DISPL         DW      0       ; y character displacement
; IRGB_CODE       DB      0
; These variables are specific for FINE_TEXTHP
; The values are stored by the INDEX_HP procedure
PIX_WIDTH       DW      0       ; Character width in pixels
PIX_LEFT        DW      0       ; Left offset of bit map
PIX_TOP         DW      0       ; Top offset of bit map
;
; Note: this procedure uses the device independent procedure named
;       PIXEL_WRITE which is compatible with VGA, XGA, and
;       SuperVGA systems
;
FINE_TEXTHP     PROC    FAR
; Save start address of font table
        MOV     CS:FINE_ADD,DI     ; Store font address
; Clear storage variables
        MOV     WORD PTR CS:PIX_WIDTH,0
        MOV     WORD PTR CS:PIX_LEFT,0
        MOV     WORD PTR CS:PIX_TOP,0
; Set up pointers
        MOV     CX,WORD PTR [SI]    ; x axis pointer
        ADD     SI,2
        MOV     DX,WORD PTR [SI]    ; y axis pointer
        ADD     SI,2
        MOV     AX,WORD PTR [SI]    ; x axis character spacing
        MOV     CS:X_DISPL,AX       ; Store in variable
        ADD     SI,2
        MOV     AX,WORD PTR [SI]    ; y axis character spacing
        MOV     CS:Y_DISPL,AX
        ADD     SI,2
        MOV     AL,BYTE PTR [SI]    ; IRGB color code
        MOV     CS:IRGB_CODE,AL     ; Store color code
        INC     SI          ; ==> to message text
; At this point
;       CX = x coordinate pixel address to start display
;       DX = y coordinate pixel
;       SI ==> first character in text message
;**********************|
;   display character  |
;**********************|
NEXT_HPFONT:
        MOV     DI,CS:FINE_ADD    ; Offset of RAM font table
        MOV     BL,[SI]         ; Character to BL
; Test for end of message code
        CMP     BL,00H                  ; Test for end of message
        JNE     NOT_HPEND
        RET
NOT_HPEND:
; Test for a space code (20H)
        CMP     BL,20H                  ; Is it a space
        JNE     NOT_SPACE
;**********************|
; space code processing|
;**********************|
; Since the space does not have a bit map in HP encoding
; Processing is based on the width of the N character
        MOV     BL,'n'          ; Character width desired
        CALL    FAR PTR INDEX_HP    ; Local procedure
; DH holds the number of horizontal bytes in the 'N' character
        JMP     NEXT_HPCHAR     ; Skip display
NOT_SPACE:
        CALL    FAR PTR INDEX_HP        ; Calculate DH, DL, and DI
        MOV     BL,CS:MAP_Y     ; Load registers
        MOV     BH,CS:MAP_X
        MOV     AL,CS:IRGB_CODE ; Color code
; At this point:
;         BH = number of horizontal bytes in the pattern
;         BL = number of pixel rows in the pattern
;         DI ==> character bit map
;         SI ==> message text
;         CX = x coordinate pixel address to start display
;         DX = y coordinate pixel
;         AL = I R G B color code
;
        PUSH    DX              ; Save character coordinates
        PUSH    CX
;**********************|
;   adjust for offset  |
;**********************|
; The character bit map is displaced from the left margin of the
; base line by the values stored in PIX_LEFT and PIX_TOP
        SUB     CX,CS:PIX_LEFT
        SUB     DX,CS:PIX_TOP
        CALL    FAR PTR FINE_PATTERN
        POP     CX              ; Restore coordinates
        POP     DX
NEXT_HPCHAR:
; Index to next character
        ADD     CX,CS:PIX_WIDTH ; Add character pixel
                                ; width
        ADD     CX,CS:X_DISPL   ; Add x displacement
        ADD     DX,CS:Y_DISPL   ; and y displacement
        INC     SI          ; Bump message pointer
        JMP     NEXT_HPFONT
FINE_TEXTHP     ENDP
;*****************************************************************
;
; Code segment data
HANDLE_1        DW      0       ; File handle for disk operations
DATA_BUF        DB      129 DUP (00H)   ; Storage for disk data
START_STR       DB      '*c33E'
                DB      0
;
READ_HPFONT     PROC    FAR
; Read a Hewlett Packard printer font in PCL format from a disk
; file into RAM
; On entry:
;      DS:DX --> ASCIIZ filename for HP soft font
;                (must be in the current path)
;      DS:DI --> RAM buffer for font storage
;
; On exit:
;       Carry clear if font read and stored in buffer
;       Carry set if file not found or disk error
;
; Open font file
        PUSH    DI              ; Save entry pointer
        CALL    OPEN_FILE
        POP     DI
        JNC     OK_OPEN         ; Go if no error code
;**********************|
;    disk open error   |
;**********************|
; Open operation failed. Set carry flag and return to caller
        STC                     ; Signal error
        RET
;**********************|
;   read font into RAM |
;**********************|
OK_OPEN:
        MOV     CS:HANDLE_1,AX     ; Store file handle
NEW_128:
        MOV     BX,CS:HANDLE_1
        LEA     DX,CS:DATA_BUF     ; Buffer for data storage
        PUSH    DS              ; Save data segment
        PUSH    CS              ; Code segment to stack
        POP     DS              ; and to DS
        PUSH    DI              ; Save buffer pointer
        CALL    READ_128        ; Read sector into buffer
        POP     DI              ; Restore buffer pointer
        POP     DS              ; and DS segment
        CMP     AX,0            ; Test for end of file
        JNE     MOVE_128        ; Go if not at end of file
;**********************|
;    end of file       |
;**********************|
        MOV     BX,CS:HANDLE_1     ; File handle
        CALL    CLOSE_FILE      ; Library routine
        JMP     END_OF_READ
;**********************|
;   move sector to     |
;    font buffer       |
;**********************|
; At this point DATA_BUF holds 128 bytes from disk file
; DI --> storage position in the font's buffer
MOVE_128:
        MOV     CX,128          ; Byte counter
        LEA     SI,CS:DATA_BUF  ; Pointer to data just read
PLACE_128:
        MOV     AL,CS:[SI]      ; Byte from DATA_BUF
        MOV     [DI],AL         ; Into font's buffer
        INC     SI              ; Bump pointers
        INC     DI
        LOOP    PLACE_128       ; Continue until all sector read
; At this point the 128 bytes newly read from the disk file are
; stored in the font's buffer
        JMP     NEW_128
END_OF_READ:
        CLC
        RET
READ_HPFONT     ENDP
;*****************************************************************
;
; COARSE_TEXT PROCEDURE
; Character generator for displaying text messages using a font
; table located in RAM. All bit maps in the font must have the
; same dimensions. Display position is defined at a byte boundary
; Coding assumes an initialized VGA graphics system
;
; Message format:
;  OFFSET  STORAGE   CONTENTS
;    0      byte     x coordinate start address (screen column)
;    1      byte     y coordinate start address (screen row)
;    3      byte     I R G B color code
;    4      ----     first character in text message
;
; Font table format:
; Each font table is preceded by two bytes that determine its
; dimensions, as follows:
;   Byte at font table - 1 = number of pixel rows in each bit map
;   Byte at font table - 2 = number of horizontal bytes in each map
;
; Embedded codes:
;                   0FFH = new line
;                    00H = end of message
; On entry:
;     DS:SI ==> message
;     DS:DI ==> RAM font table
;
; Code segment data for scratchpad variables
FONT_ADD        DW      0       ; Offset of font table in RAM
;
COARSE_TEXT     PROC    FAR
; Save start address of font table
        MOV     CS:FONT_ADD,DI     ; In memory variable
; Load font dimension into registers
        MOV     DH,[DI-2]       ; x dimension into DH (bytes)
        MOV     DL,[DI-1]       ; y dimension into DL (bits)
; Set up pointers
        MOV     CH,BYTE PTR [SI]    ; CH = column pointer
        INC     SI
        MOV     CL,BYTE PTR [SI]    ; CL = row pointer
        ADD     CL,CL           ; Double it
        MOV     BX,CX           ; Save original pointers
        INC     SI
        MOV     AL,BYTE PTR [SI]    ; IRGB Color code
        INC     SI          ; ==> to message text
; Look-up fonts and display
NEXT_FONT:
        MOV     DI,CS:FONT_ADD  ; Offset of RAM font table
        PUSH    AX          ; Save color code
        SUB     AX,AX       ; Clear high byte
        MOV     AL,BYTE PTR [SI]     ; ASCII symbol
; Test for embedded control code
        CMP     AL,0        ; End of message
        JNE     NOT_END
        POP     AX          ; Adjust stack
        RET
NOT_END:
        CMP     AL,0FFH     ; New line
        JE      BUMP_LINE
; Not an embedded control
        MUL     DH          ; Offset times number of columns
        MUL     DL          ; and times number of rows
                            ; in bit map
        ADD     DI,AX       ; DI ==> font to be displayed
        POP     AX          ; Restore color code to AL
        CALL    FAR PTR COARSE_PATTERN
        INC     SI          ; Bump message pointer
        ADD     CH,DH       ; Index to next column
        JMP     NEXT_FONT
;
BUMP_LINE:
        POP     AX          ; Adjust stack
        MOV     CX,BX       ; Restore previous start values
        INC     CL          ; Next row
        INC     CL
        MOV     BX,CX       ; Save start values
        INC     SI          ; Bump pointer
        JMP     NEXT_FONT
COARSE_TEXT     ENDP
;*****************************************************************
;
COARSE_PATTERN  PROC    FAR
; Display a bit map at a byte boundary
;
; On entry:
;         CH = Column address (range 0 to 79)
;         CL = Row address (range 0 to 59)
;         DI ==> Pointer to start of bit map
;         DH = x dimension of bit map (in bytes)
;         DL = y dimension of bit map (in bits)
;         AL = I R G B color code
;
;
; Save input register for repeat operations
        PUSH    AX
        PUSH    BX
        PUSH    CX
        PUSH    DX
        PUSH    SI
        PUSH    DI
        CALL    TILE_ADD_18     ; Device driver in VGA1 module
; Set up outer loop in DX for number of rows
        MOV     CL,DH       ; Save number of bytes
        MOV     DH,0        ; Clear high byte of row counter
; Set up inner loop counter for number of bytes per row
; CL already holds number of bytes
        MOV     CH,0        ; Clear high byte of byte counter
; Display loop loading byte code from graphic block
OUT_SHOW:
        PUSH    CX          ; Save inner loop counter
        PUSH    BX          ; And offset into buffer
IN_SHOW:
        MOV     AH,BYTE PTR [DI]     ; Byte code to set
        CALL    WRITE_PIX_18
        INC     BX          ; Bump address to next 8 pixels
        INC     DI          ; Bump byte codes pointer
        LOOP    IN_SHOW
; Horizontal line of byte codes displayed, index to next row
        POP     BX          ; Start of this row
        ADD     BX,80       ; Index to next row
        POP     CX          ; Inner loop counter
        DEC     DX          ; Outer loop counter
        JNZ     OUT_SHOW    ; Continue if not last row
;
; Restore entry parameters
        POP     DI
        POP     SI
        POP     DX
        POP     CX
        POP     BX
        POP     AX
        RET
;
COARSE_PATTERN  ENDP
;*****************************************************************
;
INDEX_HP        PROC    FAR
; Store parameters and find bit map of a character stored in an
; RAM resident printer font in Hewlett Packard PCL format
; On entry:
;      DS:DI ==> start byte of a RAM resident HP soft font
;         BL = ASCII character to find
; On exit:
;       DI --> start of bit map
;       CS:MAP_X = Number of horizontal bytes in the pattern
;       CS:MAP_Y = number of pixel rows in the pattern
;       CS:BIT_WIDTH = character width (in pixels)
;
; Displacements to start of character descriptor
;    Start of font to first header at the string '*c33E'
;    ASCII header blocks end with the letter "W"
;
; Displacements inside the character descriptor area
;    Character width (in pixels)     10 bytes
;    Character height                12 bytes
;    Bit map                         16 bytes
; Note: numerical data in HP soft fonts is stored with the low
;       order element in the high order memory address. This order
;       must be reversed when loaded into a processor register
;**********************|
; index to character   |
;**********************|
        PUSH    AX              ; Save registers
        PUSH    BX
        PUSH    CX
        PUSH    DX
        PUSH    SI
;**********************|
;  skip block leader   |
;**********************|
; Start of the first header block can be found by searching
; for the escape code string '*c33E'
; This string is stored at the variable CS:START_STR
; DS:DI ==> RAM buffer holding the font
LOOK_4_STRING:
        LEA     SI,CS:START_STR         ; Search string
        MOV     CX,5            ; 5 characters in search string
        PUSH    DI              ; Save buffer offset
SEARCH_5:
        MOV     AL,CS:[SI]      ; Get search string character
        CMP     AL,BYTE PTR DS:[DI]     ; Compare with buffer
        JNE     CHECK_NEXT      ; Continue searching
        INC     DI              ; Bump buffer pointer
        INC     SI              ; and search string pointer
        LOOP    SEARCH_5        ; Continue along the search string
; At this point the search has succeeded
        POP     DI              ; Restore start of string in
                                ; buffer
        ADD     DI,5            ; Index to end of string
        JMP     FOUND
CHECK_NEXT:
        POP     DI              ; Buffer offset
        INC     DI              ; Bump to next character
        JMP     LOOK_4_STRING
FOUND:
; DI ==> start of first header block
; BL holds the ASCII code for the desired character
; The first character encoded is the ! symbol
; There is no bit map for a space character (20H)
        SUB     BL,21H          ; Scale ASCII to character table
                                ; displacement
        MOV     CL,BL           ; Set up counter
        MOV     CH,0            ; Clear high byte
        CMP     CX,0            ; Test for first character
        JE      CHAR_FOUND      ; Go if first character
FIND_CHAR:
        CALL    NEXT_CHAR_CODE  ; Local procedure
        LOOP    FIND_CHAR
; At this point DI --> character data area
CHAR_FOUND:
        CALL    FIND_W          ; Skip header block
        MOV     AX,[DI+10]      ; Get character width (in dots)
        XCHG    AH,AL           ; Reverse storage format
        MOV     CS:PIX_WIDTH,AX ; Store character pixel width
        MOV     DL,8            ; Dots per byte
        DIV     DL              ; AX / 8
; AL now holds the character width, in bytes
; Since the unit of storage is 1 byte, the count must be
; incremented if there is a remainder
        CMP     AH,0            ; Test for remainder in byte
                                ; division
        JE      NO_REM          ; No remainder
        INC     AL              ; Compensate for fractional
                                ; units
NO_REM:
        MOV     CS:MAP_X,AL     ; Store in CS variable
        MOV     DX,[DI+12]      ; Character height in bytes
        XCHG    DH,DL           ; Reverse storage format
        MOV     CS:MAP_Y,DL     ; Store in CS variable
;**********************|
;    store character   |
;        offset        |
;**********************|
        MOV     DX,[DI+6]       ; Left offset
        XCHG    DH,DL           ; Reverse storage format
        MOV     CS:PIX_LEFT,DX  ; Store in variable
        MOV     DX,[DI+8]       ; Top offset
        XCHG    DH,DL           ; Reverse storage format
        MOV     CS:PIX_TOP,DX   ; Store in variable
        ADD     DI,16
; DI --> start of bit map
; CS:MAP_X = character width (in bytes)
; CS:MAP_Y = character height (in pixels)
        POP     SI              ; Restore registers
        POP     DX
        POP     CX
        POP     BX
        POP     AX
        RET
INDEX_HP        ENDP
;
NEXT_CHAR_CODE          PROC    NEAR
; Auxiliary procedure to index to the start of the next descriptor
; area in an HP font
; On entry:
;          DI --> the start of a character code
; On exit:
;          DI --> the next character code
;
        PUSH    AX              ; Save general registers
        PUSH    CX
        PUSH    DX
;**********************|
;  find end of ASCII   |
;    header block      |
;**********************|
; The character "W" marks the end of the header block
        CALL    FIND_W
        MOV     AX,[DI+10]      ; Get character width (in dots)
; HP storage format places the low-order byte at the high memory
; address. Storage must be inverted for compatibility with Intel
; storage format
        XCHG    AH,AL           ; Reverse storage format
        MOV     DL,8            ; Dots per byte
        DIV     DL              ; AX / 8
; AL now holds the character width, in bytes
; Since the unit of storage is 1 byte, the count must be
; incremented if there is a remainder
        CMP     AH,0            ; Test for remainder in byte
                                ; division
        JE      NO_REM1         ; No remainder
        INC     AL              ; Compensate for fractional
                                ; units
NO_REM1:
        MOV     DL,AL           ; Byte width to DL
        MOV     AX,[DI+12]      ; Character height in bytes
        XCHG    AH,AL           ; Reverse storage format
        MUL     DL              ; AX = total bytes in bit map
; AX holds the byte length of the character bit map
; SI --> start of descriptor area
        ADD     DI,16           ; Index to start of bit map
        ADD     DI,AX           ; Add in bit map length
        POP     DX              ; Restore general registers
        POP     CX
        POP     AX
        RET
NEXT_CHAR_CODE          ENDP
;
FIND_W  PROC    NEAR
; Auxiliary routine
; The character "W" marks the end of the header block
LOOK_4_W:
        CMP     BYTE PTR [DI],'W'       ; Test for end of header
        JE      HEADER_END              ; Go if end
        INC     DI              ; Bump pointer
        JMP     LOOK_4_W
HEADER_END:
        INC     DI              ; Skip block terminator
        RET
FIND_W  ENDP
;******************************************************************
;
FONT_TO_RAM     PROC    FAR
; Obtain address of VGA character map and load into user's memory
; space
; On entry:
;          AL = ROM character map desired
;               8 = 8 by 8 font
;              14 = 8 by 14 font
;              16 = 8 by 16 font
;         DS:DI --> caller's RAM buffer for font storage
;                   requires 2K for 8 by 8 font and 4K for 8 by 14
;                   and 8 by 16 font
;
        PUSH    DI              ; Save destination address
        CMP     AL,8            ; Test for 8 by 8 font
        JE      LOAD_8X8        ; Go to 8 by 8 load
        CMP     AL,14           ; Test for 8 by 14 font
        JE      LOAD_8X14       ; Go to 8 by 14 load
; Default is 8 by 16 font
        MOV     BH,6            ; BIOS entry parameter
        MOV     CX,2048         ; Word length of font table
        JMP     FONT_ADDR       ; Go to load routine
LOAD_8X8:
        MOV     BH,3            ; BIOS entry parameter
        MOV     CX,1024         ; Word length of font table
        JMP     FONT_ADDR       ; Go to load routine
LOAD_8X14:
        MOV     BH,2            ; BIOS entry parameter
        MOV     CX,2048         ; Word length of font table
;**********************|
;  get font address    |
;**********************|
FONT_ADDR:
        PUSH    CX              ; Save word length counter
        MOV     AH,17           ; BIOS character generator service
        MOV     AL,48           ; Sub-service for font address
        INT     10H
        POP     CX              ; Restore counter
        POP     DI              ; Restore address for font
; At this point ES:BP --> ROM address of character table
        MOV     SI,BP           ; Sep up pointer
MOVE_FONT:
        MOV     AX,ES:[SI]      ; Get 2 byte from ROM
        MOV     [DI],AX         ; Store them in RAM
        ADD     SI,2            ; Bump pointers
        ADD     DI,2
        LOOP    MOVE_FONT
        RET
FONT_TO_RAM     ENDP
;
;*****************************************************************
;*****************************************************************
;                   bit-map and fill primitives
;*****************************************************************
;*****************************************************************
MONO_MAP_18     PROC    FAR
; Display an encoded graphic block in the data segment
; On entry:
;      SI --> Start of control area of graphic block to be
;             displayed
;      BX --> color code for bit map
;
; Register set-up after routine initialization:
;               CX = x coordinate of block start
;               DX = y coordinate of block start
;               BL = number of rows in block
;               BH = number of bytes per block row
;               AL = color code
;               DI --> start of encoded graphic block
;
; Variables in CS (see FINE_PATTERN procedure in this module)

;            COUNT_8 ........ Counter for 8 bits
;            X_COORD ........ Storage for x coordinate
;            BYTES .......... Number of bytes per block row
;
; Initialize registers
        PUSH    BX              ; Save pointer in stack
        MOV     CX,WORD PTR [SI]        ; x coordinate
        MOV     CS:X_COORD,CX              ; Store in variable
        ADD     SI,2                    ; Bump pointer
        MOV     DX,WORD PTR [SI]        ; y coordinate
        ADD     SI,2                    ; Bump pointer
        MOV     BL,BYTE PTR [SI]        ; Number of rows
        INC     SI                      ; Bump pointer
        MOV     BH,BYTE PTR [SI]        ; Bytes per block
        MOV     CS:BYTES,BH                ; Store in variable
        INC     SI                      ; Bump pointer
        XCHG    SI,DI           ; Buffer start to DI
        POP     SI              ; Color code block pointer
        MOV     AL,[SI]         ; Color code to AL
        MOV     CS:COUNT_8,8       ; Prime bit counter
; Color code in the AL register
DISPLAY_ONE:
        MOV     AH,[DI]         ; High nibble to AH
TEST_ONE:
        TEST    AH,10000000B    ; Is high bit set?
        JZ      NEXT_ONE        ; Bit not set
; Set the pixel
        PUSH    AX              ; Save entry registers
        PUSH    BX
        CALL    PIXEL_ADD_18
        CALL    WRITE_PIX_18
        POP     BX              ; Restore registers
        POP     AX
NEXT_ONE:
        SAL     AH,1            ; Shift AH to test next bit
        INC     CX              ; Bump x coordinate counter
        DEC     COUNT_8         ; Bit counter
        JZ      NEXT_BYTE2      ; Exit if counter rewound
        JMP     TEST_ONE        ; Continue
; Index to next byte in row, if not at end of row
NEXT_BYTE2:
        DEC     BH              ; Bytes per row counter
        JZ      NEXT_ROW2       ; End of graphic row
BYTE_ENTRY2:
        INC     DI              ; Bump graphic code pointer
        MOV     CS:COUNT_8,8       ; Reset bits counter
        JMP     DISPLAY_ONE
; Index to next row
NEXT_ROW2:
; Test for last graphic row
        DEC     BL              ; Row counter
        JZ      GRAPH_END2      ; Done, exit
        MOV     BH,CS:BYTES        ; Reset bytes counter
        INC     DX              ; Bump y coordinate control
        MOV     CX,CS:X_COORD      ; Reset x coordinate control
        JMP     BYTE_ENTRY2
GRAPH_END2:
        RET
MONO_MAP_18     ENDP
;*****************************************************************
;
COLOR_MAP_18    PROC    FAR
; Display an encoded graphic block in the data segment
; On entry:
;      SI --> Start of control area preceding image bit map
;      BX --> table of byte-per-pixel color codes for image
;
; Register set-up after routine initialization:
;               CX = x coordinate of block start
;               DX = y coordinate of block start
;               BL = number of rows in block
;               BH = number of bytes per block row
;               DI --> start of encoded graphic block
;               SI --> color table
;
; Variables in CS (see FINE_PATTERN procedure in this module)

;            COUNT_8 ........ Counter for 8 bits
;            X_COORD ........ Storage for x coordinate
;            BYTES .......... Number of bytes per block row
;
; Initialize registers
        PUSH    BX              ; Save color table pointer
        MOV     CX,WORD PTR [SI]        ; x coordinate
        MOV     CS:X_COORD,CX           ; Store in variable
        ADD     SI,2                    ; Bump pointer
        MOV     DX,WORD PTR [SI]        ; y coordinate
        ADD     SI,2                    ; Bump pointer
        MOV     BL,BYTE PTR [SI]        ; Number of rows
        INC     SI                      ; Bump pointer
        MOV     BH,BYTE PTR [SI]        ; Bytes per block
        MOV     CS:BYTES,BH             ; Store in variable
        INC     SI                      ; Bump pointer
        XCHG    SI,DI           ; Buffer start to DI
        POP     SI              ; Color code block pointer
        MOV     CS:COUNT_8,8       ; Prime bit counter
; Color code in the AL register
DISPLAY_COL:
        MOV     AH,[DI]         ; High nibble to AH
TEST_COL:
        MOV     AL,[SI]         ; Color code to AL
        TEST    AH,10000000B    ; Is high bit set?
        JZ      NEXT_COL        ; Bit not set
; Set the pixel
        PUSH    AX              ; Save entry registers
        PUSH    BX
        CALL    PIXEL_ADD_18
        CALL    WRITE_PIX_18
        POP     BX              ; Restore registers
        POP     AX
NEXT_COL:
        SAL     AH,1            ; Shift AH to test next bit
        INC     CX              ; Bump x coordinate counter
        INC     SI              ; Next color code in table
        DEC     COUNT_8         ; Bit counter
        JZ      NEXT_BYTE3      ; Exit if counter rewound
        JMP     TEST_COL        ; Continue
; Index to next byte in row, if not at end of row
NEXT_BYTE3:
        DEC     BH              ; Bytes per row counter
        JZ      NEXT_ROW3       ; End of graphic row
BYTE_ENTRY3:
        INC     DI              ; Bump graphic code pointer
        MOV     CS:COUNT_8,8    ; Reset bits counter
        JMP     DISPLAY_COL
; Index to next row
NEXT_ROW3:
; Test for last graphic row
        DEC     BL              ; Row counter
        JZ      GRAPH_END3      ; Done, exit
        MOV     BH,CS:BYTES     ; Reset bytes counter
        INC     DX              ; Bump y coordinate control
        MOV     CX,CS:X_COORD   ; Reset x coordinate control
        JMP     BYTE_ENTRY3
GRAPH_END3:
        RET
COLOR_MAP_18    ENDP
;
;*****************************************************************
;
COLOR_MAP_19    PROC    FAR
; Display an encoded graphic block in the data segment while in
; mode number 19
; On entry:
;      SI --> Start of control area preceding image bit map
;             (image is encoded in byte-per-pixel format)
;
; Register set-up after routine initialization:
;               CX = x coordinate of block start
;               DX = y coordinate of block start
;               BL = number of rows in block
;               BH = number of bytes per block row
;               DI --> start of encoded graphic block
;               ES = video buffer segment base (A000H)
;
; Variables in CS (see FINE_PATTERN procedure in this module)

;            X_COORD ........ Storage for x coordinate
;            BYTES .......... Number of bytes per block row
;
; Initialize registers
        MOV     CX,WORD PTR [SI]        ; x coordinate
        MOV     CS:X_COORD,CX           ; Store in variable
        ADD     SI,2                    ; Bump pointer
        MOV     DX,WORD PTR [SI]        ; y coordinate
        ADD     SI,2                    ; Bump pointer
        MOV     BL,BYTE PTR [SI]        ; Number of rows
        INC     SI                      ; Bump pointer
        MOV     BH,BYTE PTR [SI]        ; Bytes per block
        MOV     CS:BYTES,BH             ; Store in variable
        INC     SI                      ; Bump pointer
DISPLAY_19:
        MOV     AH,[SI]         ; High nibble to AH
; Set the pixel
        PUSH    BX              ; Save entry registers
        CALL    PIXEL_ADD_19
; BX holds offset into the video buffer
        MOV     ES:[BX],AH      ; Display pixel
        POP     BX              ; Restore registers
NEXT_19:
        INC     CX              ; Bump x coordinate counter
        INC     SI              ; Next color code in table
; Index to next byte in row, if not at end of row
NEXT_BYTE19:
        DEC     BH              ; Bytes per row counter
        JZ      NEXT_ROW19      ; End of graphic row
        JMP     DISPLAY_19
; Index to next row
NEXT_ROW19:
; Test for last graphic row
        DEC     BL              ; Row counter
        JZ      GRAPH_END19     ; Done, exit
        MOV     BH,CS:BYTES     ; Reset bytes counter
        INC     DX              ; Bump y coordinate control
        MOV     CX,CS:X_COORD   ; Reset x coordinate control
        JMP     DISPLAY_19
GRAPH_END19:
        RET
COLOR_MAP_19    ENDP
;
;*****************************************************************
;                        fill primitives
;*****************************************************************
CLS_18  PROC    FAR
; Clear the entire video screen while in VGA mode number 18
; On entry:
;          AL = IRGB color code
; Make sure ES --> video buffer
        CALL    ES_TO_APA       ; Procedure in VGA1 module
        PUSH    AX      ; Save caller's machine
        PUSH    BX
        PUSH    CX
        PUSH    DX
; Set parameters for the TILE_FILL_18 procedure
        MOV     CX,0            ; Start at top left tile
        MOV     DH,80           ; Tiles per screen row
        MOV     DL,60           ; Number of screen rows
        CALL    FAR PTR TILE_FILL_18
        POP     DX              ; Restore caller's context
        POP     CX
        POP     BX
        POP     AX
        RET
CLS_18  ENDP
;*****************************************************************
CLS_19  PROC    FAR
; Clear the entire video screen while in VGA mode number 19
; On entry:
;          AL = IRGB color code
; Make sure ES --> video buffer
        CALL    ES_TO_APA       ; Procedure in VGA1 module
        PUSH    AX      ; Save caller's machine
        PUSH    BX
        PUSH    CX
        PUSH    DX
; Set parameters for the TILE_FILL_18 procedure
        MOV     CX,0            ; Start at top left tile
        MOV     DH,40           ; Tiles per screen row
        MOV     DL,25           ; Number of screen rows
        CALL    FAR PTR TILE_FILL_19
        POP     DX              ; Restore caller's context
        POP     CX
        POP     BX
        POP     AX
        RET
CLS_19  ENDP
;*****************************************************************
TILE_FILL_18    PROC     FAR
; Fill a screen area on a tile (8 by 8 pixels) boundary while in
; mode number 18
; On entry:
;         CH = x coordinate of tile (range 0 to 79)
;         CL = y coordinate of tile (range 0 to 59)
;         DH = Total horizontal tiles to fill
;         DL = Total vertical tiles to file
;         AL = I R G B color to use in fill area
;
; Make sure ES --> video buffer
        CALL    ES_TO_APA       ; Procedure in VGA1 module
; Save entry registers
        PUSH    CX
        PUSH    DX
        PUSH    AX
;
        MOV     AH,0FFH     ; Bit pattern for all bits set
        PUSH    AX
        CALL    TILE_ADD_18     ; Device driver in VGA1 module
; Start address is now set
; Next setup column counter in CX and row counter in DX
        MOV     CL,DH       ; Horizontal fill
        MOV     CH,0        ; CX = horizontal counter
        PUSH    CX          ; Save temporarily
; Row counter = DL * 8
        MOV     AX,DX
        MOV     AH,0        ; Clear high byte
        MOV     CL,8        ; Multiplier
        MUL     CL          ; AX = DL * 8
; Set up outer counter in DX
        MOV     DX,AX
        POP     CX          ; Inner loop counter
        POP     AX          ; AL = color
                            ; AH = 11111111B (all bits set)
OUTER:
        PUSH    CX          ; Save inner loop counter
        PUSH    BX          ; Save start address
INNER:
        CALL    WRITE_PIX_18
        INC     BX          ; Index to next 8-pixel block
        LOOP    INNER
; Index to next column
        POP     BX          ; Restore start of this row
        ADD     BX,80       ; Index to next row start
        POP     CX          ; Restore inner loop counter
        DEC     DX          ; Decrement total rows
        JNZ     OUTER       ; Continue if not end of count
; Restore input registers
        POP     AX
        POP     DX
        POP     CX
        RET
TILE_FILL_18    ENDP
;****************************************************************
;
TILE_FILL_19    PROC     FAR
; Fill a screen area on a tile (8 by 8 pixels) boundary while in
; mode number 19
; On entry:
;         CH = x coordinate of tile (range 0 to 39)
;         CL = y coordinate of tile (range 0 to 29)
;         DH = Total horizontal tiles to fill
;         DL = Total vertical tiles to file
;         AL = I R G B color to use in fill area
;
; Make sure ES --> video buffer
        CALL    ES_TO_APA       ; Procedure in VGA1 module
; Save entry registers
        PUSH    CX
        PUSH    DX
        PUSH    AX
;
        CALL    TILE_ADD_19     ; Device driver in VGA1 module
; Start address is now set
; Next setup column counter in CX and row counter in DX
        MOV     CL,DH       ; Horizontal fill
        MOV     CH,0        ; CX = horizontal counter
        PUSH    CX          ; Save temporarily
; Row counter = DL * 8
        PUSH    AX              ; Save pixel color
        MOV     AX,DX
        MOV     AH,0        ; Clear high byte
        MOV     CL,8        ; Multiplier
        MUL     CL          ; AX = DL * 8
; Set up outer counter in DX
        MOV     DX,AX
        POP     AX              ; Restore pixel color (in AL)
        POP     CX          ; Inner loop counter
OUTER_19:
        PUSH    CX          ; Save inner loop counter
        PUSH    BX          ; Save offset of row start
INNER_19:
        PUSH    CX              ; Save again
        MOV     CX,8            ; Counter for one tile
TILE_19:
        MOV     ES:[BX],AL      ; Write the pixel
        INC     BX          ; Index to next 8-pixel block
        LOOP    TILE_19         ; Do one 8-pixel horizontal unit
        POP     CX              ; Restore x tile counter
        LOOP    INNER_19
; Index to next column
        POP     BX              ; Restore start of row offset
        ADD     BX,320      ; Index to next row start
        POP     CX          ; Restore inner loop counter
        DEC     DX          ; Decrement total rows
        JNZ     OUTER_19    ; Continue if not end of count
; Restore input registers
        POP     AX
        POP     DX
        POP     CX
        RET
TILE_FILL_19    ENDP
;****************************************************************
;
P_CODE  ENDS
        END
