; Module name: VGA3.ASM                       Library:   GRAPHSOL.LIB
; Created: April 24, 1992
;********************************************************************
;********************************************************************
;                              VGA3.ASM
;********************************************************************
;********************************************************************
; Description:
; Library module for VGA mode number 18 geometrical primitives
;    1. Calculate coordinates for the followings figures
;        a. straight line by two points (Bresenham's algorithm)
;        b. straight line by its slope
;        c. circle
;        d. ellipse
;        e. hyperbola
;        f. parabola
;    2. Display stored coordinates in VGA, XGA, or SuperVGA
;       video system
;        a. display coordinates in first quadrant
;        b. display coordinates in second quadrant
;        c. display coordinates in third quadrant
;        d. display coordinates in fourth quadrant
;        e. display coordinates in all four quadrants
;    3. Transformations
;        a. rotate stored coordinates
;        b. clip stored coordinates
;    4. Region fill
;        a. Fill a geometrical area of continuous horizontal
;           pixels
;    5. Preliminary operations
;        a. initialize 80x87 or emulator
;        b. set device switch to VGA, XGA, or SVGA
;
; Note: The library routines that are contained in this module can
;       be used from the high level languages C, Pascal, and BASIC
;       as well as in programs coded directly in Assembly Language.
;       Appendix B explains interfacing the library routines with
;       these languages.
;*******************************************************************
;                          ROUTINES
;*******************************************************************
; Public procedures in this module:
; NAME             DESCRIPTION
; Mode number 18 geometrical calculation primitives
; BRESENHAM        Draw a spright line using Bresenham's algorithm
; LINE_BY_SLOPE    Plot and store pixel coordinates of a straight
;                  line defined by its slope
; CIRCLE           Plot and store pixel coordinates of a circle
; ELLIPSE          Idem for ellipse
; PARABOLA         Idem for parabola
; HYPERBOLA        Idem for hyperbola
;
; Mode number 18 figure display primitives
; QUAD_I           Display stored coordinates in the first quadrant
; QUAD_II          Idem in the second quadrant
; QUAD_III         Idem in the third quadrant
; QUAD_IV          Idem in the fourth quadrant
; DO_4_QUADS       General call to the four quadrant display
;                  procedures
; DISPLAY_LINE     Display straight line coordinates
;
; Geometrical transformations primitives
; ROTATE_ON        Enable rotation and set angle
; ROTATE_OFF       Disable rotation
; CLIP_ON          Enable clipping and store clipping limits
; CLIP_OFF         Disable clipping
;
; Region fill
; REGION_FILL      Fill a region of consecutive horizontal pixels
;
; Public auxiliary procedures
; INIT_X87         Initialize 80x87 hardware or emulator
; SET_DEVICE       Select VGA, XGA, or SVGA display function
;
; Local auxiliary procedures of possible use
; DEG_2_RAD        Convert degrees to radian measure
; SINE             Calculate sine
; COSINE           Calculate cosine
; TANGENT          Calculate tangent
;
;*******************************************************************
;*******************|
;      PUBLICS      |
;*******************|
PUBLIC  BRESENHAM
PUBLIC  LINE_BY_SLOPE
PUBLIC  DISPLAY_LINE
PUBLIC  CIRCLE
PUBLIC  ELLIPSE
PUBLIC  PARABOLA
PUBLIC  HYPERBOLA
PUBLIC  QUAD_I
PUBLIC  QUAD_II
PUBLIC  QUAD_III
PUBLIC  QUAD_IV
PUBLIC  DO_4_QUADS
PUBLIC  ROTATE_ON
PUBLIC  ROTATE_OFF
PUBLIC  CLIP_ON
PUBLIC  CLIP_OFF
PUBLIC  INIT_X87
PUBLIC  REGION_FILL
PUBLIC  SET_DEVICE
PUBLIC  PIXEL_WRITE
;*****************************************************************
;            externals in other libraries or modules
;*****************************************************************
; In SOLUTION.LIB
;
; In other modules of GRAPHSOL.LIB
EXTRN   PIXEL_ADD_18:   FAR
EXTRN   WRITE_PIX_18:   FAR
EXTRN   READ_PIX_18:    FAR
EXTRN   ES_TO_APA:      FAR
EXTRN   CP_PLOT:        FAR
;
; In the XGA modules
EXTRN   XGA_PIXEL_2:    FAR
;
; In the SVGA module
EXTRN   SVGA_PIX_105:   FAR
;
INCLUDELIB SOLUTION
;
;*****************************************************************
;
P_CODE  SEGMENT PUBLIC
        ASSUME  CS:P_CODE
;
;*****************************************************************
;*****************************************************************
;                        code segment data
;*****************************************************************
;*****************************************************************
; 80x87 and emulator controls and constants
ROUND_EVEN      DW      13FFH   ; Round to even (default)
;
; Variables for BRESENHAM procedure
END_Y   DW      0        ; Storage for temporary y coordinate
END_X   DW      0        ; Storage for temporary x coordinate
INCR_FOR_Y      DW      0       ; y increment
INCR_FOR_X      DW      0       ; x increment
STRT_HSEGS      DW      0       ; Start of horizontal segment
STRT_VSEGS      DW      0       ; Start of vertical segment
STRT_TOTAL      DW      0       ; Total start values
DIAG_TOTAL      DW      0       ; Total diagonal values
LINE_COLOR      DB      0       ; IRGB color code for line
;
; Coordinates of points
X1        DW   320 ; x coordinate of origin
X2        DW   240 ; y coordinate of origin
Y1        DW   0
Y2        DW   0
;
; Straight line by slope
SLOPE     DW   0    ; Slope in degrees
;
; Origin coordinates for curve display
ORIGIN_X  DW   0    ; x coordinate of origin
ORIGIN_Y  DW   0    ; y coordinate of origin
;
; Scratch-pad for temporary data
THIS_X  DW      0        ; Normalized coordinate of x
THIS_Y  DW      0        ; Normalized coordinate of y
LAST_Y  DW      0        ; Internal control for
                                ; noncontinuous y axis points
; Buffers for storing 1K of x and y coordinates of the first
; quadrant and a counter for the number of points computed
Y_BUFFER        DB      2048 DUP (00H)
X_BUFFER        DB      2048 DUP (00H)
POINTS_CNT      DW      0        ; Number of entries in buffers

; 80x87 constants
NEW87           DW      13FFH   ; New control word
STATUS_WORD     DW      0       ; 8087 status word
ONE_80          DD      180.0   ; Constant
FOUR            DD      4.0     ; Constant
;
; Rotation controls
ROT_ANGLE       DW      0       ; Valid range is 0 to 90 degrees
SINE_ROT        DQ      0       ; Sine of the rotation angle
                                ; in 8087 long real (8 bytes)
COS_ROT         DQ      0       ; Cosine of idem
XY_COOR         DW      0       ; Temporary memory storage for
                                ; x and y coordinates
; Clipping controls
CLIP_TOP        DW      0       ; Highest y allowed
CLIP_BOT        DW      0       ; Lowest y allowed
CLIP_LEFT       DW      0       ; Leftmost x allowed
CLIP_RIGHT      DW      0       ; Rightmost x allowed
CLIPPING        DB      0       ; 0 = clipping disabled
                                ; 1 = clipping enabled
;
; Variables for curve definition
RADIUS          DW      0       ; Radius in pixels
X_AXIS          DW      0       ; x semi-axis of ellipse
Y_AXIS          DW      0       ; y semi-axis of ellipse
FOCUS           DW      0       ; Focus of hyperbola
VERTEX          DW      0       ; Vertex
X_START         DW      0       ; Start x coordinate
;
; Scratchpad
X_ABS           DW      0       ; Absolute coordinate of x
Y_ABS           DW      0       ; Absolute coordinate of y
OPS_CNT         DW      0       ; Operational points counter
                                ; for circle display procedures
B_PARAM         DW      0       ; b parameter in hyperbola
VERTEX2         DW      0       ; Square of VERTEX in hyperbola
;
; Variables used by REGION_FILL procedure
FILL_COL        DB      0       ; IRGB code for fill color
BACK_COL        DB      0       ; IRGB code for background
SEED_X          DW      0       ; x coordinate of seed point
SEED_Y          DW      0       ; y coordinate of seed point
LINE_X          DW      0       ; x coordinate of start point in
                                ; present scan line
; Vertical screen limit for VGA mode number 18
LIMIT_Y         DW      480     ; y coordinate screen limit
;
; High and low limits of polygon
HIGH_X          DW      0       ; x coordinate of highest point
HIGH_Y          DW      0       ; y coordinate of highest point
LOW_Y           DW      0       ; y coordinate of lowest point
;
; Device switches for VGA, XGA, and SuperVGA
DEVICE           DB      'V'    ; Default = VGA
                                ; 'X' = XGA
                                ; 'S' = SVGA
                                ; 'P' = ColorPro plotter
;*****************************************************************
;*****************************************************************
;                      support procedures
;*****************************************************************
;*****************************************************************
PIXEL_WRITE     PROC    FAR
; Auxiliary procedures to display pixel in VGA, XGA, or SuperVGA
; systems or in ColorPro plotter
; On entry:
;         CX = x coordinate of point
;         DX = y coordinate of point
;         CS:DEVICE = 'V', 'X', 'S', or 'P'
;
        CALL    CLIP            ; Local clipping procedure
        JNC     OK_TO_DISPLAY   ; Continue if not clipped
        RET                     ; Do not display points
OK_TO_DISPLAY:
         CMP     CS:DEVICE,'X'    ; Test for XGA device
         JE      DO_IN_XGA        ; Go to XGA section
         CMP     CS:DEVICE,'S'    ; Test for SuperVGA device
         JE      DO_IN_SVGA       ; Go to SVGA section
         CMP     CS:DEVICE,'P'    ; Test for ColorPro device
         JE      DO_IN_CP         ; Go to ColorPro section
; Default handler is for VGA
        PUSH    AX              ; Save entry registers
        PUSH    BX
        CALL    PIXEL_ADD_18    ; Procedures in VGA1 module
        CALL    WRITE_PIX_18
        POP     BX              ; Restore registers
        POP     AX
        RET
;********************|
;   XGA pixel write  |
;********************|
DO_IN_XGA:
        PUSH    AX                ; Save entry registers
        PUSH    CX
        PUSH    DX
        MOV     BX,AX             ; Color to BL
        CALL    XGA_PIXEL_2       ; XGA pixel display routine
                                  ; in XGA2 module
        POP     DX
        POP     CX
        POP     AX
        RET
;********************|
;  SVGA pixel write  |
;********************|
DO_IN_SVGA:
        MOV     BX,AX             ; Pixel color to BL
        CALL    SVGA_PIX_105      ; Pixel display routine in
                                  ; the SVGA module
        MOV     AX,BX             ; Restore registers
        RET
;********************|
;  ColorPro write    |
;********************|
DO_IN_CP:
        CALL    CP_PLOT          ; Procedure in the BITIO module
        RET

PIXEL_WRITE     ENDP
;
;*****************************************************************
ADJACENT        PROC    NEAR
; Test for adjacent y coordinates
; On entry:
;          CS:SI --> buffer holding x coordinates of curve
;          CS:DI --> buffer holding y coordinates of curve
;
        MOV     DX,CS:THIS_Y
        MOV     CX,CS:THIS_X
; Test for first iteration
        CMP     CS:LAST_Y,0        ; LAST_Y = 0 only on the first
                                   ; iteration
        JE      STORE_PIX_XYS   ; No adjacency test required
TEST_4_ADJACENT:
; Condition No. 1.
;               This y < last y minus 1
        MOV     BX,CS:LAST_Y
        DEC     BX              ; Last y minus 1
        CMP     DX,BX           ; Compare to this y
        JB      FILL_IN_PIXEL
; Condition No. 2.
;               This y > last y plus 1
        MOV     BX,CS:LAST_Y
        INC     BX              ; Last y plus 1
        CMP     DX,BX           ; Compare to this y
        JA      FILL_IN_PIXEL
        JMP     STORE_PIX_XYS
; y points are not adjacent. Correct non-adjacency
; BX = last y coordinate minus 1
FILL_IN_PIXEL:
        MOV     CS:[SI],BX      ; Store y coordinate adjacent
                                ; to previous point
        MOV     CS:[DI],CX      ; Store this x coordinate
        ADD     SI,2            ; Bump pointers
        ADD     DI,2
        INC     CS:POINTS_CNT   ; Bump points counter
        MOV     CS:LAST_Y,BX    ; Update to this point
        JMP     TEST_4_ADJACENT
; Store coordinates
STORE_PIX_XYS:
        MOV     CS:[SI],DX      ; Store normalized y coordinate
        MOV     CS:[DI],CX      ; Store normalized x coordinate
; Bump both buffer pointers
        ADD     SI,2
        ADD     DI,2
        INC     CS:POINTS_CNT   ; Bump points counter
        MOV     CS:LAST_Y,DX    ; Update LAST_Y variable
        RET
ADJACENT        ENDP
;********************************************************************
;
GET_SINE        PROC    FAR
; On entry:
;       ST = angle in radian in the range 0 to pi/4
; On exit:
;       ST = sine of angle
; Find sine using formula Sin @ = y / SQR (x^2 + y^2)
        FPTAN           ;   x(adj)    |   y(opp)   |
        FMUL    ST,ST(0);   x^2       |     y      |
        FXCH            ;     y       |     x^2    |
        FLD     ST(0)   ;     y       |     y      |    x^2   |
        FXCH    ST(3)
        FSTP    ST(0)   ;     y       |     x^2    |    y     |
        FMUL    ST,ST(0);     y^2     |     x^2    |    y     |
        FADD    ST,ST(1);  y^2+x^2    |     x^2    |    y     |
        FSQRT           ; SQ(y^2+x^2) |     x^2    |    y     |
        FXCH            ;    x^2      | SQ(y^2+x^2)|    y     |
        FSTP    ST(0)   ; SQ(y^2+x^2) |     y      |
        FXCH            ;     y       | SQ(y^2+x^2)|
        FDIV    ST,ST(1);  sine r     | ? |
        FSTP    ST(1)   ;  sine r     |   empty    |
        RET
GET_SINE        ENDP
;********************************************************************
;
GET_COS PROC    FAR
; On entry:
;       ST = angle in radian in the range 0 to pi/4
; On exit:
;       ST = cosine of angle
; Find cosine using formula Cos @ = x / SQR (x^2 + y^2)
        FPTAN           ;   x(adj)    |   y(opp)   |
        FLD     ST(0)   ;      x      |      x     |     y     |
        FXCH    ST(3)
        FSTP    ST(0)   ;      x      |      y     |     x     |
        FMUL    ST,ST(0);   x^2       |     y      |     x     |
        FXCH            ;     y       |     x^2    |     x     |
        FMUL    ST,ST(0);     y^2     |     x^2    |     x     |
        FADD    ST,ST(1);  y^2+x^2    |     x^2    |     x     |
        FSQRT           ; SQ(y^2+x^2) |     x^2    |     x     |
        FXCH            ;    x^2      | SQ(y^2+x^2)|     x     |
        FSTP    ST(0)   ; SQ(y^2+x^2) |     x      |
        FXCH            ;     x       | SQ(y^2+x^2)|
        FDIV    ST,ST(1);  cosine r   | SQ(y^2+x^2 |
        FSTP    ST(1)   ;  cosine r   |   empty    |
        RET
;
GET_COS ENDP
;********************************************************************
;
DEG_2_RADS      PROC    FAR
; Convert degrees (in ST) to radian
;                       ;    ST(0)    |    ST(1)   |    ST(2) 
                        ;      d      |   empty    |   empty
        FLD     CS:ONE_80 ;   180     |     d      |
        FLDPI           ;    pi       |    180     |     d     |
        FDIV    ST,ST(1);  pi/180     |    180     |     d     |
        FSTP    ST(1)   ;  pi/180     |     d      |
        FMUL    ST,ST(1);     r       |     d      |
        FSTP    ST(1)   ;     r       |   empty    |
; Angle in ST is now in radian measure
        RET
DEG_2_RADS      ENDP
;********************************************************************
;
TANGENT PROC    FAR
; Compute tangent of an angle in the range 0 to pi/2 radian
; On entry:
;       ST = an valid angle in radian
; On exit:
;       ST = tangent of angle
; Formulas for finding tangent of r
;  a) if 0 > r < pi/4 ... tangent = tangent r
;  b) if pi/4 > r < pi/2  tan = cotan (pi/4 - reminder of r-pi/4)
; Find case a or b
; using pi/4 as a modulus for a partial remainder operation
;                       ;    ST(0)    |    ST(1)   |    ST(2) 
                        ;      r      |   empty    |   empty
        FCLEX
        FLDPI           ;    pi       |     r      |
        FDIV    CS:FOUR ;    pi/4     |     r      |
        FXCH            ;     r       |   pi/4     |
        FPREM           ;    REM      |   pi/4     |
        FSTSW   CS:STATUS_WORD
        FWAIT
; Condition code C1 is bit 9 of the status word
; If C1 = 1 then use formula b, else use formula a
; Load high byte of status word
        MOV     AH,BYTE PTR CS:STATUS_WORD+1
        TEST    AH,00000010B    ; Condition code C1
        JNZ     TAN_FORM_B      ; Bit set indicates b)
; Use formula a for tangent
        FSTP    ST(1)   ;    REM      |  empty    |
        FPTAN           ;   y(opp)    |   x(adj)   |
        FXCH            ;     x       |     y      |
        FDIV    ST,ST(1);    TAN r    | ? |
        FSTP    ST(1)   ;    tan r    |   empty    |
        RET
TAN_FORM_B:
;                       |   REM       |   pi/4    |
        FXCH            ;   pi/4      |   REM     |
        FSUB    ST,ST(1); pi/4 - REM  |   REM     |
        FSTP    ST(1)   ; pi/4 - REM  |   empty   |
; Use formula b for tangent
        FSTP    ST(1)   ;    REM      |  empty    |
        FPTAN           ;   y(opp)    |   x(adj)   |
        FDIV    ST,ST(1);  cotan r    | ? |
        FSTP    ST(1)   ;  cotan r    |   empty    |
        RET
TANGENT ENDP
;********************************************************************
;
SINE    PROC    FAR
; Compute sine of an angle in the range 0 to pi/2 radian
; which is 0 to 90 degrees
; On entry:
;       ST = an valid angle in radian
; On exit:
;       ST = sine of angle
; Formulas for finding radius of r
;  a) if 0 > r < pi/4 ... sine = sine r
;  b) if pi/4 > r < pi/2  sine = cos (pi/4 - reminder of r-pi/4)
;
; Find case a or b
; using pi/4 as a modulus for a partial remainder operation
;                       ;    ST(0)    |    ST(1)   |    ST(2) 
                        ;      r      |   empty    |   empty
        FCLEX
        FLDPI           ;    pi       |     r      |
        FDIV    CS:FOUR ;    pi/4     |     r      |
        FXCH            ;     r       |   pi/4     |
        FPREM           ;    REM      |   pi/4     |
        FSTSW   CS:STATUS_WORD
        FWAIT
; Condition code C1 is bit 9 of the status word
; If C1 = 1 then use formula b, else use formula a
; Load high byte of status word
        MOV     AH,BYTE PTR CS:STATUS_WORD+1
        TEST    AH,00000010B    ; Condition code C1
        JNZ     SINE_FOR_B      ; Bit set indicates b)
; Use formula a for sine
        FSTP    ST(1)   ;    REM      |  empty    |
        CALL    GET_SINE
        RET
;
SINE_FOR_B:
;                       |   REM       |   pi/4    |
        FXCH            ;   pi/4      |   REM     |
        FSUB    ST,ST(1); pi/4 - REM  |   REM     |
        FSTP    ST(1)   ; pi/4 - REM  |   empty   |
        CALL    GET_COS
        RET
SINE    ENDP
;********************************************************************
;
COSINE  PROC    FAR
; Compute cosine of an angle in the range 0 to pi/2 radian
; which is 0 to 90 degrees
; On entry:
;       ST = an valid angle in radian
; On exit:
;       ST = sine of angle
; Formulas for finding radius of r
;  a) if 0 > r < pi/4 ... cos = cos r
;  b) if pi/4 > r < pi/2  cos = sine (pi/4 - reminder of r-pi/4)
;
; Find case a or b
; using pi/4 as a modulus for a partial remainder operation
;                       ;    ST(0)    |    ST(1)   |    ST(2) 
                        ;      r      |   empty    |   empty
        FCLEX
        FLDPI           ;    pi       |     r      |
        FDIV    CS:FOUR ;    pi/4     |     r      |
        FXCH            ;     r       |   pi/4     |
        FPREM           ;    REM      |   pi/4     |
        FSTSW   CS:STATUS_WORD
        FWAIT
; Condition code C1 is bit 9 of the status word
; If C1 = 1 then use formula b, else use formula a
; Load high byte of status word
        MOV     AH,BYTE PTR CS:STATUS_WORD+1
        TEST    AH,00000010B    ; Condition code C1
        JNZ     COS_FORM_B      ; Bit set indicates b)
; Use formula a for sine
        FSTP    ST(1)   ;    REM      |  empty    |
        CALL    GET_COS
        RET
COS_FORM_B:
;                       |   REM       |   pi/4    |
        FXCH            ;   pi/4      |   REM     |
        FSUB    ST,ST(1); pi/4 - REM  |   REM     |
        FSTP    ST(1)   ; pi/4 - REM  |   empty   |
        CALL    GET_SINE
        RET
COSINE  ENDP
;****************************************************************
;
ROTATE_ON       PROC    FAR
; Activate the rotate function
; On entry:
;       AX = angle of rotation in the range 0 < angle < 90 degrees
;
; Store angle
        MOV     CS:ROT_ANGLE,AX
; Obtain and store sine and cosine of rotation angles
        FILD    CS:ROT_ANGLE       ; Load angle
        CALL    DEG_2_RADS         ; Convert to radian
        CALL    SINE               ; Compute sine
        FSTP    CS:SINE_ROT        ; Store sine
; Cosine
        FILD    CS:ROT_ANGLE       ; Load angle
        CALL    DEG_2_RADS         ; Convert to radian
        CALL    COSINE             ; Compute cosine function
        FSTP    CS:COS_ROT         ; Store cosine
        RET
ROTATE_ON       ENDP
;*******************************************************************
;
ROTATE_OFF      PROC    FAR
; Deactivate the rotate function
        MOV     CS:ROT_ANGLE,0
        RET
ROTATE_OFF      ENDP
;********************************************************************
;
ROTATE  PROC    NEAR
; Rotate x and y coordinates of the first quadrant of a curve
;
; On entry:
;       ROT_ANGLE = desired angle of rotation (0 to 90 degs.)
;       CX = x coordinate
;       DX = y coordinate
; On exit:
;       CX = rotated x coordinate (x')
;       DX = rotated y coordinate (y')
;
; Formulas for obtaining rotated coordinates x' and y':
;       x' = x cos @ - y sin @
;       y' = y cos @ + x sin @
;
; Test for no rotation angle
        CMP     CS:ROT_ANGLE,0
        JNE     OK_ROTATE
        RET
OK_ROTATE:
; Routine to compute x' and y'
; x' = x cos @ - y sin @
;                         ;    ST(0)    |    ST(1)   |    ST(2) 
                          ;   empty     |   empty    |   empty
        FLD     CS:SINE_ROT;   sin @    |
; Move y coordinate into 8087 stack
        MOV     CS:XY_COOR,DX      ; Store in memory
        FILD    CS:XY_COOR;      y      |    sin @   |
        FMUL    ST,ST(1)  ;  y sin @    |    sin @   |
        FSTP    ST(1)     ;  y sin @    |
        FLD     CS:COS_ROT;    cos @    |   y sin @  |
; Move x coordinate into 8087 stack
        MOV     CS:XY_COOR,CX      ; Store in memory
        FILD    CS:XY_COOR;      x      |    cos @   |  y sin @  |
        FMUL    ST,ST(1)  ;  x cos @    |    cos @   |  y sin @  |
        FSTP    ST(1)     ;  x cos @    |  y sin @   |
        FSUB    ST,ST(1)  ;     x'      |  y sin @   |
        FSTP    ST(1)     ;     x'      |
; x' remains in the 8087 stack during computation of y'
; y' = y cos @ + x sin @
;                         ;    ST(0)    |    ST(1)   |    ST(2) 
                          ;      d      |   empty    |   empty
        FLD     CS:SINE_ROT;   sin @    |     x'     |
; Move x coordinate into 8087 stack
        MOV     CS:XY_COOR,CX
        FILD    CS:XY_COOR;      x      |    sin @   |     x'   |
        FMUL    ST,ST(1)  ;  x sin @    |    sin @   |     x'   |
        FSTP    ST(1)     ;  x sin @    |      x'    |
        FLD     CS:COS_ROT;    cos @    |   x sin @  |     x'   |
; Move y coordinate into 8087 stack
        MOV     CS:XY_COOR,DX
        FILD    CS:XY_COOR;      y      |    cos @   |  x sin @  | x'
        FMUL    ST,ST(1)  ;  y cos @    |    cos @   |  x sin @  | x'
        FSTP    ST(1)     ;  y cos @    |  x sin @   |     x'    |
        FADD    ST,ST(1)  ;     y'      |  x sin @   |     x'    |
        FSTP    ST(1)     ;     y'      |     x'     |
; Store x' and y' in original buffers
        FISTP   CS:XY_COOR         ; y' to memory
        MOV     DX,CS:XY_COOR      ; DX = y'
        FISTP   CS:XY_COOR         ; x' to memory
        MOV     CX,CS:XY_COOR      ; CX = x'
        RET
ROTATE  ENDP
;*****************************************************************
;
CLIP_ON PROC    FAR
; Enable clipping and store clipping limits
; On entry:
;         AX = highest y value allowed
;         BX = lowest y value allowed
;         CX = leftmost x value allowed
;         DX = rightmost x value allowed
; The clipping window is a rectangle defined by two points with
; coordinates x,y and x',y', as follows:
;
; x = AX, y = CX -> +-------------------
;                   |      clipping     |
;                   |      window       |
;                    -------------------+ <- x' = DX, y' = BX
;
        MOV     CS:CLIPPING,1   ; Enable clipping function
; Store clipping parameters
        MOV     CS:CLIP_TOP,AX  ; x coordinate
        MOV     CS:CLIP_BOT,BX  ; y' coordinate
        MOV     CS:CLIP_LEFT,CX ; y coordinate
        MOV     CS:CLIP_RIGHT,DX; x' coordinate
        RET
CLIP_ON ENDP
;*****************************************************************
;
CLIP_OFF        PROC    FAR
; Disable clipping functions
        MOV     CS:CLIPPING,0           ; Code to disable
        RET
CLIP_OFF        ENDP
;
;*****************************************************************
;
CLIP    PROC    NEAR
; Auxiliary procedure to clip figure to stored limits, as follows:
;    CLIP_TOP ....... Highest y allowed
;    CLIP_BOT ....... Lowest y allowed
;    CLIP_LEFT ...... Leftmost x allowed
;    CLIP_RIGHT ..... Rightmost x allowed
;    CLIPPING =  0 if clipping disabled
;                1 if clipping enabled
;
; On entry:
;       CX = x coordinate
;       DX = y coordinate
; On exit:
;       CARRY SET to clip
;       CARRY CLEAR not to clip
;
;**********************|
;  test for no clip    |
;**********************|
        CMP     CS:CLIPPING,0   ; Code for disabled
        JNE     CLIP_IS_ON      ; Go if not zero
NO_CLIP:
        CLC                     ; Code for no clipping
        RET
;**********************|
;    test x limits     |
;**********************|
CLIP_IS_ON:
        CMP     CX,CS:CLIP_LEFT ; Is x too far left?
        JBE     CLIP_IT         ; Go if too far left
        CMP     CX,CS:CLIP_RIGHT        ; Is x too far right?
        JAE     CLIP_IT         ; Go if too far right
        CMP     DX,CS:CLIP_TOP  ; Is y to high?
        JBE     CLIP_IT         ; Go if too high
        CMP     DX,CS:CLIP_BOT  ; Is y too low?
        JAE     CLIP_IT         ; Go if too low
        JMP     NO_CLIP         ; No clip exit
CLIP_IT:
        STC                     ; Code to clip
        RET
CLIP    ENDP
;*****************************************************************
;*****************************************************************
;                       public procedures
;*****************************************************************
;*****************************************************************
SET_DEVICE       PROC     FAR
; Set device switch for VGA, XGA, or SVGA to be used by the
; PIXEL_WRITE routine
; On entry:
;          AL = "V" to set device to VGA 
;               "X" to set device to XGA
;               "S" to set device to SVGA
;               "P" to set device to ColorPro plotter
;
         CMP     AL,'X'           ; Test for XGA code
         JE      DEVICE_XGA       ; Set for XGA
         CMP     AL,'S'           ; Test for SVGA
         JE      DEVICE_SVGA      ; Set for SVGA
         CMP     AL,'P'           ; Test for ColorPro
         JE      DEVICE_CP        ; Set for ColorPro
; Default is VGA
         MOV     CS:DEVICE,'V'    ; Device is VGA
         RET
DEVICE_XGA:
         MOV     CS:DEVICE,'X'    ; XGA
         RET
DEVICE_SVGA:
         MOV     CS:DEVICE,'S'    ; SuperVGA
         RET
DEVICE_CP:
         MOV     CS:DEVICE,'P'    ; ColorPro plotter
         RET
SET_DEVICE       ENDP
;****************************************************************
;
INIT_X87        PROC    FAR
; Initialize 80x87 coprocessor and set rounding control to even
;
        FINIT           ; Initialize instruction
; Set program's operation mode by installing new code word in
; the 80x87 control register
        FLDCW   CS:ROUND_EVEN
        RET
INIT_X87        ENDP
;*****************************************************************
;
BRESENHAM       PROC    FAR
; Routine to draw a straight line using Bresenham's algorithm
; On entry:
;          AL = color of line
;          BX = start point x coordinate
;          CX = start point y coordinate
;          SI = end point x coordinate
;          DI = end point y coordinate
;**********************|
;  entry operations    |
;**********************|
; Move register data to variables
        MOV     CS:LINE_COLOR,AL        ; Store line color
        MOV     CS:ORIGIN_X,BX          ; Store start x
        MOV     CS:ORIGIN_Y,CX          ; Store start y
        MOV     CS:END_X,SI             ; Store end x
        MOV     CS:END_Y,DI             ; Store end y
; Set unit increments for pixel-to-pixel operation
        MOV     CX,1
        MOV     DX,1
; Determine negative or positive slope from difference between
; the y and x coordinates of the start and end points of the line
; This difference is also the length of the line
        MOV     DI,CS:END_Y
        SUB     DI,CS:ORIGIN_Y  ; Length
        JGE     POS_VERTICAL    ; Vertical length is positive
        NEG     DX              ; DX = -1 in 2's complement form
        NEG     DI              ; Make distance positive
POS_VERTICAL:
        MOV     CS:INCR_FOR_Y,DX; Increments on the y axis will
                                ; be positive or negative
; Calculate horizontal distance
        MOV     SI,CS:END_X
        SUB     SI,CS:ORIGIN_X
        JGE     POS_HORZ        ; Horizontal length is positive
        NEG     CX              ; CX = -1 in 2's complement form
        NEG     SI              ; Distance has to be positive
POS_HORZ:
        MOV     CS:INCR_FOR_X,CX ; Increments on the x axis can 
                                 ; also be positive or negative
; Compare the horizontal and vertical lengths of the line to
; determine if the straight segments will be horizontal (if
; this length is greater) or vertical (otherwise)
        CMP     SI,DI            ; SI = horizontal length
                                 ; DI = vertical length
        JGE     HORZ_SEGMENTS
; Vertical length is greater, straight segments are vertical
        MOV     CX,0             ; No horizontal segments
        XCHG    SI,DI            ; Invert lengths
        JMP     SET_CONTROLS
HORZ_SEGMENTS:
        MOV     DX,0             ; No vertical segments
SET_CONTROLS:
        MOV     CS:STRT_HSEGS,CX ; Will be 1 or 0
        MOV     CS:STRT_VSEGS,DX ; Also 1 or 0
; Calculate adjustment factor
        MOV     AX,DI            ; Smaller direction component
        ADD     AX,AX            ; Double the shorter distance
        MOV     CS:STRT_TOTAL,AX ; Straight component total
        SUB     AX,SI            ; Subtract larger direction
                                 ; component
        MOV     BX,AX            ; General component counter
        SUB     AX,SI            ; Calculate
        MOV     CS:DIAG_TOTAL,AX ; Diagonal component total
                                 ; pixels
; Prepare to draw line
        MOV     CX,CS:ORIGIN_X
        MOV     DX,CS:ORIGIN_Y
; SI = the length of the line along the longer axis
        INC     SI
        MOV     AL,CS:LINE_COLOR ; Color code for line
;*********************|
;   draw line points  |
;*********************|
LINE_POINTS:
        DEC     SI               ; Counter for total pixels
        JNZ     PIX_DRAW
        JMP     END_OF_LINE      ; Line is finished
;*********************|
;   display pixel     |
;*********************|
PIX_DRAW:
        CALL    FAR PTR PIXEL_WRITE  ; Routine to set pixel in mode 18
        CMP     BX,0            ; If BX < 0 then straight segment
                                ; diagonal segment otherwise
        JGE        DIAGONAL
; Draw straight line segments
        ADD     CX,CS:STRT_HSEGS ; Increment CX if horizontal
        ADD     DX,CS:STRT_VSEGS ; Increment DX if vertical
        ADD     BX,CS:STRT_TOTAL ; Counter plus adjustment
        JMP     LINE_POINTS
; Draw diagonal segment
DIAGONAL:
        ADD     CX,CS:INCR_FOR_X       ; X direction
        ADD     DX,CS:INCR_FOR_Y       ; Y direction
        ADD     BX,CS:DIAG_TOTAL       ; Adjust counter
        JMP     LINE_POINTS
END_OF_LINE:
        RET
BRESENHAM       ENDP
;
;*****************************************************************
;
LINE_BY_SLOPE   PROC    FAR
; Procedure to plot and store the pixel coordinates of a straight
; line of slope s, located in the fourth quadrant
; The slope must be in the range 0 < s > 90 degrees
; On entry:
;          AX = x coordinate of start point
;          BX = y coordinate of start point
;          CX = x coordinate of end point
;          DX = slope
; Working storage
;          CS:X1    = x coordinate of origin
;          CS:Y1    = y coordinate of origin
;          CS:X2    = x coordinate of end point
;          CS:SLOPE = Slope in degrees
;
; Formula:
;       y = x Tan s
;*********************|
;     preparations    |
;*********************|
; Move register data to variables
        MOV     CS:X1,AX        ; Store start x
        MOV     CS:Y1,BX        ; Store start y
        MOV     CS:X2,CX        ; Store end x
        MOV     CS:SLOPE,DX     ; Store slope
; Set registers and variables to coordinates
        LEA     SI,CS:Y_BUFFER  ; y buffer pointer
        LEA     DI,CS:X_BUFFER  ; x buffer pointer
        MOV     CS:LAST_Y,0     ; First iteration
        MOV     CS:POINTS_CNT,0 ; Reset points counter
; Calculate the normalized x coordinate for the rightmost point
;                              |  ST(0)   |   ST(1)  |   ST(2)  |
        FILD    CS:X1          ;   x1     |
        FILD    CS:X2          ;   x2     |     x1   |
        FSUB    ST,ST(1)       ; x2 - x1  |     x1   |
        FSTP    ST(1)          ; x2 - x1  |  empty   |
; Store in variable for the normalized x coordinate of
; rightmost point
        FIST    THIS_X
        FSTP    ST(0)         ; empty    |
; Obtain and store tangent of slope
;                             |  ST(0)   |   ST(1)  |   ST(2)  |
        FILD    CS:SLOPE      ; s(deg)   |
        CALL    DEG_2_RADS    ; s(rads)  |
        CALL    TANGENT       ; tan s    |
Y_BY_SLOPE:
;                             |  tan s   |
        FILD    CS:THIS_X     ;     x    |  tan s  |
        FMUL    ST,ST(1)      ;     y    |  tan s  |
; Store in variable for normalized y coordinate of this point
        FISTP   CS:THIS_Y     ;  tan s   |
;*********************|
;  test for adjacent  |
;      y values       |
;*********************|
        CALL    ADJACENT      ; Adjacency test procedure
;*********************|
; test for last pixel |
;*********************|
        CMP     CS:THIS_X,0     ; x = 0 must be calculated
        JE      EXIT_SLOPE
        DEC     CS:THIS_X
        JMP     Y_BY_SLOPE
; Adjust 8087 stack registers
EXIT_SLOPE:
        FSTP    ST(0)
        RET
LINE_BY_SLOPE   ENDP
;*****************************************************************
;
DISPLAY_LINE    PROC    FAR
; Display coordinates stored in CS:X_BUFFER and CS:Y_BUFFER in
; VGA mode number 18
; Total number of coordinates is stored in CS:POINTS_CNT
; Setup pointers and counter
        LEA     SI,CS:Y_BUFFER       ; y coordinates
        LEA     DI,CS:X_BUFFER       ; x coordinates
        MOV     CX,CS:POINTS_CNT
        MOV     CS:OPS_CNT,CX        ; Operational counter
DISP_1:
        MOV     CX,CS:X1             ; x coordinate of origin
        MOV     DX,CS:Y1             ; y coordinate of origin
; Add stored values to origin
        ADD     CX,WORD PTR CS:[DI]
        SUB     DX,WORD PTR CS:[SI]
; CS:CX = x coordinate
; CS:DX = y coordinate of point
        PUSH    AX                 ; Save color code
        CALL    FAR PTR PIXEL_WRITE   ; Local procedure
        POP     AX                 ; Restore color code
        ADD     SI,2               ; Bump coordinates pointers
        ADD     DI,2
        DEC     CS:OPS_CNT         ; Operation points counter
        JNZ     DISP_1
        RET
DISPLAY_LINE    ENDP
;*****************************************************************
;
CIRCLE  PROC    FAR
; Routine to plot and store the pixel coordinates of a circular arc
; in the fourth quadrant
;
; On entry:
;                           REGISTER        STORED IN
;       radius of circle       BX           CS:RADIUS
;
; During computations:
;          x coordinate ... CS:THIS_X
;          y coordinate ... CS:THIS_Y
;*********************|
;     preparations    |
;*********************|
; Store variables
        MOV     CS:RADIUS,BX    ; Entry parameter for radius
; Reset counters and controls
        MOV     CS:THIS_X,0           ; Start values for x
        MOV     CS:LAST_Y,0           ; and LAST_Y
; Buffer pointers:
;       SI --> Y values buffer
;       DI --> X values buffer
        LEA     SI,CS:Y_BUFFER
        LEA     DI,CS:X_BUFFER
        MOV     CS:POINTS_CNT,0        ; Reset counter
;*********************|
; calculate y values  |
;*********************|
CIRCLE_Y:
;                              |  ST(0)   |   ST(1)  |
        FILD    CS:THIS_X      ;    x      |
        FMUL    ST,ST(0)       ;    x^2    |
        FILD    CS:RADIUS      ;    r      |    x^2   
        FMUL    ST,ST(0)       ;    r^2    |    x^2   
        FSUB    ST,ST(1)       ; r^2 - x^2 |    x^2
        FSQRT                  ;Rt(r^2-x^2)|    x^2
        FISTP   CS:THIS_Y      ;    x^2    |
        FSTP    ST(0)          ;  EMPTY    |
; Test adjacency condition
        CALL    ADJACENT       ; Library procedure
        INC     CS:THIS_X      ; x increments in a circle's
                               ; fourth quadrant
        CMP     CS:THIS_Y,0    ; Test for end of execution
        JNE     CIRCLE_Y
; At this point all coordinates have been plotted
        RET
CIRCLE  ENDP
;*****************************************************************
;
ELLIPSE         PROC    FAR
; Routine to plot and store the pixel coordinates of an elliptical
; curve in the fourth quadrant
;
; On entry:
;                           REGISTER        STORED IN
;       x semiaxis (M)         BX           CS:X_AXIS
;       y semiaxis (m)         CX           CS:Y_AXIS
;
; During computations:
;          x coordinate ... CS:THIS_X
;          y coordinate ... CS:THIS_Y
;         CS:m = minor axis (X_AXIS or Y_AXIS variables)
;         CS:M = major axis (X_AXIS or Y_AXIS variables)
;
;*********************|
;     preparations    |
;*********************|
; Move data to variables
        MOV     CS:X_AXIS,BX            ; Store x semiaxis
        MOV     CS:Y_AXIS,CX            ; Store y semiaxis
; Reset counters and controls
        MOV     CS:THIS_X,0        ; Start value for x
        MOV     CS:LAST_Y,0        ; and for LAST_Y
; Buffer pointers:
;       SI --> Y values buffer
;       DI --> X values buffer
        LEA     SI,CS:Y_BUFFER
        LEA     DI,CS:X_BUFFER
        MOV     CS:POINTS_CNT,0    ; Reset counter
ELLIPSE_Y:
; Calculate primitive coordinate of y
; First solve x^2 / M^2
;                              |  ST(0)     |   ST(1)  |
        FILD    CS:X_AXIS      ;    M       |
        FMUL    ST,ST(0)       ;    M^2     |
        FILD    CS:THIS_X      ;    x       |     M^2   |
        FMUL    ST,ST(0)       ;    x^2     |     M^2   | 
        FDIV    ST,ST(1)       ; x^2/M^2    |     M^2   |
; Solve 1 - (x^2 / M^2)
        FLD1                   ;    1       | x^2/M^2   | ? |
        FSUB    ST,ST(1)       ;1-(x^2/M^2) | x^2/M^2   | ? |
; Solve  m^2 * [1-(x^2/M^2)]
        FILD    CS:Y_AXIS      ;    m       |1-(x^2/M^2)| ? | ? |
        FMUL    ST,ST(0)       ;    m^2     |1-(x^2/M^2)| ? | ? |
        FMUL    ST,ST(1)       ;m^2 * [1-(x^2/M^2)]|  ? | ? | ? |
; Find square root
        FSQRT                  ;    y       | ? | ? | ? |
        FISTP   CS:THIS_Y      ; Store y in memory
; Adjust stack
        FSTP    ST(0)          ; ? | ? |
        FSTP    ST(0)          ; ? |
        FSTP    ST(0)          ; Stack is empty
; Insure pixel adjacency condition
        CALL    ADJACENT       ; Library procedure
        INC     CS:THIS_X      ; x increments in a clockwise plot
                               ; of the first quadrant
        CMP     CS:THIS_Y,0    ; Test for end of processing
        JNE     ELLIPSE_Y
        RET
ELLIPSE ENDP
;*****************************************************************
;
PARABOLA        PROC    FAR
; Routine to plot and store the pixel coordinates of a parabolic
; curve in the fourth quadrant
;
; On entry:
;                           REGISTER        STORED IN
;    focus of parabola         BX           CS:FOCUS 
;    start x coordinate        CX           CS:X_START
;
; During computations:
;          x coordinate .......... CS:THIS_X
;          y coordinate .......... CS:THIS_Y
; Formula:
;       y = SQR. ROOT (4ax)
;       Y_ABS = SQR. ROOT (4 * FOCUS * X_ABS)
;
;*********************|
;     preparations    |
;*********************|
; Move register data to variables
        MOV     CS:FOCUS,BX     ; Store focus
        MOV     CS:X_START,CX   ; Store start value for x
; Reset counters and controls
        MOV     AX,CS:X_START        ; Start value for X
        MOV     CS:THIS_X,AX
        MOV     CS:LAST_Y,0          ; Reset LAST_Y
; Buffer pointers:
;       SI --> Y values buffer
;       DI --> X values buffer
        LEA     SI,CS:Y_BUFFER
        LEA     DI,CS:X_BUFFER
        MOV     CS:POINTS_CNT,0      ; Reset counter
PARA_Y:
; Calculate primitive coordinate of y
;       y = SQR. ROOT (4ax)
;       THIS_Y = SQR. ROOT (4 * FOCUS * THIS_X)
;                              |   ST(0)    |   ST(1)   |   |   |
        FILD    CS:THIS_X      ;     x      |
        FILD    CS:FOCUS       ;     a      |      x    |
        FMUL    ST,ST(1)       ;    ax      | ? |
        FLD1                   ;    1       |     ax    | ? |
        FADD    ST,ST(0)       ;    2       |     ax    | ? |
        FADD    ST,ST(0)       ;    4       |     ax    | ? |
        FMUL    ST,ST(1)       ;   4ax      | ? | ? |
        FSQRT                  ;    y       | ? | ? |
        FISTP   CS:THIS_Y      ; Store y in memory
; Adjust stack
        FSTP    ST(0)          ; ? |
        FSTP    ST(0)          ; Stack is empty
; Insure pixel adjacency conditions
        CALL    ADJACENT       ; Library procedure
        DEC     CS:THIS_X
        CMP     CS:THIS_Y,0    ; Test for end of processing
        JNE     PARA_Y
        RET
PARABOLA        ENDP
;*****************************************************************
;
HYPERBOLA       PROC    FAR
; Routine to plot and store the pixel coordinates of a hyperbolic
; curve in the fourth quadrant
; On entry:
;                           REGISTER        STORED IN
;    focus of hyperbola        BX           CS:FOCUS 
;    vertex of hyperbola       CX           CS:VERTEX 
;    start x coordinate        DX           CS:X_START
;
; During computations:
;          X coordinate .......... CS:THIS_X
;          Y coordinate .......... CS:THIS_Y
; Scratch-pad variables:
;          Numerator radix ..... CS:B_PARAM
;          Vertex squared ...... CS:VERTEX2
;
;*********************|
;     preparations    |
;*********************|
; Move register data to variables
        MOV     CS:FOCUS,BX     ; Store focus
        MOV     CS:VERTEX,CX    ; Store vertex
        MOV     CS:X_START,DX   ; Store start value for x
; Reset counters and controls
        MOV     AX,CS:X_START        ; Start value for X
        MOV     CS:THIS_X,AX
        MOV     CS:LAST_Y,0          ; Reset LAST_Y
; Buffer pointers:
;       SI --> Y values buffer
;       DI --> X values buffer
        LEA     SI,CS:Y_BUFFER
        LEA     DI,CS:X_BUFFER
        MOV     CS:POINTS_CNT,0      ; Reset counter
; Compute numerator radical from VERTEX and FOCUS
; Solve: B_PARAM = SQR. ROOT (FOCUS^2 - VERTEX^2)
;                             |   ST(0)    |   ST(1)   |   |   |
        FILD    CS:VERTEX     ;     a      |
        FMUL    ST,ST(0)      ;     a^2    |
        FILD    CS:FOCUS      ;     c      |    a^2    |
        FMUL    ST,ST(0)      ;     c^2    |    a^2    |
        FSUB    ST,ST(1)      ;  c^2 - a^2 |    a^2    |
        FSQRT                 ;     b      |    a^2    |
; Store b
        FISTP   CS:B_PARAM    ;    a^2     |
; Store VERTEX^2 for calculations
        FISTP   CS:VERTEX2    ; Stack is empty
HYPER_Y:
; Calculate primitive coordinate of y
;  y = b / a * SQR ROOT (x^2 - a^2)
;  or:
;  Y_ABS = B_PARAM / VERTEX * SQR ROOT (X_ABS^2 - VERTEX2)
;                             |   ST(0)    |   ST(1)   |   |   |
        FILD    CS:VERTEX2    ;   a^2      |
        FILD    CS:THIS_X     ;    x       |    a^2    |
        FMUL    ST,ST(0)      ;    x^2     |    a^2    |
        FSUB    ST,ST(1)      ; x^2-a^2    | ? |
        FSQRT                 ; SR(x^2-a^2)| ? |
        FILD    CS:B_PARAM    ;     b      | # | ? |
        FILD    CS:VERTEX     ;     a      |     b     | # | ? |
        FDIV    ST(1),ST      ;     b      |    b/a    | # | ? |
        FSTP    ST(0)         ;    b/a     | # | ? |
        FMUL    ST,ST(1)      ;    y       | ? | ? |
        FISTP   CS:THIS_Y     ; Store y in memory
; Adjust stack
        FSTP    ST(0)         ; ? |
        FSTP    ST(0)         ; Stack is empty
; Insure pixel adjacency condition
        CALL    ADJACENT      ; Library procedure
        DEC     CS:THIS_X
        CMP     CS:LAST_Y,0   ; Test for end of processing
        JNE     HYPER_Y
        RET
HYPERBOLA       ENDP
;*****************************************************************
;                        display procedures
;*****************************************************************
;
QUAD_I          PROC    FAR
; Display figure with stored coordinates at X_BUFFER and 
; Y_BUFFER with total number of coordinates at POINTS_CNT
; On entry:
;           AL = color code in IRGB format (for mode 18)
;           BX = x coordinate of origin
;           CX = y coordinate of origin
;
; Move entry parameters to variables
        MOV     CS:ORIGIN_X,BX          ; x start
        MOV     CS:ORIGIN_Y,CX          ; y start
        PUSH    AX                      ; Save entry parameters
        PUSH    BX
        PUSH    CX
;
        LEA     SI,CS:Y_BUFFER
        LEA     DI,CS:X_BUFFER
        MOV     CX,CS:POINTS_CNT
        MOV     CS:OPS_CNT,CX
QUAD_1:
        PUSH    AX                 ; Save color code
        MOV     CX,CS:ORIGIN_X     ; X offset of this circle
        MOV     AX,CS:[DI]         ; Computed absolute coordinate
        ADD     CX,AX              ; Add x in quadrant I
        MOV     DX,CS:ORIGIN_Y     ; Y offset of this circle
        MOV     AX,CS:[SI]         ; Computed y abs. coordinate
        SUB     DX,AX              ; Subtract y in quadrant I
        POP     AX
; CX = x coordinate of point
; DX = y coordinate of point
        CALL    ROTATE
        CALL    FAR PTR PIXEL_WRITE  ; Local procedure
        ADD     SI,2
        ADD     DI,2               ; Bump counters forward
        DEC     CS:OPS_CNT         ; Operation points counter
        JNZ     QUAD_1
        POP     CX              ; Restore entry values
        POP     BX
        POP     AX
        RET
QUAD_I          ENDP
;********************************************************************
;
QUAD_IV         PROC    FAR
; Move entry parameters to variables
        MOV     CS:ORIGIN_X,BX          ; x start
        MOV     CS:ORIGIN_Y,CX          ; y start
        PUSH    AX                      ; Save entry parameters
        PUSH    BX
        PUSH    CX
;
        LEA     SI,CS:Y_BUFFER
        LEA     DI,CS:X_BUFFER
        MOV     CX,CS:POINTS_CNT
        MOV     CS:OPS_CNT,CX
; Set pointers to end of buffer
        ADD     CX,CX           ; Two bytes per point
; Adjust pointer to last entry in table
        DEC     CX
        DEC     CX
        ADD     SI,CX
        ADD     DI,CX
;
QUAD_4:
        PUSH    AX              ; Save color code
        MOV     CX,CS:ORIGIN_X     ; X offset of this circle
        MOV     AX,CS:[DI]         ; Computed absolute coordinate
        ADD     CX,AX           ; Add x in quadrant I
        MOV     DX,CS:ORIGIN_Y     ; Y offset of this circle
        MOV     AX,CS:[SI]         ; Computed y absolute
        ADD     DX,AX           ; Subtract y in quadrant I
        POP     AX
;
        CALL    ROTATE
        CALL    FAR PTR PIXEL_WRITE     ; Local procedure
        SUB     SI,2
        SUB     DI,2            ; Bump counters forward
        DEC     CS:OPS_CNT         ; Operation points counter
        JNZ     QUAD_4
        POP     CX              ; Restore entry values
        POP     BX
        POP     AX
        RET
QUAD_IV         ENDP
;********************************************************************
;
QUAD_III        PROC    FAR
; Move entry parameters to variables
        MOV     CS:ORIGIN_X,BX          ; x start
        MOV     CS:ORIGIN_Y,CX          ; y start
        PUSH    AX                      ; Save entry parameters
        PUSH    BX
        PUSH    CX
;
        LEA     SI,CS:Y_BUFFER
        LEA     DI,CS:X_BUFFER
        MOV     CX,CS:POINTS_CNT
        MOV     CS:OPS_CNT,CX
;
QUAD_3:
        PUSH    AX              ; Save color code
        MOV     CX,CS:ORIGIN_X     ; X offset of this circle
        MOV     AX,CS:[DI]         ; Computed absolute
        SUB     CX,AX           ; Sub x in quadrant III
        MOV     DX,CS:ORIGIN_Y     ; Y offset of this circle
        MOV     AX,CS:[SI]         ; Computed y absolute
        ADD     DX,AX           ; Add y in quadrant III
        POP     AX
        CALL    ROTATE
        CALL    FAR PTR PIXEL_WRITE        ; Local procedure
        ADD     SI,2
        ADD     DI,2            ; Bump counters forward
        DEC     CS:OPS_CNT         ; Operation points counter
        JNZ     QUAD_3
        POP     CX              ; Restore entry values
        POP     BX
        POP     AX
        RET
QUAD_III        ENDP
;********************************************************************
;
QUAD_II         PROC    FAR
; Move entry parameters to variables
        MOV     CS:ORIGIN_X,BX          ; x start
        MOV     CS:ORIGIN_Y,CX          ; y start
        PUSH    AX                      ; Save entry parameters
        PUSH    BX
        PUSH    CX
;
        LEA     SI,CS:Y_BUFFER
        LEA     DI,CS:X_BUFFER
        MOV     CX,CS:POINTS_CNT
        MOV     CS:OPS_CNT,CX
; Set pointers to end of buffer
        ADD     CX,CX           ; Two bytes per point
; Adjust pointer to last entry in table
        DEC     CX
        DEC     CX
        ADD     SI,CX
        ADD     DI,CX
;
QUAD_2:
        PUSH    AX              ; Save color code
        MOV     CX,CS:ORIGIN_X     ; X offset of this circle
        MOV     AX,CS:[DI]         ; Computed absolute
        SUB     CX,AX           ; Subtract x in quadrant II
        MOV     DX,CS:ORIGIN_Y     ; Y offset of this circle
        MOV     AX,CS:[SI]         ; Computed y absolute
        SUB     DX,AX           ; Subtract y in quadrant II
        POP     AX
        CALL    ROTATE
        CALL    FAR PTR PIXEL_WRITE        ; Local procedure
        SUB     SI,2
        SUB     DI,2            ; Bump counters forward
        DEC     CS:OPS_CNT         ; Operation points counter
        JNZ     QUAD_2
        POP     CX              ; Restore entry values
        POP     BX
        POP     AX
        RET
QUAD_II         ENDP
;********************************************************************
;
DO_4_QUADS      PROC    FAR
        CALL    QUAD_I
        CALL    QUAD_IV
        CALL    QUAD_III
        CALL    QUAD_II
        RET
DO_4_QUADS      ENDP
;********************************************************************
;
REGION_FILL     PROC    FAR
; Fill a surface with contiguous horizontal elements
; On entry:
;         AL = fill color
;         CX = x coordinate of seed point
;         DX = y coordinate of seed point
;
;**********************|
;  store entry data    |
;**********************|
        MOV     CS:FILL_COL,AL          ; Store color codes
        MOV     CS:SEED_X,CX            ; Store CX
        MOV     CS:SEED_Y,DX            ; and DX
;***********************|
;   store background    |
;        color          |
;***********************|
; Read pixel at seed point coordinates and store value
        CALL    PIXEL_ADD_18    ; Buffer offset to BX
        CALL    READ_PIX_18     ; Procedure in VGA1
; CL now holds color of present pixel
        MOV     CS:BACK_COL,CL  ; Store background in variable
;**********************|
;  find figure low and |
;     high limits      |
;**********************|
        CALL    GET_HIGH        ; Local procedures
        CALL    GET_LOW
; Setup to start fill at highest point
        MOV     CX,HIGH_X       ; x coordinate of high point
        MOV     DX,HIGH_Y       ; y coordinate
        MOV     CS:LINE_X,CX    ; Store this line start
;**********************|
;    scan one line     |
;**********************|
START_POINT:
; End of figure?
        CMP     DX,CS:LOW_Y     ; Test for lowest y coordinate
        JB      DO_LINE         ; Go if not at limit
;**********************|
;     end of fill      |
;**********************|
        RET
; Move to left margin of scan line
DO_LINE:
        CALL    GET_LEFT_LIM    ; Local procedure
        MOV     CS:LINE_X,CX    ; Store this x coordinate
;**********************|
;     read pixel       |
;**********************|
TEST_PIX:
        CALL    BOUNDARY        ; Local procedure
        JZ      NOT_BOUND_PIX   ; Go if pix matches background
;**********************|
;    boundary pixel    |
;  index to next line  |
;**********************|
; Right boundary of scan line has been reached. Index to next
; scan line
        INC     DX              ; y = y + 1
        MOV     CX,CS:LINE_X    ; x coordinate of previous scan
                                ; line
        JMP     START_POINT     ; Next scan
;**********************|
;    not boundary pix  |
;       fill it        |
;**********************|
NOT_BOUND_PIX:
        MOV     AL,CS:FILL_COL  ; Fill color to AL
        CALL    WRITE_PIX_18    ; Procedure in VGA1 module
;**********************|
;      next pixel      |
;**********************|
NEXT_PIXEL:
        INC     CX              ; x = x + 1
        JMP     TEST_PIX        ; Continue
REGION_FILL     ENDP
;*****************************************************************
;
BOUNDARY        PROC    NEAR
; Auxiliary procedure for REGION_FILL
; On exit:
;        Zero flag set if pixel matches the background
;        Zero flag clear if pixel is a boundary
;
        CALL    PIXEL_ADD_18    ; Buffer offset to BX
; Register set-up
;      CX = x coordinate of present pixel
;      DX = y coordinate of present pixel
;      BX = byte offset into video buffer
;      AH = bit mask
        PUSH    AX              ; Save mask and offset
        PUSH    CX
        PUSH    DX
        CALL    READ_PIX_18     ; Procedure in VGA1
; CL now holds color of present pixel
; Compare but hold jump until stack is restored
        CMP     CL,CS:BACK_COL  ; Does pixel match background
        POP     DX              ; Restore address registers
        POP     CX
        POP     AX              ; Restore pixel mask
        RET
BOUNDARY        ENDP
;*****************************************************************
;
GET_HIGH        PROC    NEAR
; Get y coordinate of high vertex of polygon
; On entry:
;       SEED_X = x coordinate of seed point
;       SEED_Y = y coordinate of seed point
;
;**********************|
;      initialize      |
;**********************|
        MOV     CX,CS:LIMIT_Y   ; Highest y coordinate
        MOV     CS:HIGH_Y,CX            ; Set limit
; Load operational registers
        MOV     CX,CS:SEED_X    ; x coordinate of seed to CX
        MOV     DX,CS:SEED_Y    ; y coordinate of seed to DX
;**********************|
; find left line limit |
;**********************|
; Search line to the left until a boundary pixel is found
LEFT_LIM:
        DEC     CX
        CALL    BOUNDARY        ; Test pixel
        JZ      LEFT_LIM        ; Go if not a boundary
; At this point CX holds the x coordinate of a boundary pixel
        INC     CX              ; Move one pix right
;**********************|
;   search upward      |
;**********************|
UP_SEARCH:
        PUSH    DX              ; Save y coordinate
NEXT_UP:
        DEC     DX              ; Next pixel up
        CALL    BOUNDARY        ; Test pixel
        JZ      NEXT_UP         ; Continue up if not boundary
; At this point a vertical boundary has been found (in DX)
        CMP     DX,CS:HIGH_Y            ; Test previous high
        JB      NEW_HIGH        ; Go is new point is higher
; New high point is lower than previous one
        JMP     SKIP_RIGHT      ; Go to next pixel
NEW_HIGH:
; Update highest point stored
        MOV     CS:HIGH_Y,DX    ; Store y of this point
        MOV     CS:HIGH_X,CX    ; Store x coordinate of this point
SKIP_RIGHT:
        POP     DX              ; Restore previous y
;**********************|
;   move to next pix   |
;**********************|
        INC     CX              ; Next horizontal position
        CALL    BOUNDARY        ; End of the scan line?
        JZ      UP_SEARCH       ; Go if not end of scan line
; End of scan line. CS:HIGH_Y holds highest y coordinate
; of polygon. Since this is the coordinate of a boundary, the
; highest fillable y coordinate is one less
        MOV     DX,CS:HIGH_Y            ; Low of boundary
        INC     DX              ; Move down one pix
        MOV     CS:HIGH_Y,DX    ; Store new value
        RET
GET_HIGH        ENDP
;*****************************************************************
;
GET_LOW         PROC    NEAR
; Get y coordinate of low vertex of polygon
; On entry:
;       SEED_X = x coordinate of seed point
;       SEED_Y = y coordinate of seed point
;
;**********************|
;      initialize      |
;**********************|
        MOV     CS:LOW_Y,0              ; Set limit
; Load operational registers
        MOV     CX,CS:SEED_X    ; x coordinate of seed to CX
        MOV     DX,CS:SEED_Y    ; y coordinate of seed to DX
;**********************|
; find left line limit |
;**********************|
; Search line to the left until a boundary pixel is found
        CALL    GET_LEFT_LIM    ; Local procedure
;**********************|
;   search downward    |
;**********************|
DOWN_SEARCH:
        PUSH    DX              ; Save y coordinate
NEXT_DOWN:
        INC     DX              ; Next pixel
        CALL    BOUNDARY        ; Test pixel
        JZ      NEXT_DOWN       ; Continue if not boundary
; At this point a vertical boundary has been found (in DX)
        CMP     DX,CS:LOW_Y             ; Test previous low
        JA      NEW_LOW         ; Go if new point is lower 
; New high point is lower than previous one
        JMP     SKIP_RIGHT_1    ; Go to next pixel
NEW_LOW:
; Update highest point stored
        MOV     CS:LOW_Y,DX            ; This is new low
SKIP_RIGHT_1:
        POP     DX              ; Restore previous y
;**********************|
;   move to next pix   |
;**********************|
        INC     CX              ; Next horizontal position
        CALL    BOUNDARY        ; End of the scan line?
        JZ      DOWN_SEARCH     ; Go if not end of scan line
; End of scan line. CS:LOW_Y  holds lowest y coordinate
; of polygon
        RET
GET_LOW         ENDP
;*****************************************************************
;
GET_LEFT_LIM    PROC    NEAR
; Search line to the left until a boundary pixel is found
; On entry:
;          CX = x coordinate in a scan line
;          DX = y coordinate in a scan line
; On exit:
;          CX = leftmost not-boundary pix in the line
LEFT_LIMIT:
        CALL    BOUNDARY        ; Test pixel
        JZ      NOT_LEFT_LIM    ; Go if not a boundary
; At this point CX holds the x coordinate of a boundary pixel
; Skip all boundary pixels until the first background pix is
; found
FIRST_BACK:
        INC     CX              ; Move one pix right
        CALL    BOUNDARY        ; Is this a boundary
        JNZ     FIRST_BACK      ; Continue right if a boundary
        RET
NOT_LEFT_LIM:
        DEC     CX
        JMP     LEFT_LIMIT      ; Continue
GET_LEFT_LIM    ENDP
;
P_CODE  ENDS
        END
